/* Configuration parser.
 *
 * PegSoft configuration parser library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file conf/parser.h
 * \brief Configuration parser header file.
 */

#ifndef CONF_PARSER_H
#define CONF_PARSER_H

/*************************************************************************/

#include <iosfwd>
#include <map>
#include <string>

#include <misc/string.h>

namespace conf
{

/*************************************************************************/

class ConfDir;
class ConfGlobal;
class ConfLogger;

/*************************************************************************/

/// A parser for config files.

/** Conf is the base class used to parse configuration directives from
  * any istream object. Each directive is represented by a ConfDir
  * object, and must be added to the Conf list of directives so they can
  * be processed when encountered in the stream. 
  */

class Conf
{
	public:
#if DEBUG
		/// Debug message types (used by debug() and debugf()).
		enum
		{
			dbgParse = 1,
			dbgSubstitution = 2,
			dbgProcessLow = 3,
			dbgProcess = 4,
			dbgUserDefined = 256
		};
#endif
		
		/// Directive list type.
		typedef std::map<std::string, ConfDir *, misc::insensitive_less> DirList;
	
		/// Constructor.
		/** Constructs the parser.
		  *
		  * \param[in] logger The ConfLogger object that will be used by
		  * the debug() and error() functions to log messages. It is
		  * not copied, and must therefore remain instantiated until
		  * the parser is destroyed. */
		Conf(ConfLogger &logger) : m_directives(), m_global(0),
				m_logger(&logger), m_context(), m_line(0), m_ready(false),
				m_reloading(false) { }
		/// Copy constructor.
		Conf(Conf const &conf) : m_directives(conf.m_directives),
				m_global(conf.m_global), m_logger(conf.m_logger),
				m_context(conf.m_context), m_line(conf.m_line),
				m_ready(conf.m_ready), m_reloading(conf.m_reloading) { }
		/// Destructor.
		~Conf() { };
		/// Assignment operator.
		Conf &operator=(Conf const &right);
		
		/// Adds a directive to the list of valid directives.
		/** Adds a directive to the list of valid directives. This
		  * directive will then be set when encountered by the
		  * configuration parser. 
		  *
		  * This function assumes that no directive with the same
		  * name is already in the list.
		  *
		  * \param[in] dir The configuration directive to be added. It is
		  * NOT copied and therefore must be valid till the remove()
		  * function is called, or the object is destroyed.
		  * \return true if the directive has been successfully added,
		  * false otherwise.
		  * \sa remove() */
		bool add(ConfDir &dir);
		/// Removes a directive from the list of valid directives.
		/** Removes a directive from the list of valid directives.
		  *
		  * \param[in] dir The configuration directive to be removed.
		  * \return true if the directive was found in the list, false
		  * otherwise.
		  * \sa add() */
		bool remove(ConfDir &dir);
		/// Finds a directive in the list of valid directives.
		/** Finds a directive in the list of valid directives.
		  *
		  * \param[in] name The name of the directive to find.
		  * \return A pointer to the directive with the given
		  * name, or NULL if it's not found.
          * \sa add() */
        ConfDir *find(std::string const &name);
		
		/// Initializes the parser.
		/** Initializes the parser and calls the begin() member
		  * functions for all directives in the list. This function
		  * MUST be called before using the parse() and parse_file()
		  * member functions (you may call those as many times as you
		  * want without having to call begin() again). Don't forget
		  * to call end() when ALL parsing is finished.
		  *
		  * \param[in] global Sets the ConfGlobal object associated
		  * to the parser. It will be automatically unset when
		  * end() is called. The begin() and end() functions of the
		  * global directives will be automatically called. The object
		  * is not copied, and must therefore remain valid until 
		  * end() is called.
		  * \param[in] reload Specifies whether the configuration
		  * is being reloaded (true) or not (false).
		  *
		  * \return true if initialization went well, false
		  * otherwise (in this case, you can't parse anything, and
		  * error() should have been called with an appropriate
		  * error message).
		  * 
		  * \sa ConfDir::begin(), parse(), parsefile(), end(), global(),
		  * error() */
		bool begin(ConfGlobal &global, bool reload);
		/// Parses a configuration file.
		/** Parses the given configuration file. The parser context
		  * will be set to the name of the file.
		  * \pre begin() has been called.
		  * \param[in] filename The file to parse.
		  * \return true if parsing succeeded, false otherwise
		  * (error() should have been called with an appropriate
		  * error message in that case).
		  * \sa begin(), error() */
		bool parse_file(std::string const &filename);
		/// Parses a string.
		/** Parses the given string. 
		  * \param[in] str The string to parse.
		  * \param[in] context An arbitrary string giving indication
		  * on the parsing location.
		  * \return true if parsing succeeded, false otherwise
		  * (error() should have been called with an appropriate
		  * error message in that case).
		  * \sa begin(), error() */
		bool parse_string(std::string const &str, std::string const &context);
		/// Parses a stream.
		/** Parses the given stream.
		  * \pre begin() has been called.
		  * \param[in] is The input stream to parse.
		  * \param[in] context An arbitrary string giving indication
		  * on the parsing location.
		  * \return true if parsing succeeded, false otherwise
		  * (error() should have been called with an appropriate
		  * error message in that case).
		  * \sa begin(), error() */
		bool parse(std::istream &is, std::string const &context);
		/// Finalizes the parser.
		/** Finalizes the parser (does any cleanup task) and calls
		  * the end() member functions of all directives in the list
		  * and all directives of the ConfGlobal object associated
		  * to this object. This function MUST be called after all
		  * parsing has been done through the parse() and parse_file()
		  * functions.
		  *
		  * \return true if finalization went well, false otherwise
		  * (error() should have been called with an appropriate
		  * error message in that case).
		  * \sa ConfDir::end(), parse(), parse_file() */
		bool end();
		
		/// Logs a warning message.
		/** Logs a warning message using logger().warning(), with current
		  * context and line number.
		  * \param[in] msg The warning message to be logged.
		  * \return Nothing. 
		  */
		void warning(std::string const &msg) const;
		/// Like warning() but with a printf-like syntax
		void warningf(char const *format, ...) const FORMAT(printf,2,3);
		
		/// Logs an error message.
		/** Logs an error message using logger().error(), with current
		  * context and line number.
		  * \param[in] err The error message to be logged.
		  * \return Nothing. 
		  */
		void error(std::string const &err) const;
		/// Like error() but with a printf-like syntax
		void errorf(char const *format, ...) const FORMAT(printf,2,3);
#if DEBUG
		/// Logs a debug message.
		/** Logs a debug message using logger().debug(), with current
		  * context and line number.
		  * \param[in] type The debug message type, that may be
		  * used by the function to know when to log the message.
		  * Possible values are:
		  *   - \a Conf::dbgParse: the debug message comes from the parser.
		  *   - \a Conf::dbgSubstitution: the debug message is related to
		  *     variable substitution.
		  *   - \a Conf::dbgProcessLow: the debug message comes from the
		  *     directive processor, and it is believed that there won't
		  *     won't be too many messages of that type.
		  *   - \a Conf::dbgProcess: the debug message comes from the
		  *     directive processor.
		  *   - \a Conf::dbgUserDefined: this value, and values
		  *     higher than it, may be used as custom message types.
		  * \param[in] msg The debug message to be logged.
		  * \return Nothing. */
		void debug(int type, std::string const &msg) const;
		/// Like debug() but with a printf-like syntax
		void debugf(int type, char const *format, ...) const 
				FORMAT(printf,3,4);
#endif

		/// Returns the directives list.
		DirList const &directives() const { return m_directives; }
		/// Returns the ConfGlobal object associated with the parser.
		/** Returns the ConfGlobal object associated with the parser.
		  * This object is used by the parser when it needs to look
		  * for defines, global directives and config files search path.
		  * \return The ConfGlobal object associated with the parser.
		  */
		ConfGlobal &global() const { return *m_global; }
		/// Returns the ConfLogger object associated with the parser.
		/** Returns the ConfLogger object associated with the parser.
		  * This object is used by the debug() and error() member
		  * functions to log messages. */
		ConfLogger &logger() const { return *m_logger; }
		/// Returns the current context.
		/** Returns the current context, i.e. a string that helps identify
		  * the current parsing location (for example, a directive name
		  * or a file).
		  * \return The current context.
		  */ 
		std::string const &context() const { return m_context; }
		/// Same as context() but returns a char const *
		char const *c_context() const { return m_context.c_str(); }
		/// Returns the current line number.
		/** Returns the current line number, relative to the current
		  * context.
		  * \return The current line number.
		  * \sa context()
		  */
		unsigned int line() const { return m_line; }
		/// Returns whether begin(), but not end(), has been called yet.
		bool ready() const { return m_ready; }
		/// Returns whether the configuration is being reloaded.
		bool reloading() const { return m_reloading; }
	private:
		DirList m_directives;
		ConfGlobal *m_global;
		ConfLogger *m_logger;
	
		std::string m_context;
		unsigned int m_line;
		
		bool m_ready;
		bool m_reloading;
		
		bool parse_real(std::istream &is, std::string const &context);
#if DEBUG
		void print_directive(unsigned int count,
				std::string const params[]) const;
#endif
		bool process_directive(unsigned int &count, std::string params[]);
};

/*************************************************************************/

/// Logging interface for the configuration parser.

/** ConfLogger is an interface that provides functions used by the 
  * Conf::debug() and Conf::error() functions, allowing programs to make
  * whatever use they want from the messages sent by the parser.
  *
  * You can't use this class directly; you must inherit it and override
  * its member functions. */

class ConfLogger
{
	public:
		/// Destructor.
		virtual ~ConfLogger() { }
		
		/// Logs a warning message (called by Conf::warning()).
		virtual void warning(std::string const &context, unsigned int line,
				std::string const &msg) = 0;
		/// Logs an error message (called by Conf::error()).
		virtual void error(std::string const &context, unsigned int line,
				std::string const &error) = 0;
#if DEBUG
		/// Logs a debug message (called by Conf::debug()).
		virtual void debug(std::string const &context, unsigned int line,
				int type, std::string const &msg) = 0;
#endif
};

/*************************************************************************/

/// An implementation of ConfLogger using stdout and stderr as output target.

class ConfStdLogger : public ConfLogger
{
	public:
		/// Logs a warning message to stderr.
		virtual void warning(std::string const &context, unsigned int line,
				std::string const &msg);
		/// Logs an error message to stderr.
		virtual void error(std::string const &context, unsigned int line,
				std::string const &error);
#if DEBUG
		/// Logs a debug message to stdout.
		virtual void debug(std::string const &context, unsigned int line,
				int type, std::string const &msg);
#endif
};

/*************************************************************************/

inline Conf &Conf::operator=(Conf const &right)
{
	if (&right != this) 
	{
		m_directives = right.m_directives;
		m_global = right.m_global;
		m_logger = right.m_logger;
		m_context = right.m_context;
		m_line = right.m_line;
		m_ready = right.m_ready;
		m_reloading = right.m_reloading;
	}
	
	return *this;
}

/*************************************************************************/

} /* namespace conf */

/*************************************************************************/

#endif /* CONF_PARSER_H */
