/* Configuration directives.
 *
 * PegSoft configuration parser library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

#include <system/headers.h>

#if HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif

#if HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif

#include <misc/misc.h>

#include "conf/parser.h"
#include "conf/directives.h"
#include "conf/global.h"

using namespace std;
using namespace misc;

namespace conf
{

/*************************************************************************/

ConfDir &ConfDir::operator=(ConfDir const &right)
{
	if (&right != this) 
	{
		m_name = right.m_name;
		m_reloadable = right.m_reloadable;
		
		m_ready = right.m_ready;
		
		m_minparams = right.m_minparams;
		m_maxparams = right.m_maxparams;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfDir::set(Conf &conf, unsigned short count, std::string const params[])
{	
	if (!m_ready)
	{
		conf.errorf("Directive %s not ready", m_name.c_str());
		return false;
	}
	
	if (count < m_minparams)
	{
		conf.errorf("Cannot process directive %s: not enough parameters",
				m_name.c_str());
		return false;
	}
	
	if (count > m_maxparams)
	{
		conf.errorf("Cannot process directive %s: too many parameters",
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/
/*************************************************************************/

bool ConfBool::assign(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!params[0].empty())
	{
		switch (params[0][0])
		{
			case 'N':
			case 'n':
				if (!check(conf, false))
					return false;
				
				m_value = false;			
#if DEBUG
				conf.debugf(Conf::dbgProcess, "Set %s to false",
						m_name.c_str());
#endif
				return true;
			case 'Y':
			case 'y':
				if (!check(conf, true))
					return false;
				
				m_value = true;
#if DEBUG
				conf.debugf(Conf::dbgProcess, "Set %s to true",
						m_name.c_str());
#endif
				return true;
			default:
				break;
		}
	}
	
	conf.errorf("Invalid value %s for directive %s", (params[0].empty() ? 
			"\"\"" : params[0].c_str()), m_name.c_str());
	return false;
}

/*************************************************************************/

ConfDuration &ConfDuration::operator=(ConfDuration const &right)
{
	if (&right != this)
	{
		ConfValue<time_t>::operator=(right);
		
		m_mindur = right.m_mindur;
		m_maxdur = right.m_maxdur;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfDuration::assign(Conf &conf, unsigned short count,
		std::string const params[])
{
	time_t secs;
	
	if (!duration_scan(params[0].c_str(), &secs))
	{
		conf.errorf("Can't set directive %s: %s is not a valid duration",
				m_name.c_str(), params[0].c_str());
		return false;
	}
	
	if (!check(conf, secs))
		return false;
	
	m_value = secs;

#if DEBUG
	char buf[128];
	conf.debugf(Conf::dbgProcess, "Set %s to %s", m_name.c_str(),
			duration_expr(buf, sizeof(buf), m_value));
#endif

	return true;
}

/*************************************************************************/

bool ConfDuration::check(Conf &conf, time_t value) const
{
	if ((value < m_mindur) || (value > m_maxdur))
	{
		char buf1[128], buf2[128];
		
		conf.errorf("Can't set directive %s: value must be between "
				"%s and %s", m_name.c_str(),
				duration_expr(buf1, 128, m_mindur),
				duration_expr(buf2, 128, m_maxdur));
		return false;
	}
	
	return true;
}

/*************************************************************************/

char const *ConfDuration::label(size_t index) const
{
	if (m_label.empty())
	{
		char buf[128];
		m_label = duration_expr(buf, sizeof(buf), m_value);
	}
	
	return m_label.c_str();
}

/*************************************************************************/

bool ConfHostname::assign(Conf &conf, unsigned short count,
		std::string const params[])
{
	int err;
	struct in_addr value;
	
	if (!resolve_host(params[0].c_str(), &value, &err))
	{
		char buf[256];
		
		if (!resolve_error(buf, sizeof(buf), err))
			strncpyzt(buf, "Invalid hostname or IP address", sizeof(buf));
		
		conf.errorf("Invalid value %s for directive %s: %s", params[0].c_str(),
				m_name.c_str(), buf);
		return false;
	}
	
	m_value = value;

#if DEBUG
	conf.debugf(Conf::dbgProcess, "Set %s to %s (%s)", m_name.c_str(),
			params[0].c_str(), inet_ntoa(m_value));
#endif
	return true;
}

/*************************************************************************/

char const *ConfHostname::label(size_t index) const
{
	if (m_label.empty())
		m_label = inet_ntoa(m_value);
	
	return m_label.c_str();
}

/*************************************************************************/

ConfStr &ConfStr::operator=(ConfStr const &right)
{
	if (&right != this)
	{
		ConfValue<std::string>::operator=(right);
		
		m_minlen = right.m_minlen;
		m_maxlen = right.m_maxlen;
		m_validchars = right.m_validchars;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfStr::assign(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!check(conf, params[0]))
		return false;
	
	m_value = params[0];

#if DEBUG
	conf.debugf(Conf::dbgProcess, "Set %s to %s", m_name.c_str(),
			m_value.c_str());
#endif
	return true;
}

/*************************************************************************/

bool ConfStr::check(Conf &conf, std::string const &value) const
{
	if ((value.size() < m_minlen) || (value.size() > m_maxlen))
	{
		conf.errorf("Can't set directive %s: string length must be between "
				"%lu and %lu characters", m_name.c_str(),
				(unsigned long) m_minlen, (unsigned long) m_maxlen);
		return false;
	}
	
	if (!m_validchars.empty() && 
			strspn(value.c_str(), m_validchars.c_str()) != value.size())
	{
		conf.errorf("Can't set directive %s: invalid characters in string",
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/
/*************************************************************************/

ConfBlock::ConfBlock(ConfBlock const &cd) : ConfDir(cd),
		m_conf(cd.m_conf ? new Conf(*cd.m_conf) : 0)
{
	
}

/*************************************************************************/

ConfBlock::~ConfBlock()
{
	if (m_conf)
		delete m_conf;
}

/*************************************************************************/

ConfBlock &ConfBlock::operator=(ConfBlock const &right)
{
	if (&right != this)
	{
		ConfDir::operator=(right);
		
		if (m_conf)
			delete m_conf;
		m_conf = (right.m_conf ? new Conf(*right.m_conf) : 0);
	}
	
	return *this;
}

/*************************************************************************/

bool ConfBlock::begin(Conf &conf)
{
	if (m_ready)
		return true;
	
	if (m_conf)
	{
		delete m_conf;
		m_conf = 0;
	}
	
	Conf *new_conf = new Conf(conf.logger());
	
	if (!new_conf)
	{
		conf.errorf("Can't create new Conf object for directive %s: out of memory",
				m_name.c_str());
		return false;
	}
	
#if DEBUG
	conf.debugf(Conf::dbgProcess, "Entering begin() for %s", m_name.c_str());
#endif
	
	if (!add(conf, *new_conf) || 
			!new_conf->begin(conf.global(), conf.reloading()) ||
			!ConfDir::begin(conf))
	{
		delete new_conf;
		
		conf.errorf("Can't create directive %s: initialization failed",
				m_name.c_str());
		return false;
	}

#if DEBUG
	conf.debugf(Conf::dbgProcess, "Leaving begin() for %s", m_name.c_str());
#endif

	m_conf = new_conf;
	
	return ConfDir::begin(conf);
}

/*************************************************************************/

bool ConfBlock::end(Conf &conf)
{
	if (!m_ready)
		return true;
	
	if (m_conf)
	{
#if DEBUG
		conf.debugf(Conf::dbgProcess, "Entering end() for %s", m_name.c_str());
#endif

		if (!m_conf->end())
		{
			conf.errorf("Directive %s is not fully configured",
					m_name.c_str());
			return false;
		}
		
		/* This should take place AFTER m_conf->end() has been called,
		 * otherwise some useful methods like
		 * ConfValue<T>::has_value_changed() would have incorrect values. */
		if (conf.reloading() && !apply_changes(conf))
			return false;
			
		/*if (m_conf)
			delete m_conf;
		m_conf = 0;*/

#if DEBUG
		conf.debugf(Conf::dbgProcess, "Leaving end() for %s", m_name.c_str());
#endif
	}
	
	return ConfDir::end(conf);
}

/*************************************************************************/

bool ConfBlock::set(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;

#if DEBUG
	conf.debugf(Conf::dbgProcess, "Processing %s block", m_name.c_str());
#endif

	if (!m_conf->parse_string(params[0], m_name))
	{
		conf.errorf("Unable to process directive %s", m_name.c_str());
		return false;
	}
	
#if DEBUG
	conf.debugf(Conf::dbgProcess, "Processed %s block", m_name.c_str());
#endif
	
	return true;
}

/*************************************************************************/

ConfBoolBlock &ConfBoolBlock::operator=(ConfBoolBlock const &right)
{
	if (&right != this)
	{
		ConfBlock::operator=(right);
		
		m_set = right.m_set;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfBoolBlock::set(Conf &conf, unsigned short count,
				std::string const params[])
{	
	if (!ConfBlock::begin(conf) || !ConfBlock::set(conf, count, params))
		return false;
	
	m_set = true;
	return true;
}

/*************************************************************************/
/*************************************************************************/

ConfHostPort &ConfHostPort::operator=(ConfHostPort const &right)
{
	if (&right != this)
	{
		ConfBlock::operator=(right);
		
		hostname = right.hostname;
		port = right.port;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfHostPort::add(Conf &conf, Conf &new_conf)
{
	if (!new_conf.add(hostname) || !new_conf.add(port))
		return false;
	
	return true;
}

/*************************************************************************/

ConfRatio &ConfRatio::operator=(ConfRatio const &right)
{
	if (&right != this)
	{
		ConfValue<misc::Ratio>::operator=(right);
		
		m_minx = right.m_minx;
		m_maxx = right.m_maxx;
		m_miny = right.m_miny;
		m_maxy = right.m_maxy;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfRatio::assign(Conf &conf, unsigned short count,
		std::string const params[])
{
	int32_t x, y;
	
	char const *str = params[0].c_str();
	char *end;
	
	x = strtol(str, &end, 10);
	
	if (*end != ':' || *(++end) == 0)
	{
		conf.errorf("Can't set directive %s: %s is not a valid ratio",
				m_name.c_str(), params[0].c_str());
		return false;
	}
	
	str = end;
	y = strtol(str, &end, 10);
	
	if (*end != 0)
	{
		conf.errorf("Can't set directive %s: %s is not a valid ratio",
				m_name.c_str(), params[0].c_str());
		return false;
	}
	
	misc::Ratio ratio(x, y);
	
	if (!check(conf, ratio))
		return false;
	
	m_value = ratio;

#if DEBUG
	conf.debugf(Conf::dbgProcess, "Set %s to %ld:%ld", m_name.c_str(),
			(long) m_value.x(), (long) m_value.y());
#endif

	return true;
}

/*************************************************************************/

bool ConfRatio::check(Conf &conf, misc::Ratio value) const
{
	if (value.x() < m_minx || value.x() > m_maxx)
	{
		conf.errorf("Can't set directive %s: first part of the ratio "
				"must be between %ld and %ld", m_name.c_str(),
				(long) m_minx, (long) m_maxx);
		return false;
	}
	
	if (value.y() < m_miny || value.y() > m_maxy)
	{
		conf.errorf("Can't set directive %s: second part of the ratio "
				"must be between %ld and %ld", m_name.c_str(),
				(long) m_miny, (long) m_maxy);
		return false;
	}
	
	return true;
}

/*************************************************************************/

char const *ConfRatio::label(size_t index) const
{
	if (m_label.empty())
	{
		char buf[128];
		snprintf(buf, sizeof(buf), "%ld:%ld",
				(long) m_value.x(), (long) m_value.y());
		
		m_label = buf;
	}
	
	return m_label.c_str();
}

/*************************************************************************/

ConfRules &ConfRules::operator=(ConfRules const &right)
{
	if (&right != this)
	{
		ConfBlock::operator=(right);
		
		m_rules = right.m_rules;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfRules::begin(Conf &conf)
{
	if (conf.reloading())
		m_rules.clear();
	
	return ConfBlock::begin(conf);
}

/*************************************************************************/

ConfRules::Rule &ConfRules::Rule::operator=(ConfRules::Rule const &right)
{
	if (&right != this)
	{
		m_type = right.m_type;
		m_value = right.m_value;
	}
	
	return *this;
}

/*************************************************************************/

ConfRules::ConfRule &ConfRules::ConfRule::operator=(ConfRules::ConfRule const &right)
{
	if (&right != this)
	{
		ConfDir::operator=(right);
		
		m_container = right.m_container;
		m_type = right.m_type;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfRules::ConfRule::set(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
	if (!check(params[0]))
		return false;
	
	m_container->push_back(ConfRules::Rule(m_type, params[0]));
	return true;
}

/*************************************************************************/
/*************************************************************************/

bool ConfError::set(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
	conf.error(m_msg);
	return false;
}

/*************************************************************************/

ConfWarning &ConfWarning::operator=(ConfWarning const &right)
{
	if (&right != this)
	{
		ConfDir::operator=(right);
		
		m_msg = right.m_msg;
	}
	
	return *this;
}

/*************************************************************************/

bool ConfWarning::set(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
	conf.warning(m_msg);
	return true;
}

/*************************************************************************/
/*************************************************************************/

bool ConfElse::set(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
	if (can_process()) {
#if DEBUG
		conf.debugf(Conf::dbgProcess, "Processing %s block",
				m_name.c_str());
#endif
		Conf else_conf(conf);
		if (!else_conf.parse_string(params[0], m_name))
		{
			conf.errorf("Can't process %s block", m_name.c_str());
			return false;
		}
	}
#if DEBUG
	else {
		conf.debugf(Conf::dbgProcess, "Ignoring %s block", m_name.c_str());
	}
#endif

	return true;
}

/*************************************************************************/

bool ConfIf::set(Conf &conf, unsigned short count, std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
	if (can_process(conf, params[0])) {
#if DEBUG
		conf.debugf(Conf::dbgProcess, "Processing %s block",
				m_name.c_str());
#endif
		conf.global().else_directive().can_process(false);
		Conf if_conf(conf);
		if (!if_conf.parse_string(params[1], m_name))
		{
			conf.errorf("Can't process %s block", m_name.c_str());
			return false;
		}
	} else {
#if DEBUG
		conf.debugf(Conf::dbgProcess, "Ignoring %s block", m_name.c_str());
#endif
		conf.global().else_directive().can_process(true);
	}


	return true;
}

/*************************************************************************/

bool ConfIfnset::can_process(Conf &conf, std::string const &expr) const
{
	return !(conf.global().var_get(expr, 0));
}

/*************************************************************************/

bool ConfIfset::can_process(Conf &conf, std::string const &expr) const
{
	return (conf.global().var_get(expr, 0));
}

/*************************************************************************/

bool ConfInclude::set(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
#if DEBUG
	conf.debugf(Conf::dbgProcessLow, "Including file %s",
			params[0].c_str());
#endif	

	Conf in_conf(conf);
	if (!in_conf.parse_file(params[0]))
	{
		conf.errorf("Can't parse file %s", params[0].c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool ConfPath::set(Conf &conf, unsigned short count, std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
	if (!conf.global().incl_paths_add(params[0]))
	{
		
		if (errno)
			conf.errorf("Cannot add %s to include path: %s", params[0].c_str(),
					strerror(errno));
		else
			conf.errorf("Cannot add %s to include path", params[0].c_str());
		return false;
	}
	
#if DEBUG
	conf.debugf(Conf::dbgProcess, "Added %s to include paths",
			params[0].c_str());
#endif

	return true;
}

/*************************************************************************/

bool ConfVarSet::set(Conf &conf, unsigned short count, std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
	if (!conf.global().var_set(params[0], ((count == 2) ? params[1] : "")))
	{
		conf.errorf("Cannot set variable `%s' to `%s'", params[0].c_str(),
				params[1].c_str());
		return false;
	}
	
#if DEBUG
	conf.debugf(Conf::dbgProcess, "Set variable %s to `%s'",
			params[0].c_str(), params[1].c_str());
#endif

	return true;
}

/*************************************************************************/

bool ConfUnset::set(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfDir::set(conf, count, params))
		return false;
	
	
	if (conf.global().var_unset(params[0]))
	
#if !DEBUG
		;
#else
		conf.debugf(Conf::dbgProcess, "Unset variable %s",
				params[0].c_str());
	else
		conf.debugf(Conf::dbgProcess,
				"Ignoring unset for non-existing variable %s",
				params[0].c_str());
#endif

	return true;
}

/*************************************************************************/

} /* namespace conf */

/*************************************************************************/
