/* Hash list.
 *
 * Containers collection (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file containers/hashlist.h
  * \brief Hash list container header file. */

#ifndef CONTAINERS_HASHLIST_H
#define CONTAINERS_HASHLIST_H

/*************************************************************************/

#include <iterator>
#include <list>

/*************************************************************************/

/// Namespace for the containers collection.

namespace containers
{

/*************************************************************************/

/// Hash and comparaison class.

/** This class is used in conjunction with the HashList class to
  * determine the hash values of and compare or match a value against
  * arbitrary keys.
  *
  * You can't use this class directly; instead, create a derived
  * class and override the compare(), hash() and match() functions.
  *
  * \param T The type of value in the hash list.
  * \param KeyT The type of the key. */

template <class T, class KeyT>
class Hash
{
	public:
		/// Destructor
		virtual ~Hash() { }

		/// Compares a value against a key.
		/** Compares a value against the given key.
		  *
		  * \param[in] key The key which must be compared to the
		  * value.
		  * \param[in] value The value.
		  * \return -1, 0 or 1 if the value is respectively less than,
		  * equal to, or greater than the key. */
		virtual int compare(KeyT const &key, T const &value) const = 0;
		
		/// Hash the given key.
		/** Hash the given key.
		  *
		  * \param[in] key The key to hash.
		  * \param[in] max The number of potential different hash
		  * this function can return.
		  * \return The hash of the given key, which must
		  * be in the 0..max-1 range. */
		virtual uint32_t hash(KeyT const &key, uint32_t max) const = 0;
		
		/// Determines whether a key matches a value.
		/** Determines whether a key matches the given value.
		  *
		  * \param[in] key The key to match the value against.
		  * \param[in] value The value.
		  * \return true if the the value and key match, false
		  * otherwise. */
		virtual bool match(KeyT const &key, T const &value) const = 0;
};

/*************************************************************************/

/// Hash list.

/** This template class implements a hash list which stores values of
  * arbitrary types using keys to speedup their retrieval. It uses a
  * specialized Hash class that is responsible to compute the hash of
  * the keys and match and compare keys against the values.
  *
  * \param T The type of values stored in the list.
  * \param KeyT The type of the keys used to retrieve a value.
  * \param SIZE The size of the hash list (the number of possible values,
  * from 0 to SIZE-1, the associated Hash class will return). */
  
template <class T, class KeyT, uint32_t SIZE>
class HashList
{
	private:
		typedef std::list<T> list_type;
	public:
	
		/* Typedefs */
		
		/// Value type.
		typedef T value_type;
		/// Reference type
		typedef T & reference;
		/// Const reference type
		typedef T const & const_reference;
		/// Pointer type
		typedef T * pointer;
		/// Const pointer type
		typedef T const * const_pointer;
		/// Size type.
		typedef typename list_type::size_type size_type;
		/// Difference type.
		typedef typename list_type::difference_type difference_type;
		
		/// Key type.
		typedef KeyT key_type;
		/// Type of the Hash class.
		typedef Hash<T, KeyT> hash_type;
	
		/* Classes */
		
		class ConstIterator;
		
		/// HashList iterator class
		class Iterator;
		friend class Iterator;
		class Iterator : public 
				std::iterator<std::bidirectional_iterator_tag, T,
					typename HashList<T, KeyT, SIZE>::difference_type,
					T *, T &>
		{
			private:
				/// Container type
				typedef HashList<T, KeyT, SIZE> container_type;
				/// Iterator type
				typedef Iterator iterator_type;
				/// Internal iterator type
				typedef typename container_type::list_type::iterator literator_type;
			public:
				friend class ConstIterator;
				friend class HashList<T, KeyT, SIZE>;
			
				/* Typedefs */
				/// Iterator category.
				typedef std::bidirectional_iterator_tag iterator_category;
				/// Value type
				typedef T value_type;
				/// Difference type
				typedef typename container_type::difference_type difference_type;
				/// Reference type
				typedef T & reference;
				/// Pointer type
				typedef T * pointer;
			
				/* Functions */
			
				/// Possible values for the constructor state parameter.
				enum
				{
					/// An iterator suitable for the begin() function.
					isBegin = 0,
					/// An iterator suitable for the end() function.
					isEnd = 1
				};
			
				/// Default constructor.
				Iterator() : m_container(0), m_index(0), m_iter() { }
				/// Constructor setting the container and state (isBegin or isEnd).
				Iterator(container_type &container, int state);
				/// Constructor setting the container, index and internal iterator.
				Iterator(container_type &container, uint32_t index,
						literator_type const &iter) : m_container(&container),
						m_index(index), m_iter(iter) { }
				/// Copy constructor.
				Iterator(iterator_type const &it) : m_container(it.m_container),
						m_index(it.m_index), m_iter(it.m_iter) { }
				/// Assignment operator.
				Iterator &operator=(iterator_type const &right);
				
				/// * unary operator.
				reference operator*() const { return *m_iter; }
				/// -> operator.
				pointer operator->() const { return &(*m_iter); }
				
				/// == operator
				bool operator==(iterator_type const &right) const;
				/// != operator
				bool operator!=(iterator_type const &right) const
						{ return !(*this == right); }
				
				/// ++ operator (prefix)
				iterator_type &operator++();
				/// ++ operator (suffix)
				iterator_type operator++(int);
				
				/// -- operator (prefix)
				iterator_type &operator--();
				/// -- operator (suffix)
				iterator_type operator--(int);
			private:
				container_type *m_container;
				uint32_t m_index;
				literator_type m_iter;
		};
		
		/// HashList const iterator class
		class ConstIterator;
		friend class ConstIterator;
		class ConstIterator : public 
				std::iterator<std::bidirectional_iterator_tag, T,
					typename HashList<T, KeyT, SIZE>::difference_type,
					T const *, T const &>
		{
			private:
				/// Container type
				typedef HashList<T, KeyT, SIZE> const container_type;
				/// ConstIterator type
				typedef ConstIterator iterator_type;
				/// Internal iterator type
				typedef typename container_type::list_type::const_iterator
						literator_type;
			public:
				/* Typedefs */
				/// ConstIterator category.
				typedef std::bidirectional_iterator_tag iterator_category;
				/// Value type
				typedef T value_type;
				/// Difference type
				typedef typename container_type::difference_type difference_type;
				/// Reference type
				typedef T const & reference;
				/// Pointer type
				typedef T const * pointer;
			
				/* Functions */
			
				/// Possible values for the constructor state parameter.
				enum
				{
					/// An iterator suitable for the begin() function.
					isBegin = 0,
					/// An iterator suitable for the end() function.
					isEnd = 1
				};
			
				/// Default constructor.
				ConstIterator() : m_container(0), m_index(0), m_iter() { }
				/// Constructor setting the container and state (isBegin or isEnd).
				ConstIterator(container_type &container, int state);
				/// Constructor setting the container, index and internal iterator.
				ConstIterator(container_type &container, uint32_t index,
						literator_type const &iter) : m_container(&container),
						m_index(index), m_iter(iter) { }
				/// Copy constructor.
				ConstIterator(iterator_type const &it) : m_container(it.m_container),
						m_index(it.m_index), m_iter(it.m_iter) { }
				/// Copy constructor for normal iterator.
				ConstIterator(typename container_type::iterator const &it) : 
						m_container(it.m_container), m_index(it.m_index),
						m_iter(it.m_iter) { }
				/// Assignment operator.
				ConstIterator &operator=(iterator_type const &right);
				
				/// * unary operator.
				reference operator*() const { return *m_iter; }
				/// -> operator.
				pointer operator->() const { return &(*m_iter); }
				
				/// == operator
				bool operator==(iterator_type const &right) const;
				/// != operator
				bool operator!=(iterator_type const &right) const
						{ return !(*this == right); }
				
				/// ++ operator (prefix)
				iterator_type &operator++();
				/// ++ operator (suffix)
				iterator_type operator++(int);
				
				/// -- operator (prefix)
				iterator_type &operator--();
				/// -- operator (suffix)
				iterator_type operator--(int);
			private:
				container_type *m_container;
				uint32_t m_index;
				literator_type m_iter;
		};
		
		/* Typedefs */
		
		/// Normal iterator type.
		typedef Iterator iterator;
		/// Const iterator type.
		typedef ConstIterator const_iterator;
		/// Reverse iterator
		typedef std::reverse_iterator<iterator> reverse_iterator;
		/// Const reverse iterator
		typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
		
		/* Constants */
		
		/// Hash size.
		static int const hash_size = SIZE;
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] hash The class Hash object associated with
		  * the list and used to compute the hash and do
		  * value and key match and comparison on the values in
		  * the list. No copy of Hash object is done. It
		  * must remain valid until the HashList object is destroyed. */
		explicit HashList(hash_type const &hash) : m_hash(&hash), m_lists() { }
		/// Copy constructor
		explicit HashList(HashList const &hl);
		/// Assignment operator
		HashList &operator=(HashList const &right);
		
		/// Adds a value to the hash list.
		/** Adds a value to the hash list that can be retrieved
		  * with the given key.
		  *
		  * \param[in] key The key associated with the value.
		  * \param[in] value The value to add to the list.
		  * \return An iterator to the newly added value, or
		  * end() if a value with the same key is already in
		  * the list. */
		iterator add(KeyT const &key, T const &value);

		/// Adds a value to the hash list in sorted order.
		/** Adds a value to the hash list in sorted order. The 
		  * Hash::compare() function is used to sort the value
		  * using its key. This allows to use functions such
		  * as find_sorted() that work faster on sorted lists,
		  * but all elements of the list must have been added
		  * through the add_sorted() function.
		  *
		  * \param[in] key The key associated with the value.
		  * \param[in] value The value to add to the list.
		  * \return An iterator to the newly added value, or
		  * end() if a value with the same key is already in
		  * the list. */
		iterator add_sorted(KeyT const &key, T const &value);
		
		/// Clears the hash list.
		void clear();
		
		/// Erases the value pointed to by the given iterator.
		/** Erases the value pointed to by the given iterator.
		  *
		  * \param[in] it The iterator pointing to the value
		  * to erase. */
		void erase(iterator it);
		
		/// Erases values pointed to by the given iterator range.
		/** Erases values pointed to by the given iterator range.
		  *
		  * \param[in] first The first iterator of the range.
		  * \param[in] last The last iterator of the range. */
		void erase(iterator first, iterator last);
		
		/// Finds a value matching the given key.
		/** Finds a value matching the given key.
		  *
		  * \param[in] key The key of the value to find.
		  * \return An iterator to the requested value, or
		  * end() if no matching value was found. */
		iterator find(KeyT const &key);
		/// Const version of find().
		const_iterator find(KeyT const &key) const;
		
		/// Finds a value matching the given key in a sorted list.
		/** Finds a value matching the given key in a sorted list.
		  * All values in the hash list should have been added
		  * with the add_sorted() function. Finding a value in
		  * a sorted list is usually faster when no value matching
		  * the key exists in the list, and runs at the same
		  * speed otherwise.
		  *
		  * \param[in] key The key of the value to find.
		  * \return An iterator to the requested value, or
		  * end() if no matching value was found. */
		iterator find_sorted(KeyT const &key);
		/// Const version of find().
		const_iterator find_sorted(KeyT const &key) const;
		
		/// Removes a value matching the given key.
		/** Removes a value matching the given key.
		  *
		  * \param[in] key The key of the value to remove.
		  * \return true if a value has been removed, false
		  * otherwise. */
		bool remove(KeyT const &key);
		
		/// Removes a value matching the given key in a sorted list.
		/** Removes a value matching the given key in a sorted list.
		  * All values in the hash list should have been added
		  * with the add_sorted() function. Removing a value in
		  * a sorted list is usually faster when no value matching
		  * the key exists in the list, and runs at the same
		  * speed otherwise.
		  *
		  * \param[in] key The key of the value to remove.
		  * \return true if a value has been removed, false
		  * otherwise. */
		bool remove_sorted(KeyT const &key);
		
		/// Returns an iterator for the beginning of the hash list.
		iterator begin() { return iterator(*this, iterator::isBegin); }
		/// Returns an iterator to the end of the hash list.
		iterator end() { return iterator(*this, iterator::isEnd); }
		/// Returns a const iterator for the beginning of the hash list.
		const_iterator begin() const 
				{ return const_iterator(*this, const_iterator::isBegin); }
		/// Returns a const iterator to the end of the hash list.
		const_iterator end() const 
				{ return const_iterator(*this, const_iterator::isEnd); }
		
		/// Returns a reverse iterator to the beginning of the hash list.
		reverse_iterator rbegin() { return reverse_iterator(end()); }
		/// Returns a reverse iterator to the end of the hash list.
		reverse_iterator rend() { return reverse_iterator(begin()); }
		/// Returns a const reverse iterator to the beginning of the hash list.
		const_reverse_iterator rbegin() const { return const_reverse_iterator(end()); }
		/// Returns a const reverse iterator to the end of the hash list.
		const_reverse_iterator rend() const { return const_reverse_iterator(begin()); }
		
		/// Returns the actual number of elements in the hash list.
		size_type size() const;
		/// Returns whether the hash list is empty.
		bool empty() const;
		/// Returns the maximum size of the hash list.
		size_type max_size() const;
	private:
		hash_type const *m_hash;
		list_type m_lists[hash_size];
};

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
HashList<T, KeyT, SIZE>::HashList(HashList const &hl) : m_hash(hl.m_hash),
		m_lists() 
{ 
	for (int i = 0; i < hash_size; ++i)
		m_lists[i] = hl.m_lists[i];
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
HashList<T, KeyT, SIZE> &HashList<T, KeyT, SIZE>::operator=(HashList const &right)
{
	if (this != &right)
	{
		m_hash = right.m_hash;
		for (int i = 0; i < hash_size; ++i)
			m_lists[i] = right.m_lists[i];
	}
	
	return *this;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::iterator
		HashList<T, KeyT, SIZE>::add(KeyT const &key, T const &value)
{
	/* Ensure a value with the same key doesn't exist (yet). */
	if (find(key) != end())
		return end();
	
	uint32_t index = m_hash->hash(key, hash_size);
	return iterator(*this, index,
			m_lists[index].insert(m_lists[index].end(), value));
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::iterator
		HashList<T, KeyT, SIZE>::add_sorted(KeyT const &key, T const &value)
{
	/* Ensure a value with the same key doesn't exist (yet). */
	if (find_sorted(key) != end())
		return end();
	
	uint32_t index = m_hash->hash(key, hash_size);
	typename list_type::iterator it, end;
	
	/* Try to insert it at an appropriate place. */
	for (it = m_lists[index].begin(), end = m_lists[index].end();
			it != end; ++it)
		if (m_hash->compare(key, (*it)) < 0)
			return iterator(*this, index, m_lists[index].insert(it, value));
	
	/* When the list is empty OR this value should be the last,
	 * we add it at the end. */
	return iterator(*this, index,
			m_lists[index].insert(m_lists[index].end(), value));
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
void HashList<T, KeyT, SIZE>::clear()
{
	for (int i = 0; i < hash_size; ++i)
		m_lists[i].clear();
}

/*************************************************************************/


template <class T, class KeyT, uint32_t SIZE>
bool HashList<T, KeyT, SIZE>::empty() const
{	
	for (int i = 0; i < hash_size; ++i)
		if (!m_lists[i].empty())
			return false;
	
	return true;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
void HashList<T, KeyT, SIZE>::erase(iterator it)
{	
	m_lists[it.m_index].erase(it.m_iter);
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
void HashList<T, KeyT, SIZE>::erase(iterator first, iterator last)
{	
	for (iterator it = first; it != last; ++it)
		m_lists[it.m_index].erase(it.m_iter);
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::iterator 
		HashList<T, KeyT, SIZE>::find(KeyT const &key)
{
	uint32_t index = m_hash->hash(key, hash_size);
	typename list_type::iterator it, end;
	
	for (it = m_lists[index].begin(), end = m_lists[index].end(); it != end; ++it)
		if (m_hash->match(key, (*it)))
			return iterator(*this, index, it);
	
	return this->end();	
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::const_iterator 
		HashList<T, KeyT, SIZE>::find(KeyT const &key) const
{
	uint32_t index = m_hash->hash(key, hash_size);
	typename list_type::const_iterator it, end;
	
	for (it = m_lists[index].begin(), end = m_lists[index].end(); it != end; ++it)
		if (m_hash->match(key, (*it)))
			return const_iterator(*this, index, it);
	
	return this->end();	
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::iterator 
		HashList<T, KeyT, SIZE>::find_sorted(KeyT const &key)
{
	int result;
	uint32_t index = m_hash->hash(key, hash_size);
	typename list_type::iterator it, end;
	
	for (it = m_lists[index].begin(), end = m_lists[index].end(); it != end; ++it)
	{
		if ((result = m_hash->compare(key, (*it))) == 0)
			return iterator(*this, index, it);
		else if (result < 0)
			return this->end();
	}
	
	return this->end();	
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::const_iterator 
		HashList<T, KeyT, SIZE>::find_sorted(KeyT const &key) const
{
	int result;
	uint32_t index = m_hash->hash(key, hash_size);
	typename list_type::const_iterator it, end;
	
	for (it = m_lists[index].begin(), end = m_lists[index].end(); it != end; ++it)
	{
		if ((result = m_hash->compare(key, (*it))) == 0)
			return const_iterator(*this, index, it);
		else if (result < 0)
			return this->end();
	}
	
	return this->end();	
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::size_type HashList<T, KeyT, SIZE>::max_size() const
{
	size_type sz = 0, prev = 0;
	
	for (int i = 0; i < hash_size; ++i)
	{
		sz += m_lists[i].max_size();
		if (sz < prev)
			return prev;
		else
			prev = sz;
	}
	
	return sz;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
bool HashList<T, KeyT, SIZE>::remove(KeyT const &key)
{
	iterator it = find(key);
	
	if (it == end())
		return false;
	
	erase(it);
	return true;	
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
bool HashList<T, KeyT, SIZE>::remove_sorted(KeyT const &key)
{
	iterator it = find_sorted(key);
	
	if (it == end())
		return false;
	
	erase(it);
	return true;	
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::size_type HashList<T, KeyT, SIZE>::size() const
{
	size_type sz = 0;
	
	for (int i = 0; i < hash_size; ++i)
		sz += m_lists[i].size();
	
	return sz;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
HashList<T, KeyT, SIZE>::ConstIterator::ConstIterator(container_type &container,
		int state) : m_container(&container), m_index(0), m_iter()
{
	switch (state)
	{
		case isBegin:
			for (int i = 0; i < container.hash_size; ++i)
			{
				if (!container.m_lists[i].empty())
				{
					m_index = i;
					m_iter = container.m_lists[m_index].begin();
					return;
				}
			}
			/* Fall through */
		case isEnd:
			m_index = (container.hash_size - 1);
			m_iter = container.m_lists[m_index].end();
			break;
		default:
			break;
	}
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::ConstIterator &
		HashList<T, KeyT, SIZE>::ConstIterator::operator=(iterator_type const &right)
{
	if (this != &right)
	{
		m_container = right.m_container;
		m_index = right.m_index;
		m_iter = right.m_iter;
	}
	
	return *this;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
bool HashList<T, KeyT, SIZE>::ConstIterator::operator==(iterator_type const &right) const
{
	if (this == &right)
		return true;
	
	return ((m_container == right.m_container) && (m_index == right.m_index) &&
			(m_iter == right.m_iter));
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::ConstIterator &
		HashList<T, KeyT, SIZE>::ConstIterator::operator++()
{
	if (m_container)
	{
		if (m_iter != m_container->m_lists[m_index].end())
			++m_iter;
		
		/* If the current iterator is the end() of its list */
		if (m_iter == m_container->m_lists[m_index].end())
		{
			/* FIrst ensure we're not at the last index */
			if (m_index == (m_container->hash_size - 1))
				return *this; /* Don't increment it further */
			
			/* Look for the next index */
			for (int i = m_index + 1; i < m_container->hash_size; ++i)
			{
				if (!m_container->m_lists[i].empty())
				{
					m_index = i;
					m_iter = m_container->m_lists[m_index].begin();
					return *this;
				}
			}
			
			/* If we fall there, we're at the end */
			m_index = (m_container->hash_size - 1);
			m_iter = m_container->m_lists[m_index].end();
		}
	}
	
	return *this;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::ConstIterator
		HashList<T, KeyT, SIZE>::ConstIterator::operator++(int)
{
	ConstIterator before(*this);
	operator++();
	return before;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::ConstIterator &
		HashList<T, KeyT, SIZE>::ConstIterator::operator--()
{
	if (m_container)
	{
		/* If the current iterator is the first of its list */
		if (m_iter == m_container->m_lists[m_index].begin())
		{
			/* FIrst ensure we're not at the first index */
			if (m_index == 0)
				return *this; /* Don't decrement it further */
			
			/* Look for the next index */
			for (int i = m_index - 1; i > 0; --i)
			{
				if (!m_container->m_lists[i].empty())
				{
					m_index = i;
					m_iter = --(m_container->m_lists[m_index].end());
					return *this;
				}
			}
			
			/* If we fall there, we're at the beginning of the hash list -- 
			 * don't do anything. */
		} else {
			--m_iter;
		}
	}
	
	return *this;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::ConstIterator
		HashList<T, KeyT, SIZE>::ConstIterator::operator--(int)
{
	ConstIterator before(*this);
	operator--();
	return before;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
HashList<T, KeyT, SIZE>::Iterator::Iterator(container_type &container,
		int state) : m_container(&container), m_index(0), m_iter()
{
	switch (state)
	{
		case isBegin:
			for (int i = 0; i < container.hash_size; ++i)
			{
				if (!container.m_lists[i].empty())
				{
					m_index = i;
					m_iter = container.m_lists[m_index].begin();
					return;
				}
			}
			/* Fall through */
		case isEnd:
			m_index = (container.hash_size - 1);
			m_iter = container.m_lists[m_index].end();
			break;
		default:
			break;
	}
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::Iterator &
		HashList<T, KeyT, SIZE>::Iterator::operator=(iterator_type const &right)
{
	if (this != &right)
	{
		m_container = right.m_container;
		m_index = right.m_index;
		m_iter = right.m_iter;
	}
	
	return *this;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
bool HashList<T, KeyT, SIZE>::Iterator::operator==(iterator_type const &right) const
{
	if (this == &right)
		return true;
	
	return ((m_container == right.m_container) && (m_index == right.m_index) &&
			(m_iter == right.m_iter));
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::Iterator &
		HashList<T, KeyT, SIZE>::Iterator::operator++()
{
	if (m_container)
	{
		if (m_iter != m_container->m_lists[m_index].end())
			++m_iter;
		
		/* If the current iterator is the end() of its list */
		if (m_iter == m_container->m_lists[m_index].end())
		{
			/* FIrst ensure we're not at the last index */
			if (m_index == (m_container->hash_size - 1))
				return *this; /* Don't increment it further */
			
			/* Look for the next index */
			for (int i = m_index + 1; i < m_container->hash_size; ++i)
			{
				if (!m_container->m_lists[i].empty())
				{
					m_index = i;
					m_iter = m_container->m_lists[m_index].begin();
					return *this;
				}
			}
			
			/* If we fall there, we're at the end */
			m_index = (m_container->hash_size - 1);
			m_iter = m_container->m_lists[m_index].end();
		}
	}
	
	return *this;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::Iterator
		HashList<T, KeyT, SIZE>::Iterator::operator++(int)
{
	Iterator before(*this);
	operator++();
	return before;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::Iterator &
		HashList<T, KeyT, SIZE>::Iterator::operator--()
{
	if (m_container)
	{
		/* If the current iterator is the first of its list */
		if (m_iter == m_container->m_lists[m_index].begin())
		{
			/* FIrst ensure we're not at the first index */
			if (m_index == 0)
				return *this; /* Don't decrement it further */
			
			/* Look for the next index */
			for (int i = m_index - 1; i > 0; --i)
			{
				if (!m_container->m_lists[i].empty())
				{
					m_index = i;
					m_iter = --(m_container->m_lists[m_index].end());
					return *this;
				}
			}
			
			/* If we fall there, we're at the beginning of the hash list -- 
			 * don't do anything. */
		} else {
			--m_iter;
		}
	}
	
	return *this;
}

/*************************************************************************/

template <class T, class KeyT, uint32_t SIZE>
typename HashList<T, KeyT, SIZE>::Iterator
		HashList<T, KeyT, SIZE>::Iterator::operator--(int)
{
	Iterator before(*this);
	operator--();
	return before;
}

/*************************************************************************/

} /* namespace containers */

/*************************************************************************/

#endif /* CONTAINERS_HASHLIST_H */
