/* Database<->XML.
 *
 * PegSoft databases library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file databases/xml.h
  * \brief Database XML import/export classes. */

#ifndef DATABASES_XML_H
#define DATABASES_XML_H

/*************************************************************************/

#include <string>
#include <vector>

#include <tinyxml/tinyxml.h>

#include <databases/record.h>
#include <modules/interface.h>

/*************************************************************************/

namespace databases
{

/*************************************************************************/

class XMLIndex;

/*************************************************************************/

/// XML export interface.

/** This interface may be exposed by databases modules through the
  * Module::factory() function to provide a mean for other modules,
  * and the dbexport tool that may be provided with the program, to
  * export all databases that were ever created, so that they can
  * later be imported through the XMLImport interface.
  *
  * The export_data() function must be overriden in derived classes. */

class XMLExport : public modules::ErrorInterface
{
	public:
		/// The interface identifier
		static modules::PSIID const IID = 0x42e2ed91;
		
		/// Exports databases to XML files.
		/** Exports databases to XML files. The tags used in the
		  * XML files are currently undocumented, sorry... But it is
		  * very straightforward and examining XML files generated by
		  * some of the "standard" database modules will probably get
		  * you started without having to read the source code ;).
		  * The XMLWriter class may be used as a framework to make
		  * the task easier.
		  *
		  * \param[in] dir The output directory for XML files. If it is
		  * an empty string, files must be output to current working
		  * directory
		  * \param[in] index The index to which the files you create
		  * should be added to (you should check that a file hasn't
		  * been exported yet by looking for its presence in the index).
		  * \param[in] database The name of a database to export. If
		  * it is an empty string (the default), then all databases
		  * must be exported.
		  * \return true if successful, false otherwise (in that
		  * case, error() should return an appropriate error
		  * string describing the error.) */
		virtual bool export_data(std::string const &dir,
				XMLIndex &index, std::string const &database = "") = 0;
	private:
};

/*************************************************************************/

/// XML import interface.

/** This interface may be exposed by databases modules through the
  * Module::factory() function to provide a mean for other modules,
  * and the dbimport tool that may be provided with the program, to
  * import XML databases so that they can be used by the corresponding
  * database module normally.
  *
  * The import_data() function must be overriden in derived classes. */

class XMLImport : public modules::ErrorInterface
{
	public:
		/// The interface identifier
		static modules::PSIID const IID = 0x42e3f23f;
		
		/// Imports a database from an XML file.
		/** Imports a database from an XML file. The tags used in the
		  * XML files are currently undocumented, sorry... But it is
		  * very straightforward and examining XML files generated by
		  * some of the "standard" database modules will probably get
		  * you started without having to read the source code ;).
		  * The XMLReader class may be used as a framework to make
		  * the task easier.
		  *
		  * \param[in] dir The input directory for the XML file. If it
		  * is an empty string, the file is located in the current working
		  * directory.
		  * \param[in] database The name of the database to import.
		  * \return true if successful, false otherwise (in that
		  * case, error() should return an appropriate error
		  * string describing the error.) */
		virtual bool import_data(std::string const &dir,
				std::string const &database) = 0;
	private:
};

/*************************************************************************/

/// XML index.
/** This class implements an XML index file, that may be used to load/save 
  * respectively the files to import or the exported files. */

class XMLIndex
{
	public:
		/* Typedefs */
		
		/// Container type.
		typedef std::vector<std::string> container_type;
		/// Const iterator type.
		typedef container_type::const_iterator const_iterator;
		/// Const reverse iterator type.
		typedef container_type::const_reverse_iterator const_reverse_iterator;
		
		/* Functions */
		
		/// Constructor.
		XMLIndex() : m_files(), m_error() { }
		/// Copy constructor.
		XMLIndex(XMLIndex const &xi) : m_files(xi.m_files), m_error(xi.m_error) { }
		/// Assignment operator.
		XMLIndex &operator=(XMLIndex const &right);
		
		/// Loads index from an XML index file.
		/** Loads index from an XML index file.
		  *
		  * \param[in] filename The name of the file to load from.
		  * \return true if successful, false otherwise. The error()
		  * function can be used to get feedback on what happened. */
		bool load(std::string const &filename);
		
		/// Saves index to an XML index file.
		/** Saves index to an XML index file.
		  *
		  * \param[in] filename The name of the file to save to.
		  * \return true if successful, false otherwise. The error()
		  * function can be used to get feedback on what happened. */
		bool save(std::string const &filename);
		
		/// Adds a file to the index.
		/** Adds a file to the index. If the given file is already in the
		  * index, then it is not added, though the function is still
		  * successful.
		  *
		  * \param[in] filename The file to add.
		  * \return true if successful, false otherwise. */
		bool add(std::string const &filename);
		
		/// Removes a file from the index.
		/** Removes a file from the index.
		  *
		  * \param[in] filename The file to remove.
		  * \return true if successful, false otherwise. */
		bool remove(std::string const &filename);
		
		/// Clears the index.
		/** Clears the index.
		  *
		  * return true if successful, false otherwise. */
		void clear();
		
		/// Determines whether a file exists in the index.
		/** Determines whether a file exists in the index.
		  *
		  * \param[in] filename The file to check for.
		  * \return true if it exists, false otherwise. */
		bool exists(std::string const &filename) const;
		
		/// Returns a const_iterator to the first element.
		const_iterator begin() const { return m_files.begin(); }
		/// Returns a const_iterator to the last element.
		const_iterator end() const { return m_files.end(); }
		
		/// Returns a const_reverse_iterator to the first element.
		const_reverse_iterator rbegin() const { return m_files.rbegin(); }
		/// Returns a const_reverse_iterator to the last element.
		const_reverse_iterator rend() const { return m_files.rend(); }
		
		/// Returns whether the index is empty.
		bool empty() const { return m_files.empty(); }
		/// Returns the number of files in the index.
		container_type::size_type size() const { return m_files.size(); }
		
		/// Returns a string describing the last error that occured after load()/save().
		std::string const &error() const { return m_error; }
	private:
		container_type m_files;
		std::string m_error;
};

/*************************************************************************/

/// Class for reading XML database files.

/** This class may be used to read database XML files that have been
  * generated through the XMLExport interface, possibly using the XMLWriter
  * class. It's generally used when implementing the XMLImport interface.
  *
  * A class derived from XMLReader::Callback should be created so that
  * it'll receive the data as it is parsed from the XML file. */

class XMLReader
{
	public:
		/* Classes */
		
		/// The callback class used by XMLReader::load().
		/** Create a class derived from this abstract class so that
		  * you may process the data as you wish. */
		
		class Callback
		{
			public:
				/// Destructor.
				virtual ~Callback() { }
			
				/// Sets the database information.
				/** Sets the database information.
				  *
				  * \param[in] version The database version (same as
				  * Database::version(Database::vtExport)).
				  * \return true if successful, false otherwise. */
				virtual bool database(uint16_t version) = 0;
				
				/// Adds field information.
				/** Adds field information.
				  *
				  * \param[in] name The field name (same as Field::name()).
				  * \param[in] type The field type (same as Field::type()).
				  * \param[in] size The field size (same as Field::size()).
				  * \param[in] attrs The field attributes (same as
				  * Field::attrs()).
				  * \param[in] defval The field default value (same as
				  * Field::default_value()).
				  * \return true if successful, false otherwise. */
				virtual bool add_field(char const *name, uint16_t type,
						size_t size, uint16_t attrs, char const *defval) = 0;
				
				/// Sets the number of records in the database.
				/** Sets the number of records in the database.
				  *
				  * \param[in] count The number of records.
				  * \return true if successful, false otherwise. */
				virtual bool count(uint32_t count) = 0;
				
				/// Begins a record.
				/** Begins a record. All calls to add_value() and other
				  * record-related functions until end_record() is called
				  * is relative to this record.
				  *
				  * \return true if successful, false otherwise. */
				virtual bool begin_record() = 0;
				
				/// Adds a value to the current record.
				/** Adds a value to the current record.
				  * 
				  * \param[in] field The field name.
				  * \param[in] value The value.
				  * \return true if successful, false otherwise. */
				virtual bool add_value(char const *field,
						char const *value) = 0;
				
				/// Same as add_value(), but the value is an integer.
				virtual bool add_ivalue(char const *field, int32_t value) = 0;
				/// Same as add_value(), but the value is an unsigned integer.
				virtual bool add_uivalue(char const *field,
						uint32_t value) = 0;
				/// Same as add_value(), but the value is a Unix timestamp.
				virtual bool add_utvalue(char const *field, time_t value) = 0;
				/// Same as add_value(), but the value is a record identifier.
				virtual bool add_rivalue(char const *field,
						Record::identifier_type value) = 0;
				
				/// Ends a record.
				/** Ends a record.
				  *
				  * \return true if successful, false otherwise. */
				virtual bool end_record() = 0;
		};
	
		/* Functions */
	
		/// Constructor.
		XMLReader() : m_error() { }
		/// Copy constructor.
		XMLReader(XMLReader const &xr) : m_error() { }
		/// Assignment operator.
		XMLReader &operator=(XMLReader const &right);
		
		/// Loads the given XML file.
		/** LOads the given XML file.
		  *
		  * \param[in] filename The name of the file to load, either
		  * absolute or relative to the current working directory.
		  * \param[in] cb The callback object that will be used to
		  * pass the parsed data to the caller.
		  * \return true if successful, false otherwise. */
		bool load(std::string const &filename, Callback &cb);
		
		/// Returns the last error string.
		char const *error() const { return m_error.c_str(); }
	private:
		std::string m_error;
};

/*************************************************************************/

/// Class for writing XML file from databases.

/** This class may be used to write database XML files that can then
  * be imported through the XMLImport interface. It is generally used
  * when implementing the XMLExport interface as it saves a lot of time
  * in writing a suited XML file. */

class XMLWriter
{
	public:
		/// Constructor.
		XMLWriter() : m_doc(), m_header("header"), m_data("data"),
				m_record(0) { }
		/// Copy constructor.
		XMLWriter(XMLWriter const &xw) : m_doc(xw.m_doc),
				m_header(xw.m_header), m_data(xw.m_data),
				m_record(xw.m_record ? new TiXmlElement(*xw.m_record) : 0) { }
		/// Assignment operator.
		XMLWriter &operator=(XMLWriter const &right);
		/// Destructor.
		~XMLWriter();
		
		/// Sets the database information.
		/** Sets the database information.
		  *
		  * \param[in] version The database version (same as
		  * Database::version(Database::vtExport)).
		  * \return true if successful, false otherwise. */
		bool database(uint16_t version);
		
		/// Adds a field to the database header.
		/** Adds a field to the database header.
		  *
		  * \param[in] name The field name (same as Field::name()).
		  * \param[in] type The field type (same as Field::type()).
		  * \param[in] size The field size (same as Field::size()).
		  * \param[in] attrs The field attributes (same as Field::attrs()).
		  * \param[in] defval The field default value (same as
		  * Field::default_value()).
		  * \return true if successful, false otherwise. */
		bool add_field(char const *name, uint16_t type, size_t size,
				uint16_t attrs, char const *defval);
		
		/// Sets the number of records in the database.
		/** Sets the number of records in the database.
		  *
		  * \param[in] count The number of records.
		  * \return true if successful, false otherwise. */
		bool count(uint32_t count);
		
		/// Begins a record.
		/** Begins a record. All calls to add_value() and other
		  * record-related functions until end_record() is called
		  * is relative to this record.
		  *
		  * \return true if successful, false otherwise. */
		bool begin_record();
		
		/// Adds a value to the database data.
		/** Adds a value to the database data.
		  * 
		  * \param[in] field The field name.
		  * \param[in] value The value.
		  * \return true if successful, false otherwise. */
		bool add_value(char const *field, char const *value);
		
		/// Same as add_value(), but the value is an integer.
		bool add_ivalue(char const *field, int32_t value);
		/// Same as add_value(), but the value is an unsigned integer.
		bool add_uivalue(char const *field, uint32_t value);
		/// Same as add_value(), but the value is a Unix timestamp.
		bool add_utvalue(char const *field, time_t value);
		/// Same as add_value(), but the value is a record identifier.
		bool add_rivalue(char const *field, Record::identifier_type value);
		
		/// Ends a record.
		/** Ends a record. Must have been preceded by a call
		  * to begin_record() and one or more calls to add_value()
		  * and other related values.
		  *
		  * \return true if successful, false otherwise. */
		bool end_record();
		
		/// Frees the current record without adding it to the file.
		/** Frees the current record without adding it to the file.
		  * Use this method if you don't wish to write a record 
		  * for which begin_record() was already called.  */
		void dump_record();
		
		/// Saves the XML file.
		/** Saves the XML file.
		  *
		  * \param[in] filename The name of the file to write, either
		  * absolute or relative to the current working directory.
		  * \return true if successful, false otherwise. */
		bool save(std::string const &filename);
		
		/// Returns the error that occured if save() returns false.
		char const *error() const { return m_doc.ErrorDesc(); }
	private:
		TiXmlDocument m_doc;
		TiXmlElement m_header;
		TiXmlElement m_data;
		TiXmlElement *m_record;
};

/*************************************************************************/

} /* namespace databases */

/*************************************************************************/

#endif /* DATABASES_XML_H */
