/* Core module.
 *
 * PegSoft modules subsystem library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file modules/core.h
 * \brief Basic core module header file.
 */

#ifndef MODULES_CORE_H
#define MODULES_CORE_H

/*************************************************************************/

#include <string>

#include <modules/module.h>

/*************************************************************************/

namespace cmdline { class CmdLineParser; }
namespace conf { class ConfLogger; }
namespace xlog { class Logger; class LogLevel; }
namespace scheduler { class Scheduler; }

/*************************************************************************/

/// Namespace for all modules declarations.

namespace modules
{

/*************************************************************************/

class ModManager;

/*************************************************************************/

/// Core module class
/** This is the core module class, i.e. the module that handles all
  * the program basics.
  *
  * It encapsulates all the core features one might expect such as the
  * command-line parser, the conf logger, the modules manager, the
  * scheduler or the program logger. There's also a few functions that
  * control the program flow.
  *
  * Other modules will always be able to rely on a module derived from
  * this class as the module manager ensures this module exists and can't
  * be unloaded (yes, that wouldn't make much sense, really. :P) */

class Core : public Module
{
	public:
		/// Constructor.
		Core();
	
		/// Exit the program immediately.
		/** When something goes so wrong that there's no way to
		  * continue execution normally, call this to exit
		  * the program.
		  *
		  * \param[in] msg A message to log before exiting. If 0,
		  * exitmsg is logged instead.
		  *
		  * \sa exitmsg */
		virtual void fatal(char const *msg = 0) = 0;
		
		/// Same as fatal but works like printf().
		void fatalf(char const *msg, ...) FORMAT(printf,2,3);
		
		/// Logs a message.
		/** Logs a message through the program logger (the
		  * Logger::log() function is used).
		  *
		  * \param[in] level The log level of the message.
		  * \param[in] format A format string for the message, similar
		  * to the one used by the printf() family of functions.
		  * \param[in] ... Parameters for the format string.
		  * \sa LogLevel, Logger, Logger::log() */
		void log(xlog::LogLevel const &level, char const *format, ...)
				FORMAT(printf, 3, 4);
		
		/// Logs a message with a system error.
		/** Logs a message through the program logger (the
		  * Logger::log() function is used). If errno is non-zero,
		  * then a system error message (as returned by the 
		  * strerror() function) will be appended to it.
		  *
		  * \param[in] level The log level of the message.
		  * \param[in] format A format string for the message, similar
		  * to the one used by the printf() family of functions.
		  * \param[in] ... Parameters for the format string.
		  * \sa LogLevel, Logger, Logger::log() */
		void log_perror(xlog::LogLevel const &level,
				char const *format, ...) FORMAT(printf, 3, 4);
		
		/// Schedule flags.
		/** This enum lists all possible flags for use with the
		  * schedule() function. */
		enum
		{
			/// Reload configuration.
			sfReload =  0x00000001,
			/// Send a sync message to modules.
			sfSync =    0x00000002,
			/// Restart the program.
			sfRestart = 0x00000004,
			/// Stops the program.
			sfStop =    0x00000008,
			/// Stops the program abruptly (ignores all other flags if set).
			sfFatal =   0x00000010
		};
		
		/// Schedule program-related action.
		/** This function schedules a program-related action to be
		  * executed once, at the next main loop iteration, allowing
		  * them to be executed in a safe manner whatever the context
		  * is when the scheduling is done.
		  *
		  * This function may be called multiple times to schedule
		  * different actions. The order in which the actions are
		  * done is undefined in that case, although sfFatal is ALWAYS
		  * processed first.
		  *
		  * \param[in] action May be either sfReload, sfSync, sfRestart,
		  * sfStop or sfFatal (for the last three, don't forget to
		  * set exitmsg accordingly).
		  *
		  * \sa sfReload, sfSync, sfRestart, sfStop, sfFatal */
		virtual void schedule(unsigned int action) { m_action |= action; }
		
		/// Returns the parsed command-line parameters.
		/** Returns the command-line parameters, already parsed.
		  * Should be used by modules that want to handle their
		  * own command-line parameters (a parameter is never
		  * invalid so all parameters given at program start are
		  * always there). Modules must use only long options
		  * though, because they can't predict which short options
		  * will be already in use by the core.
		  * \return The command-line parameters.
		  * \sa cmdline::CmdLineParser
		  */
		virtual cmdline::CmdLineParser &clp() = 0;
		
		/// Returns the configuration logger object.
		/** Returns the configuration logger object, that may be used
		  * by modules when they need to create configuration parsers.
		  * \return The configuration logger. */
		virtual conf::ConfLogger &conf_logger() = 0;
		
		/// Returns current debug level.
		/** Returns current debug level, usually between 0 and
		  * 4 (but it can be more if user requests it). Debug
		  * messages must be logged only if debuglevel is non-zero.
		  * \return The debug level.
		  */
		virtual unsigned int debuglevel() const { return m_debuglevel; }
		/// Sets debug level to a new value.
		/** Sets debug level to a new value.
		  * \param[in] lev The new debug level.
		  * \return The new debug level.
		  * \sa debuglevel() */
		virtual unsigned int debuglevel(unsigned int lev)
				{ return (m_debuglevel = lev); }
		
		/// Program exit message size.
		/** Maximum size of the exit message.
		  * \sa exitmsg */
		static int const exitmsg_size = 256;
		/// Program exit message.
		/** The message that will be logged before exiting the program.
		  * Must be set prior to ask for program termination or restart
		  * through fatal(), fatalf() or schedule(). */
		char exitmsg[exitmsg_size];
		
		/// Program flags.
		/** This enum lists all possible flags for use with the
		  * flags() function family. */
		enum
		{
			/// The program won't go into background.
			pfForeground = 0x00000001,
			/// The program runs in background.
			pfBackground = 0x00000002,
			/// The program configuration file(s) ha(s/ve) been loaded.
			pfConfigured = 0x00000004,
			/// The program should not write anything to disk
			pfReadOnly   = 0x00000008,
			/// Nothing shall be written to stdout
			pfQuiet      = 0x00000010,
			/// Nothing shall be written to stdout and stderr
			pfVeryQuiet  = 0x00000020
		};
		
		/// Returns current program flags.
		/** Returns current program flags controlling
		  * runtime behaviour. 
		  * \return The current program flags, which consist
		  * of an or'ed set of pfForeground, pfBackground, pfConfigured
		  * or pfReadOnly.
		  * \sa pfForeground pfBackground pfConfigured pfReadOnly
		  */
		virtual unsigned int flags() const { return m_flags; }
		/// Adds new flags to current program flags.
		/** Adds new flags to the current program flags that control
		  * runtime behaviour.
		  * \param[in] flg Program flags to be added, which consist
		  * of an or'ed set of pfForeground, pfBackground, pfConfigured
		  * or pfReadOnly.
		  * \return The new current program flags.
		  * \sa pfForeground pfBackground pfConfigured pfReadOnly
		  */
		virtual unsigned int flags_add(unsigned int flg)
				{ return (m_flags |= flg); }
		/// Removes flags from current program flags.
		/** Removes flags from the current program flags that
		  * control runtime behaviour.
		  * \param[in] flg Program flags to be removed, which consist
		  * of an or'ed set of pfForeground, pfBackground, pfConfigured
		  * or pfReadOnly.
		  * \return New current program flags.
		  * \sa pfForeground pfBackground pfConfigured pfReadOnly
		  */
		virtual unsigned int flags_remove(unsigned int flg)
				{ return (m_flags &= ~flg); }
		
		/// Returns the program logger object.
		/** Returns the program logger object, that may be used by the
		  * program to log messages through various means, as well as
		  * register log modules.
		  * \return The program logger. */
		virtual xlog::Logger &logger() = 0;
		
		/// Returns the modules manager object.
		/** Returns the modules manager object, that is used to 
		  * load, unload and maintain the modules used by the
		  * program.
		  * \return The modules manager */
		virtual ModManager &modules() = 0; 
		
		/// Returns whether the program is started.
		/** Returns whether the program is started (i.e. all
		  * initialization finished, in the main loop).
		  */
		virtual bool started() const { return m_started; }
		/// Returns when the program started.
		/** Returns when the program started (i.e. when the core
		  * module was created). */
		virtual time_t start_time() const { return m_starttime; }
		
		/// Returns the program scheduler object.
		/** Returns the program scheduler object, that modules may use
		  * to schedule tasks to be run periodically. It's basically their
		  * main way to hook the main loop to get things to be done when
		  * they need to.
		  * \return The scheduler. */
		virtual scheduler::Scheduler &tasks() = 0;
		
		/// Returns the version string
		/** Returns the program version string, in free format.
		  * \return A version string. */
		virtual std::string const version_string() const = 0;
	protected:
		/// Scheduled action (through schedule()).
		unsigned int m_action;
		/// Debug level.
		unsigned int m_debuglevel;
		/// Program flags.
		unsigned int m_flags;
		
		/// Started?
		bool m_started;
		/// When did it start?
		time_t m_starttime;
};

/*************************************************************************/

} /* namespace modules */

/*************************************************************************/

#endif /* MODULES_CORE_H */
