/* IP sockets base class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif

#if HAVE_NETDB_H
# include <netdb.h>
#endif

#include "sockets/ipsocket.h"

using namespace std;

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

IPSocket &IPSocket::operator=(IPSocket const &right)
{
	if (this != &right)
	{
		Socket::operator=(right);
		
		if (m_local)
			delete m_local;
		m_local = (right.m_local ? new IPInfo(*right.m_local) : 0);
		
		if (m_remote)
			delete m_remote;
		m_remote = (right.m_remote ? new IPInfo(*right.m_remote) : 0);
	}
	
	return *this;
}

/*************************************************************************/

IPSocket::~IPSocket()
{
	if (m_local)
		delete m_local;
	if (m_remote)
		delete m_remote;
}

/*************************************************************************/

bool IPSocket::assign(int fd, uint32_t status)
{
	int flags;
	
	if ((flags = fcntl(fd, F_GETFL)) == -1)
		return false;
	
	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1)
		return false;
	
	return Socket::assign(fd, status);
}

/*************************************************************************/

bool IPSocket::bind()
{
	if (status() & ssBound)
		return true;
	
	if (!(status() & ssLocalInfo))
		return false;
	
	if (!(status() & ssCreated) && !create())
		return false;
	
	return (::bind(fd(), m_local->addr(), m_local->addrlen()) != -1);
}

/*************************************************************************/

bool IPSocket::create()
{
	if (!Socket::create())
		return false;
	
	int flags;
	
	if ((flags = fcntl(fd(), F_GETFL)) == -1)
		return false;
	
	return (fcntl(fd(), F_SETFL, flags | O_NONBLOCK) != -1);
}

/*************************************************************************/

char const *IPSocket::laddress() const
{
	/* That's pretty straightforward. */
	if (status() & ssLocalInfo)
		return m_local->cname();
	
	if (status() & ssCreated)
	{
#if HAVE_IPV6
		struct sockaddr_storage sa;
#else
		struct sockaddr_in sa;
#endif
		socklen_t len = sizeof(sa);
		
		if (getsockname(fd(), (struct sockaddr *)&sa, &len) == -1)
			return 0;
		
#if HAVE_IPV6
		if (family() == PF_INET6)
		{
			static char buf[INET6_ADDRSTRLEN + 1];
			return inet_ntop(family(),
					(void const *)&((struct sockaddr_in6 *)&sa)->sin6_addr, buf,
					INET6_ADDRSTRLEN);
		}
#endif
		return inet_ntoa(((struct sockaddr_in *)&sa)->sin_addr);
	}
	
	return 0;
}

/*************************************************************************/

unsigned short IPSocket::lport() const
{
	if (status() & ssLocalInfo)
		return port(m_local);
	
	if (status() & ssCreated)
	{
#if HAVE_IPV6
		struct sockaddr_storage sa;
#else
		struct sockaddr_in sa;
#endif
		socklen_t len = sizeof(sa);
		
		if (getsockname(fd(), (struct sockaddr *)&sa, &len) == -1)
			return 0;
		
#if HAVE_IPV6
		if (family() == PF_INET6)
			return ntohs(((struct sockaddr_in6 *)&sa)->sin6_port);
#endif
		return ntohs(((struct sockaddr_in *)&sa)->sin_port);
	}
	
	return 0;
}

/*************************************************************************/

unsigned short IPSocket::port(IPInfo const *ipi) const
{
#if HAVE_IPV6
	if (ipi->family() == AF_INET6)
		return ntohs(((struct sockaddr_in6 const *)ipi->addr())->sin6_port);
#endif

	return ntohs(((struct sockaddr_in const *)ipi->addr())->sin_port);
}

/*************************************************************************/

char const *IPSocket::raddress() const
{
	/* That's pretty straightforward. */
	if (status() & ssRemoteInfo)
		return m_remote->cname();
	
	if (status() & ssConnected)
	{
#if HAVE_IPV6
		struct sockaddr_storage sa;
#else
		struct sockaddr_in sa;
#endif
		socklen_t len = sizeof(sa);
		
		if (getpeername(fd(), (struct sockaddr *)&sa, &len) == -1)
			return 0;
		
#if HAVE_IPV6
		if (family() == PF_INET6)
		{
			static char buf[INET6_ADDRSTRLEN + 1];
			return inet_ntop(family(),
					(void const *)&((struct sockaddr_in6 *)&sa)->sin6_addr,
					buf, INET6_ADDRSTRLEN);
		}
#endif
		return inet_ntoa(((struct sockaddr_in *)&sa)->sin_addr);
	}
	
	return 0;
}

/*************************************************************************/

bool IPSocket::resolve(char const *address, char const *port,
		char const **error, IPInfo **ipi)
{
	if (!address || !port)
	{
		if (error)
			*error = "Missing address or port";
		return false;
	}
	
	/* Assume *ipi == 0 or pointer to valid object */
	
#if HAVE_IPV6
	struct addrinfo hints, *res;
	memset(&hints, 0, sizeof(hints));
	
	hints.ai_flags = AI_CANONNAME;
	hints.ai_family = family();
	hints.ai_socktype = type();
	hints.ai_protocol = protocol();
	
	int errnum;
	
	if ((errnum = getaddrinfo(address, port, &hints, &res)) != 0)
	{
		if (error)
			*error = gai_strerror(errnum);
		return false;
	}
	
	for (struct addrinfo *it = res; it; it = it->ai_next)
	{
		if (!*ipi || it->ai_family == (*ipi)->family())
			*ipi = new IPInfo(it->ai_family, it->ai_addrlen, it->ai_addr,
					res->ai_canonname, *ipi);
	}
	
	freeaddrinfo(res);
#else
	unsigned short nport = 0;
	char *end;
	
	nport = htons(strtol(port, &end, 10));
	if (*end)
	{
	
		struct servent *se;
	
		if (!(se = getservbyname(port, 0)))
		{
			if (error)
				*error = "Could not resolve port";
			return false;
		}
	
		nport = se->s_port;
	}
	
	struct hostent *he;
	
	if (!(he = gethostbyname(address)))
	{
		if (error)
			*error = hstrerror(h_errno);
		return false;
	}
	
	if (he->h_addrtype != AF_INET)
	{
		if (error)
			*error = "Invalid address family";
		return false;
	}
	
	for (int i = 0; he->h_addr_list[i]; ++i)
	{	
		struct sockaddr_in sin;
		sin.sin_family = AF_INET;
		sin.sin_port = nport;
		memcpy(&sin.sin_addr, he->h_addr_list[i], he->h_length);
	
		*ipi = new IPInfo(AF_INET, sizeof(sin), (struct sockaddr *)&sin,
				he->h_name, *ipi);
	}
#endif
	
	return true;
}

/*************************************************************************/

unsigned short IPSocket::rport() const
{
	if (status() & ssRemoteInfo)
		return port(m_remote);

	if (status() & ssConnected)
	{
#if HAVE_IPV6
		struct sockaddr_storage sa;
#else
		struct sockaddr_in sa;
#endif
		socklen_t len = sizeof(sa);
		
		if (getpeername(fd(), (struct sockaddr *)&sa, &len) == -1)
			return 0;
		
#if HAVE_IPV6
		if (family() == PF_INET6)
			return ntohs(((struct sockaddr_in6 *)&sa)->sin6_port);
#endif
		return ntohs(((struct sockaddr_in *)&sa)->sin_port);
	}

	return 0;
}

/*************************************************************************/

bool IPSocket::set_local(char const *address, char const *port,
		char const **error)
{
	if (status() & (ssBound | ssConnecting | ssConnected | ssListening))
		return true;
	
	IPInfo *ipi = 0; 
	
	/* Try to resolve what we have. */
	if (!resolve(address, port, error, &ipi))
		return false;
	
	/* Success? So, set everything up. */
	
	if (!(status() & (ssRemoteInfo | ssCreated)))
		family(ipi->family());
	
	if (m_local)
		delete m_local;
	m_local = ipi;
	
	status_add(ssLocalInfo);
	return true;
}

/*************************************************************************/

bool IPSocket::set_local(char const *address, unsigned short port,
		char const **error)
{
	char sport[6];
	snprintf(sport, sizeof(sport), "%u", port);

	return set_local(address, sport, error);
}

/*************************************************************************/

bool IPSocket::set_remote(char const *address, char const *port,
		char const **error)
{
	if (status() & (ssConnecting | ssConnected | ssListening))
		return true;
	
	IPInfo *ipi = 0; 
	
	/* Try to resolve what we have. */
	if (!resolve(address, port, error, &ipi))
		return false;
	
	/* Success? So, set everything up. */
	
	if (!(status() & (ssLocalInfo | ssCreated)))
		family(ipi->family());
	
	if (m_remote)
		delete m_remote;
	m_remote = ipi;
	
	status_add(ssRemoteInfo);
	return true;
}

/*************************************************************************/

bool IPSocket::set_remote(char const *address, unsigned short port,
		char const **error)
{
	char sport[6];
	snprintf(sport, sizeof(sport), "%u", port);

	return set_remote(address, sport, error);
}

/*************************************************************************/

IPSocket::IPInfo::IPInfo(int family, size_t addrlen,
		struct sockaddr const *addr, char const *name, IPInfo *next) :
		m_family(family), m_addrlen(addrlen), m_addr(), m_name(name),
		m_next(next)
{
	memcpy(&m_addr, addr, m_addrlen);
}

/*************************************************************************/

IPSocket::IPInfo::IPInfo(IPInfo const &ipi) : m_family(ipi.m_family),
		m_addrlen(ipi.m_addrlen), m_addr(), m_name(ipi.m_name), m_next(0)
{
	memcpy(&m_addr, &ipi.m_addr, m_addrlen);
	
	if (ipi.m_next)
		m_next = new IPInfo(*ipi.m_next);
}

/*************************************************************************/

IPSocket::IPInfo::~IPInfo()
{
	if (m_next)
		delete m_next;
}

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/
