/* poll()-based monitor class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <algorithm>

#include "my_poll.h"

using namespace std;

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

Poll::Poll(size_t hint) : Monitor("poll", hint), m_fds(), m_pfds()
{
	m_fds.reserve(hint);
	m_pfds.resize(hint);
}

/*************************************************************************/

bool Poll::add(Socket *s, int fd, unsigned int events)
{
	if (!Monitor::add(s, fd, events))
		return false;
	
	struct pollfd pfd;
	memset(&pfd, 0, sizeof(pfd));
	
	pfd.fd = fd;
	
	if (events & evRead)
		pfd.events |= POLLIN;
	if (events & evWrite)
		pfd.events |= POLLOUT;
	
	if (m_fds.size() == m_fds.capacity())
		m_fds.reserve(m_fds.capacity() + hint());
	
	m_fds.push_back(pfd);
	return true;
}

/*************************************************************************/

void Poll::do_events()
{
	if (m_fds.size() == 0)
		return;
	
	while (m_pfds.size() < m_fds.size())
		m_pfds.resize(m_pfds.size() * 2);
	
	memcpy(&m_pfds[0], &m_fds[0], sizeof(pollfd) * m_fds.size());
	
	int ret;
	
	while ((ret = poll(&m_pfds[0], m_fds.size(), 0)) == -1)
	{
		if (errno == EINTR)
		{
			memcpy(&m_pfds[0], &m_fds[0], sizeof(pollfd) * m_fds.size());
			continue;
		}
		
		/* A bad file descriptor means our program is flawed. */
		if (errno == EBADF)
			abort();
		
		return;
	}
	
	/* Do we need to bother ? */
	if (ret == 0)
		return;
	
	/* Now, send the event messages. */
	for (fds_list::size_type i = 0; i < m_fds.size() && ret; ++i)
	{
		if (m_pfds[i].revents)
			--ret;
		
		/* Do the write event first so the sockets can be set
		 * to connected state if read_event() is then called. */
		if (m_pfds[i].revents & (POLLOUT | POLLHUP | POLLERR))
		{
			Socket *s = get_socket(m_pfds[i].fd);
			if (s && !s->write_event())
				continue;
		}
		
		if (m_pfds[i].revents & (POLLIN | POLLHUP | POLLERR))
		{
			Socket *s = get_socket(m_pfds[i].fd);
			if (s)
				s->read_event();
		}
	}
}

/*************************************************************************/

void Poll::remove(Socket *s, int fd)
{
	fds_list::iterator it, end;
		
	for (it = m_fds.begin(), end = m_fds.end(); it != end; ++it)
	{
		if (it->fd == fd)
		{
			m_fds.erase(it);
			break;
		}
	}
	
	Monitor::remove(s, fd);
}

/*************************************************************************/

bool Poll::update(Socket *s, int fd, unsigned int events)
{
	if (!Monitor::update(s, fd, events))
		return false;
	
	int pollev = 0;
	if (events & evRead)
		pollev |= POLLIN;
	if (events & evWrite)
		pollev |= POLLOUT;
	
	fds_list::iterator it, end;
		
	for (it = m_fds.begin(), end = m_fds.end(); it != end; ++it)
	{
		if (it->fd == fd)
		{
			if (pollev == 0)
			{
				m_fds.erase(it);
				return true;
			}
			
			it->events = pollev;
			return true;
		}
	}
	
	struct pollfd pfd;
	memset(&pfd, 0, sizeof(pfd));
	
	pfd.fd = fd;
	pfd.events = pollev;
	
	if (m_fds.size() == m_fds.capacity())
		m_fds.reserve(m_fds.capacity() + hint());
	
	m_fds.push_back(pfd);
	return true;
}

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/
