/* Dynamic buffer class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/dbuffer.h
  * \brief Dynamic buffer header file. */

#ifndef SOCKETS_DBUFFER_H
#define SOCKETS_DBUFFER_H

/*************************************************************************/

#include <list>

#include <sockets/buffer.h>

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

/// Dynamic buffer class.
/** Dynamic buffer class.
  *
  * This buffer, while still having a maximum size, allocates memory
  * progressively by small chunks and frees them as needed so the
  * buffer never grows bigger than needed, which makes it ideal when
  * used with a large number of sockets.
  *
  * When freeing memory chunks, the buffer may choose to "recycle" it
  * instead, which means that they are placed in a pool for being
  * reused by this or another buffer who need a chunk of the same
  * size. */

class DBuffer : public Buffer
{
	private:
		/* Classes */
		
		class Chunk;
	public:
		/* Constants */
		
		/// The chunk unit
		static size_t const chunk_unit = 1024;
		
		/* Typedefs */
		
		typedef std::list<Chunk *> chunks_list;
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] chunkmul This number, which must be between
		  * 1 and 64, is multiplied by DBuffer::chunk_unit to set
		  * the size of the memory chunks that are allocated
		  * dynamically.
		  * \param[in] maxsize The maximum size of the buffer, in
		  * bytes. This is not allocated immediately; see the
		  * description of the class for more info. */
		explicit DBuffer(uint8_t chunkmul, size_t maxsize);
		/// Destructor.
		virtual ~DBuffer();
		
		/// Implements Buffer::read().
		virtual bool read(size_t max,
				char const **data, size_t *len) const;
		/// Implements Buffer::write().
		virtual bool write(char const *data, size_t len);
		/// Implements Buffer::erase().	
		virtual bool erase(size_t len);
		/// Implements Buffer::clear().
		virtual bool clear();
		
		/// Implements Buffer::empty().
		virtual bool empty() const { return !m_size; }
		/// Implements Buffer::size().
		virtual size_type size() const { return m_size; }
		/// Implements Buffer::max_size().
		virtual size_type max_size() const { return m_maxsize; }
	protected:
		/// Implements Buffer::read_byte().
		virtual bool read_byte(size_type pos, char *ch) const;
	private:
		/* Classes */
		
		class Chunk
		{
			public:
				/// Constructor.
				Chunk(size_t maxsize) : m_data(new char [maxsize]), m_first(0),
						m_next(m_data), m_size(0), m_maxsize(maxsize),
						m_remaining(maxsize) { }
				/// Destructor.
				~Chunk() { delete [] m_data; }
				
				/// Writes data
				bool write(char const *data, size_t len);
				/// Erases data
				bool erase(size_t len);
				/// Clears the chunk.
				void clear();
				
				/// Returns the data
				char const *data() const { return m_data; }
				/// Returns the first byte of the data
				char const *first() const { return m_first; }
				/// Returns the next byte to write the data to
				char *next() { return m_next; }
				
				/// Is the chunk empty?
				size_t empty() const { return !m_size; }
				/// Returns the size
				size_t size() const { return m_size; }
				/// Returns the maximum size.
				size_t max_size() const { return m_maxsize; }
				/// Returns the remaining size
				size_t remaining() const { return m_remaining; }
			private:
				char *m_data;
				
				char *m_first;
				char *m_next;
				
				size_t m_size;
				size_t const m_maxsize;
				size_t m_remaining;
				
				Chunk(Chunk const &);
				Chunk &operator=(Chunk const &);
		};
		
		/* Static data */
		
		/// Chunk pool
		static chunks_list chunks_pool[64];
		
		/* Data */
		
		/// The chunks for this buffer.
		chunks_list m_chunks;
		
		/// The size of a chunk
		size_t m_chunksize;
		/// The index in the pool to check.
		uint8_t m_poolidx;
		
		/// Size of the buffer
		size_t m_size;
		/// Maximum size of the buffer (capacity)
		size_t m_maxsize;
		
		/* Functions */
		
		/// Copy constructor.
		DBuffer(DBuffer const &);
		/// Assignment operator.
		DBuffer &operator=(DBuffer const &);
		
		/// Allocates a chunk, puts it in the list as last.
		Chunk *alloc();
		/// Recycles the first chunk, removes it from the list if needed.
		void recycle();
};

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_DBUFFER_H */
