/* Static buffer class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/sbuffer.h
  * \brief Static buffer header file. */

#ifndef SOCKETS_SBUFFER_H
#define SOCKETS_SBUFFER_H

/*************************************************************************/

#include <sockets/buffer.h>

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

/// Static buffer class.
/** Static buffer class.
  *
  * This buffer allocates memory to store its data once it is created,
  * which is efficient because it never needs to reallocate memory
  * afterwards, but makes it impractical for very large buffers or
  * a great number of sockets. */

class SBuffer : public Buffer
{
	public:
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] maxsize The maximum size of the buffer, in
		  * bytes. A memory chunk of that size is allocated by
		  * the constructor. */
		explicit SBuffer(size_t maxsize);
		/// Copy constructor.
		SBuffer(SBuffer const &sb);
		/// Destructor.
		virtual ~SBuffer();
		/// Assignment operator.
		SBuffer &operator=(SBuffer const &right);
		
		/// Implements Buffer::read().
		virtual bool read(size_t max,
				char const **data, size_t *len) const;
		/// Implements Buffer::write().
		virtual bool write(char const *data, size_t len);
		/// Implements Buffer::erase().	
		virtual bool erase(size_t len);
		/// Implements Buffer::clear().
		virtual bool clear();
		
		/// Implements Buffer::empty().
		virtual bool empty() const { return !m_size; }
		/// Implements Buffer::size().
		virtual size_type size() const { return m_size; }
		/// Implements Buffer::max_size().
		virtual size_type max_size() const { return m_maxsize; }
	protected:
		/// Implements Buffer::read_byte().
		virtual bool read_byte(size_type pos, char *ch) const;
	private:
		/// The buffer
		char *m_buf;
		
		/// First byte in the buffer (0 if empty)
		char const *m_first;
		/// Next byte to write (always == m_buf if empty, == m_first if full)
		char *m_next;
		/// The point past m_buf's end.
		char const *m_end;
		
		/// Size of the buffer
		size_t m_size;
		/// Maximum size of the buffer (capacity)
		size_t m_maxsize;
};

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_SBUFFER_H */
