/* TCP socket class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/tcpsocket.h
  * \brief TCP socket class header file. */

#ifndef SOCKETS_TCPSOCKET_H
#define SOCKETS_TCPSOCKET_H

/*************************************************************************/

#include <sockets/tcpclient.h>

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

/// TCP socket class
/** This class adds connection abilities to the TCPClient class. */

class TCPSocket : public TCPClient
{
	public:
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] rbuf If non-zero, points to the read buffer of
		  * the socket.
		  * \param[in] wbuf If non-zero, points to the write buffer of
		  * the socket.
		  * \sa TCPClient::rbuf(), TCPClient::wbuf() */
		explicit TCPSocket(Buffer *rbuf = 0, Buffer *wbuf = 0) :
				TCPClient(rbuf, wbuf), m_next(0) { }
		/// Copy constructor.
		TCPSocket(TCPSocket const &tcps) : TCPClient(tcps), m_next(0) { }
		/// Assignment operator.
		TCPSocket &operator=(TCPSocket const &right);
		
		/// Connects the socket.
		/** Connects the socket to the remote address and port that have
		  * been specified through the set_remote() function.
		  *
		  * This function creates the socket file descriptor if it
		  * is not created already, and calls the watch() function to
		  * monitor it. The ssCreated and ssMonitored status flags are
		  * thus guaranteed to be set if this function succeeds. 
		  *
		  * The socket may not be actually fully connected when this
		  * function succeeds; if it is, tne ssConnected and ssWritable
		  * status flags are set, the on_connect() function is called,
		  * the on_write() function is then called if the ssWritable
		  * flag is still set, else the ssConnecting flag is set. In
		  * both cases, the watch() function is called to monitor the
		  * socket so that incoming events are caught.
		  *
		  * \return true if successful, false otherwise. In the latter
		  * case, errno is set appropriately, unless you didn't call
		  * set_remote() or watch() fails. */
		virtual bool connect();
		
		/// Connects to the next peer.
		/** Connects to the next peer. The list of peers is specified
		  * through the add_remote() function.
		  *
		  * This function creates the socket file descriptor if it
		  * is not created already, and calls the watch() function to
		  * monitor it. The ssCreated and ssMonitored status flags are
		  * thus guaranteed to be set if this function succeeds. 
		  *
		  * The socket may not be actually fully connected when this
		  * function succeeds; if it is, tne ssConnected and ssWritable
		  * status flags are set, the on_connect() function is called,
		  * the on_write() function is then called if the ssWritable
		  * flag is still set, else the ssConnecting flag is set. In
		  * both cases, the watch() function is called to monitor the
		  * socket so that incoming events are caught.
		  *
		  * \param[out] last If non-zero, this is set to true if the
		  * peer is the last in the list, false otherwise. 
		  * \return true if successful, false otherwise. In the latter
		  * case, errno is set appropriately, unless you didn't call
		  * set_remote() or watch() fails. */
		virtual bool connect_next(bool *last = 0);
		
		/// Same as IPSocket::set_remote(), but resets the next connection to 0 if needed.
		virtual bool set_remote(char const *address, char const *port,
				char const **error = 0);
		/// Same as IPSocket::set_remote(), numeric port version.
		virtual bool set_remote(char const *address, unsigned short port,
				char const **error = 0) 
				{ return TCPClient::set_remote(address, port, error); }
		
		/// Adds remote address and port.
		/** Adds remote address and port to the existing list of peers
		  * that this socket can be connected to. It can't be changed
		  * once the ssListening, ssConnecting or ssConnected status
		  * flags are set on the socket.
		  *
		  * This function calls the set_remote() function if the
		  * ssRemoteInfo status flag is not set. Else it adds the
		  * new peers to the TOP of the list.
		  *
		  * \param[in] address An address to connect to, either as a
		  * hostname that will be resolved or an IP address expressed
		  * as a character string.
		  * \param[in] port A port to connect to, either as a service
		  * listed in /etc/services or a port number expressed as
		  * a character string.
		  * \param[out] error (optional) Points to a character wtring
		  * describing the error that happened if the function fails.
		  * \return true if successful, false otherwise. */
		virtual bool add_remote(char const *address, char const *port,
				char const **error = 0);
		
		/// Overrides IPSocket::raddress().
		virtual char const *raddress() const;
		
		/// Overrides IPSocket::rport().
		virtual unsigned short rport() const;
	protected:
		/// Connects to the given peer.
		virtual bool connect(IPInfo const *ipi);
		
		/// Returns the next peer to try to connect to.
		IPInfo const *next() const { return m_next; }
	private:
		IPInfo const *m_next;
};

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_TCPSOCKET_H */
