/* TCP client class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <algorithm>

#include "sockets/buffer.h"
#include "sockets/monitor.h"
#include "sockets/tcpclient.h"

using namespace std;

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

TCPClient &TCPClient::operator=(TCPClient const &right)
{
	if (this != &right)
	{
		IPSocket::operator=(right);
		m_linesep = right.m_linesep;
	}
	
	return *this;
}

/*************************************************************************/

bool TCPClient::assign(int fd, uint32_t status)
{	
	if (!IPSocket::assign(fd, status))
		return false;

	uint32_t wrstatus = (this->status() & (ssReadable | ssWritable));
	
	if ((this->status() & ssConnected) && !on_connect(0))
		return false;
	if ((this->status() & ssWritable) && !on_write())
		return false;
	if ((this->status() & ssReadable) && !on_read())
		return false;
	
	if (wrstatus != (this->status() & (ssReadable | ssWritable)))
	{
		if (!watch())
			return false;
	}
	
	return true;
}

/*************************************************************************/

bool TCPClient::close()
{
	if (status() & ssConnected)
		shutdown(fd(), SHUT_RDWR);
	
	if (rbuf())
		rbuf()->clear();
	if (wbuf())
		wbuf()->clear();
	
	return IPSocket::close();
}

/*************************************************************************/

ssize_t TCPClient::printf(uint32_t options, char const *format, ...)
{
	if (!(status() & (ssConnecting | ssConnected)))
	{
		errno = ENOTCONN;
		return -1;
	}
	
	va_list args;
	va_start(args, format);
	ssize_t ret = vprintf(options, format, args);
	va_end(args);
	
	return ret;
}

/*************************************************************************/

ssize_t TCPClient::read(char *buf, size_t len)
{
	if (!(status() & ssConnected))
		return 0;
	
	return read_real(buf, len);
}

/*************************************************************************/

bool TCPClient::read_event()
{
	status_add(ssReadable);

	if (!on_read())
		return false;
	
	if (status() & ssReadable)
		watch();
	
	return true;
}

/*************************************************************************/

ssize_t TCPClient::read_inbuf(size_t max)
{
	if (!(status() & ssConnected))
		return 0;
	
	if (!rbuf())
		return -1;
	
	if (rbuf()->remaining() == 0)
		return 0;
	
	if (max == 0)
		max = rbuf()->remaining();
	else if (max > rbuf()->remaining())
		max = rbuf()->remaining();
	
	char buf[8192];
	ssize_t ret, total_read = 0;
	
	while (max)
	{
		while ((ret = read_real(buf, min(max, sizeof(buf)))) == -1)
		{
			if (errno == EINTR)
				continue;
			
			if (errno == EAGAIN || errno == EWOULDBLOCK)
				return (total_read ? total_read : -1);
			
			return -1;
		}
		
		if (ret == 0)
			return (total_read ? total_read : 0);
		
		if (!rbuf()->write(buf, ret))
			return -1;
		
		total_read += ret;
		max -= ret;
	}
	
	return total_read;
}

/*************************************************************************/

ssize_t TCPClient::read_line(char *buf, size_t len) const
{
	if (!len || !rbuf())
		return -1;
	
	if (rbuf()->empty())
		return 0;
	
	Buffer::const_iterator first, last, it, end;
	
	for (first = rbuf()->begin(), it = last = first, end = rbuf()->end();
			it != end; ++it)
	{
		/* Make sure we don't overflow */
		if ((size_t)(it - first) >= len)
			return -1;
		
		if (*it == '\r')
		{
			/* If it's an empty line, remove 1 character from the
			 * buffer, and retry. */
			if (it == rbuf()->begin())
			{
				rbuf()->erase(1);
				return read_line(buf, len);
			}
			
			continue;
		}
		
		if (*it == '\n')
		{
			/* Woohoo! We have a line! */
			ssize_t ret = (++last - first);
			
			*(copy(first, last, buf)) = 0;
			rbuf()->erase(++it);
			
			return ret;
		}
		
		++last;
	}
	
	return 0;
}

/*************************************************************************/

ssize_t TCPClient::read_real(void *buf, size_t len)
{
	ssize_t ret;
	
	while ((ret = recv(fd(), buf, len, 0)) == -1)
	{
		if (errno == EINTR)
			continue;
		
		if ((errno == EAGAIN || errno == EWOULDBLOCK) &&
				(status() & ssReadable))
		{
			status_remove(ssReadable);
			watch();
		}
		
		return -1;
	}
	
	return ret;
}

/*************************************************************************/

ssize_t TCPClient::vprintf(uint32_t options, char const *format,
				va_list args)
{
	if (!(status() & (ssConnecting | ssConnected)))
	{
		errno = ENOTCONN;
		return -1;
	}
	
	va_list rargs;
	char buf[8192], *stat_buf = buf, *real_buf = buf;
	
	va_copy(rargs, args);
	
	size_t len = vsnprintf(buf, sizeof(buf), format, args);
	
	if (len >= sizeof(buf))
	{
		/* OK, so we actually need more memory ! */
		real_buf = new char [len + 1] ;
		len = vsnprintf(real_buf, len + 1, format, rargs);
	}
	
	va_end(rargs);
	
	ssize_t ret;
	
	if (!(options & poLine))
		ret = write(real_buf, len, ((options & poInBuf) ? true : false));
	else
		ret = write_line(real_buf, ((options & poInBuf) ? true : false));
	
	if (real_buf != stat_buf)
		delete [] real_buf;
	
	return ret;
}

/*************************************************************************/

ssize_t TCPClient::write(char const *buf, size_t len, bool inbuf)
{	
	if (!(status() & (ssConnecting | ssConnected)))
		return 0;
	
	if (!wbuf())
		return -1;
	
	if ((status() & ssConnected) && !on_writing())
		return -1;
	
	size_t total_written = 0;
	
	/* Try to write directly first, if possible. */
	if (!inbuf && (status() & ssConnected) && (wbuf()->size() == 0))
	{
		ssize_t ret;
		
		if ((ret = write_real(buf, len)) == -1)
		{
			if (errno != EAGAIN && errno != EWOULDBLOCK)
				return -1;
		}
		
		if ((size_t)ret == len)
			return ret;
		
		if ((size_t)ret < len)
		{
			buf += ret;
			len -= ret;
			
			total_written += ret;
		}
	}
	
	/* Now write to the buffer. */
	
	/* Is there enough room? */
	if (wbuf()->remaining() < len)
		return total_written;
	
	/* OK, so write... */
	if (!wbuf()->write(buf, len))
		return -1;
	
	total_written += len;
	return total_written;
}

/*************************************************************************/

bool TCPClient::write_event()
{
	status_add(ssWritable);

	if (status() & ssConnecting)
	{
		status_remove(ssConnecting);
		
		int error = 0;
		socklen_t errlen = sizeof(error);
		
		if (getsockopt(fd(), SOL_SOCKET, SO_ERROR, &error, &errlen) == -1)
			error = errno;
		
		if (error == 0)
			status_add(ssConnected);
		else
			status_remove(ssWritable);
		
		if (!on_connect(error))
			return false;
		
		if (error != 0)
			return true; /* Shouldn't happen if on_connect is
			              * well implemented. */
	}
	
	if (status() & ssWritable)
	{
		if (!on_write())
			return false;
	}
	
	if (status() & ssWritable)
		watch();
	
	return true;
}

/*************************************************************************/

ssize_t TCPClient::write_line(char const *line, bool inbuf)
{
	if (!(status() & (ssConnecting | ssConnected)))
		return 0;
	
	size_t len = strlen(line);

	if (wbuf()->remaining() < (len + m_linesep.size()))
		return 0;
	
	if (!wbuf()->write(line, len) ||
			!wbuf()->write(m_linesep.data(), m_linesep.size()))
		return -1;
	
	if (!inbuf && (status() & ssConnected) && write_outbuf() == -1)
		return -1;
	
	return (len + m_linesep.size());
}

/*************************************************************************/

ssize_t TCPClient::write_outbuf(size_t max)
{
	if (!(status() & ssConnected))
		return 0;
	
	if (!wbuf())
		return -1;
	
	if (wbuf()->size() == 0)
		return 0;
	
	if (!max || max > wbuf()->size())
		max = wbuf()->size();
	
	char const *data;
	size_t len, total_written = 0;
	
	while (max)
	{
		if (!wbuf()->read(max, &data, &len))
			return -1;
		
		ssize_t ret = write_real(data, len);
		
		if (ret == -1)
		{
			if (errno == EAGAIN || errno == EWOULDBLOCK)
				return total_written;
			else
				return -1;
		}
		
		if (!wbuf()->erase(ret))
			return -1;
		
		total_written += ret;
		max -= ret;
		
		if ((size_t)ret < len)
			return total_written;
	}
	
	return total_written;
}

/*************************************************************************/

ssize_t TCPClient::write_real(void const *buf, size_t len)
{
	ssize_t ret;
	
	while ((ret = send(fd(), buf, len, 0)) == -1)
	{
		if (errno == EINTR)
			continue;
		
		if ((errno == EAGAIN || errno == EWOULDBLOCK) &&
				(status() & ssWritable))
		{
			status_remove(ssWritable);
			watch();
		}
		
		return -1;
	}
	
	return ret;
}

/*************************************************************************/

ssize_t TCPClient::write_unbuffered(char const *buf, size_t len)
{
	if (!(status() & ssConnected))
		return 0;
	
	return write_real(buf, len);
}

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/
