/* Epona -- main source file.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#if HAVE_SIGNAL_H
# include <signal.h>
#endif

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#include <fstream>
#include <iostream>
#include <new>

#include <core/directives.h>
#include <core/epona.h>
#include <core/version.h>
#include <log/logger.h>
#include <log/levels.h>
#include <mail/mail.h>
#include <mail/transport.h>
#include <misc/misc.h>
#include <modules/manager.h>
#include <scheduler/scheduler.h>
#include <sockets/monitor.h>

#include "blacklist.h"
#include "botserv.h"
#include "chanserv.h"
#include "cmdline.h"
#include "conf.h"
#include "epona.h"
#include "nickserv.h"
#include "operserv.h"
#include "timeout.h"

#include "services.h"

using namespace std;
using namespace misc;
using namespace modules;
using namespace xlog;
using namespace core;

/******** Global variables! ********/

/* Core module */
Epona *epona;

/* Input buffer - global, so we can dump it if something goes wrong */
char inbuf[BUFSIZE];

/* Socket for talking to server */
int servsock = -1;

/* Set to 1 if we are waiting for input */
int waiting = 0;

/*************************************************************************/

LOGLEVEL (DebugProgram, xlog::LogDebug, "Debug::Program");

LOGLEVEL (InfoProgram, xlog::LogInfo, "Info::Program");

LOGLEVEL (NoticeFlow, xlog::LogNotice, "Notice::Flow");
LOGLEVEL (NoticeFlowRestart, NoticeFlow, "Notice::Flow::Restart");
LOGLEVEL (NoticeFlowStart, NoticeFlow, "Notice::Flow::Start");
LOGLEVEL (NoticeFlowStop, NoticeFlow, "Notice::Flow::Stop");

LOGLEVELH(ErrorPID, xlog::LogError, "Error::PIDfile", "Error: pid file: ");

/*************************************************************************/

/* Run expiration routines */

static void expire_all(void) 
{
	if (epona->debuglevel())
		epona->log(DebugProgram(), "Running expire routines");
	
	waiting = -21;
	aliases.expire();
	waiting = -22;
	channels.expire();
	
	waiting = -25;
	akills.expire();
	waiting = -26;
	sglines.expire();
	waiting = -28;
	sqlines.expire();
	waiting = -27;
	szlines.expire();
}

/*************************************************************************/

static void introduce_service(char const *who,
		Epona::Directives::Service const &service, char const *modes,
		bool qline)
{
	if (!who || !stricmp(who, service.nick))
		s_nick(service.nick, service.username, service.hostname,
				service.realname, modes, true);
}

/*************************************************************************/

#if defined(IRC_ULTIMATE) || defined(IRC_UNREAL)
# define SRV_STD_MODES   "+oS"
# define SRV_INVIS_MODES "+ioS"
#else
# define SRV_STD_MODES   "+o"
# define SRV_INVIS_MODES "+io"
#endif

void introduce_user(const char *user)
{
	/* Watch out for infinite loops... */
#define LTSIZE 20
	static int lasttimes[LTSIZE];
	
	if (lasttimes[0] >= time(NULL)-3)
		epona->fatal("introduce_user() loop detected");
	
	memmove(lasttimes, lasttimes+1, sizeof(lasttimes)-sizeof(int));
	lasttimes[LTSIZE-1] = time(NULL);
#undef LTSIZE

	introduce_service(user, epona->cd().nickserv.service, SRV_STD_MODES, true);
	introduce_service(user, epona->cd().chanserv.service, SRV_STD_MODES, true);
	introduce_service(user, epona->cd().memoserv.service, SRV_STD_MODES, true);
	introduce_service(user, epona->cd().botserv.service, SRV_STD_MODES, true);
	introduce_service(user, epona->cd().helpserv, SRV_STD_MODES, true);
	introduce_service(user, epona->cd().operserv.service, SRV_INVIS_MODES, true);
	introduce_service(user, epona->cd().global.service, SRV_INVIS_MODES, true);
	
	/* We make the bots go online */
	BotInfo *bi;
	Bots::const_iterator it, end;
	
	for (it = bots.begin(), end = bots.end(); it != end; ++it)
	{
		bi = (BotInfo *)(*it);
	
		if (!user || !stricmp(user, bi->nick()))
			s_nick(bi->nick(), bi->username(), bi->hostname(), bi->realname(),
						BOT_MODES, true);
	}
}

/*************************************************************************/

static void pid_remove()
{
	epona->pid_erase(epona->cd().pidfile);
}

/*************************************************************************/

Epona::Epona(int ac, char const *const *av, char const *const *envp) : Core(),
		m_clp(new ECmdLineParser(ac, av, envp)),
		m_conflogger(new EConfLogger()),
		m_cd(new hidden::Directives(*m_conflogger)),
		m_conffile(DEFAULT_CONF_FILE), m_global(0),
		m_logger(new xlog::Logger(*this, m_cd->log)), m_modules(0),
		m_tasks(new scheduler::Scheduler)
{
}

/*************************************************************************/

Epona::~Epona()
{
	delete m_tasks;
	if (m_modules)
		delete m_modules;
	delete m_logger;
	if (m_global)
		delete m_global;
	delete m_cd;
	delete m_conflogger;
	delete m_clp;
	
	epona = 0;
}

/*************************************************************************/

void Epona::fatal(char const *msg)
{
	if (!msg)
		msg = exitmsg;
	if (!*msg)
		strncpyzt(exitmsg, "Unknown reason", exitmsg_size);
	
	if (servsock >= 0) 
	{
		s_squit(m_cd->services.sname, msg);
		disconn(servsock);
	}
	
	log(LogFatal(), "%s", msg);
	exit(EXIT_FAILURE);
}

/*************************************************************************/

void Epona::load()
{	
	/* Parse command-line options. */
	if (!m_clp->parse()) 
	{
		log(LogFatal(), "%s", m_clp->error().c_str());
		exit(EXIT_FAILURE);
	}

	/* Parse configuration files. */
	if (!config(false))
		fatal("Configuration error");
	m_flags |= pfConfigured;
	
	/* Announce ourselves */
	log(NoticeFlowStart(), "%s %s starting up...", program_name,
			version_string().c_str());
	
	/* Configure signal handling */
	init_signals();
	
	/* Configure mail library */
	mail::conf_options(m_cd->mail.sender.name, m_cd->mail.sender.address,
			m_cd->mail.limit.authentified.value(),
			m_cd->mail.limit.unauthentified.value(),
			m_cd->mail.sender_limit.value());
	
	/* Configure sockets monitor */
	sockets::setup_monitors(256);
	
	/* Initialize modules manager */
	m_modules = new ModManager(*this, m_cd->module);
	m_modules->global(*m_global);
	
	/* Load modules as specified in the config file */
	load_modules();
	
	ns_init();
	cs_init();
	ms_init();
	bs_init();
	os_init();
}

/*************************************************************************/

void Epona::load_modules()
{
	typedef cmdline::CmdLineParser::ParamsList::const_iterator PIType;
	
	pair<PIType, PIType> m_range, module_range;
	
	m_range = m_clp->params().equal_range("m");
	module_range = m_clp->params().equal_range("module");
	
	/* Are there modules specified on the command-line? */
	if ((m_range.first != m_range.second) || 
			(module_range.first != module_range.second))
	{
		/* So load them. */
		PIType it, end;
		
		for (it = m_range.first, end = m_range.second; it != end; ++it)
			if (!m_modules->load(it->second))
				fatalf("Unable to load module %s: %s", it->second.c_str(),
						m_modules->error().c_str());
		
		for (it = module_range.first, end = module_range.second; it != end; ++it)
			if (!m_modules->load(it->second))
				fatalf("Unable to load module %s: %s", it->second.c_str(),
						m_modules->error().c_str());
		
		flags_add(efTakeOver);
	} else 
	{
		/* Load the mail transport module, if needed. */
		if (m_cd->mail)
		{
			if (!m_modules->load(m_cd->mail.transport))
				fatalf("Unable to load mail transport module %s: %s",
						m_cd->mail.transport.value().c_str(),
						epona->modules().error().c_str());
		}				
		
		/* Load modules as specified in the config file */
		if (!m_modules->load(m_cd->load_module))
			fatalf("Unable to load all requested modules: %s",
					m_modules->error().c_str());
		
		/* These call serve as Module::load() substitutes until they are
		 * really modules. */
		if (!nickserv_load() || !botserv_load() || !chanserv_load() ||
				!operserv_load() || !news_load() || !exceptions_load())
			fatal("Can't initialize services");
	}
}

/*************************************************************************/

void Epona::notify(Module &sender, int msg, void *arg)
{
	conf::ConfStringA::container_type::const_iterator it, end;
	
	switch (msg)
	{
		case mnLoaded:
			for (it = m_cd->lock_module->begin(),
					end = m_cd->lock_module->end(); it != end; ++it)
			{
				if ((*it) == sender.name())
					sender.references_add();
			}
			
			if (sender.name() == m_cd->mail.transport.value())
			{
				mail::Transport *transport = (mail::Transport *)
						sender.interface(mail::Transport::IID);
				
				if (!transport)
					fatalf("%s is not a mail transport module",
							sender.name().c_str());
				
				sender.references_add();
				
				mail::conf_transport(transport);
			}
			
			break;
		
		default:
			break;
	}
	
	nickserv_notify(sender, msg, arg);
	botserv_notify(sender, msg, arg);
	chanserv_notify(sender, msg, arg);
	operserv_notify(sender, msg, arg);
	news_notify(sender, msg, arg);
	exceptions_notify(sender, msg, arg);
}

/*************************************************************************/

void Epona::pid_erase(char const *filename)
{
	if (filename)
		if (remove(filename) == -1)
			log_perror(ErrorPID(), "Can't remove PID file %s", filename);
}

/*************************************************************************/

void Epona::pid_write(char const *filename)
{
	if (!filename)
		return;
	
	/* First check there is no PID file already existing. */
	
	int rpid;
	ifstream in(filename, ios::in);
	
	if (in >> rpid)
	{
		if (kill(rpid, SIGCHLD) == 0 || (errno == EPERM))
		{
			fatal("Epona is already running, aborting.");
			exit(EXIT_FAILURE);
		}
	}
	
	/* Write the PID file */
	ofstream out(filename, ios::out | ios::trunc);
	
	if (!(out << (int) getpid() << endl))
		fatalf("Can't write PID file %s: %s", filename, strerror(errno));
	
	/* Clean up the PID file. */
	atexit(pid_remove);
}

/*************************************************************************/

void Epona::restart()
{
	if (!*exitmsg)
		strncpyzt(exitmsg, "Unknown reason", exitmsg_size);

	log(NoticeFlowRestart(), "Restarting: %s", exitmsg);
	
	m_modules->stop();
	
	s_squit(m_cd->services.sname, exitmsg);
	disconn(servsock);
	
	execve(EPONA_BIN, const_cast<char *const *>(m_clp->argv()),
			const_cast<char *const *>(m_clp->envp()));
	
	log_perror(LogFatal(), "Restart failed");
	exit(EXIT_FAILURE);
}

/*************************************************************************/

void Epona::run()
{
	load();
	start();
	
	/* We have a line left over from earlier, so process it first. */
	process();
	
	volatile time_t last_blu; /* When did we last update the blacklist? */
	volatile time_t last_expire; /* When did we last expire nicks/channels? */
	volatile time_t last_check;  /* When did we last check timeouts? */

	/* Set up timers. */
	last_blu = last_expire = last_check = time(NULL);

	m_started = true;

	/*** Main loop. ***/

	while (!(m_action & sfFatal)) 
	{
		time_t t = time(NULL);

		if (m_debuglevel >= 2)
			log(DebugProgram(), "debug: Top of main loop");
	
		if (!(m_flags & efNoExpire) &&
				((m_action & sfSync) || t - last_expire >= m_cd->expire_timeout))
		{
			expire_all();
			last_expire = t;
		}
		
		if ((m_action & sfSync) || (t - last_blu >= m_cd->blacklist.update)) 
		{
			waiting = -18;
			blacklist_update();
			
			last_blu = t;
		}
		
		if (m_action & sfSync)
		{
			m_modules->broadcast(Module::msStarted, *this, Module::mnSync, 0);
			m_action &= ~sfSync;
		}
		
		if (m_action & (sfRestart | sfStop))
			break;
		
		if (m_action & sfReload)
		{
			m_flags &= ~pfConfigured;
			if (!config(true))
				fatal("Configuration reload failed");
			m_flags |= pfConfigured;
			
			m_modules->global(*m_global);
			if (!m_modules->reload())
				fatal("Configuration reload failed");
			
			m_modules->broadcast(Module::msStarted,
					*this, Module::mnReconfig, 0);
			
			m_action &= ~sfReload;
		}
		
		waiting = -1;
		m_tasks->run();
		
		waiting = -3;
		sockets::do_events();
		
		waiting = -2;
		if (t-last_check >= m_cd->timeout_check) 
		{
			check_timeouts();
			last_check = t;
		}
		
		waiting = 1;
		int i = (int)(long)sgets2(inbuf, sizeof(inbuf), servsock);
		waiting = 0;
		if (i > 0) {
			process();
		} else if (i == 0) {
			snprintf(exitmsg, exitmsg_size,
					"Read error from server: %s", strerror(errno));
			break;
		}
		waiting = -4;
	}

	/* Check for fatal exit */
	if (m_action & sfFatal)
		fatal();	

	/* Check for restart instead of exit */
	if (m_action & sfRestart)
		restart(); /* Does not return. */

	/* Stop all modules and exit */
	stop();
}

/*************************************************************************/

void Epona::start()
{
	/* Detach ourselves if requested. */
	if (!(m_flags & pfForeground)) 
	{
		log(InfoProgram(), "Now going into background.");
		
		int i;
		
		if ((i = fork()) < 0) {
			log_perror(LogFatal(), "fork()");
			exit(EXIT_FAILURE);
		} else if (i != 0) {
			exit(EXIT_SUCCESS);
		}
	
		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
		
#if HAVE_SETSID
		if (setsid() == -1) 
		{
			log_perror(LogFatal(), "setsid()");
			exit(EXIT_FAILURE);
		}
#elif HAVE_SETPGID
		if (setpgid(0, 0) == -1) 
		{
			log_perror(LogFatal(), "setpgid()");
			exit(EXIT_FAILURE);
		}
#endif
		m_flags |= pfBackground;
	}

	/* Write our PID to the PID file. */
	if (!(flags() & efTakeOver))
		pid_write(m_cd->pidfile);

    /* Start modules */
	m_modules->start();
    
	/* If the --module command-line parameter was specified, now is a good
	 * time to stop if m_modules->start() ever returns. */
	if (flags() & efTakeOver)
		exit(EXIT_SUCCESS);

	/* Initialize multi-language support */
	lang_init();

	/* These functions replace Module::start() until they are really modules */
	nickserv_start();
	botserv_start();
	chanserv_start();
	operserv_start();
	news_start();
	exceptions_start();

	log(InfoProgram(), "Databases loaded");

	/* Initialize hostname blacklist */
	if (m_cd->blacklist.enable)
		blacklist_init();

    /* Connect to the remote server */
	servsock = conn(m_cd->services.connect.hostname, m_cd->services.connect.port,
    		m_cd->services.bind.hostname, m_cd->services.bind.port);
	if (servsock < 0)
		fatalf("Can't connect to server: %s", strerror(errno));
	
	s_linkneg(m_cd->services.password, m_cd->services.sname,
			m_cd->services.description);

	sgets2(inbuf, sizeof(inbuf), servsock);
	if (strnicmp(inbuf, "ERROR", 5) == 0) 
	{
	    /* Close server socket first to stop wallops, since the other
	     * server doesn't want to listen to us anyway */
	    disconn(servsock);
		servsock = -1;
		fatalf("Remote server returned: %s", inbuf);
	}

	/* Bring in our pseudo-clients */
	introduce_user(NULL);
    
	/* Announce ourselves to the logfile. */
	if (m_debuglevel || (m_flags & pfReadOnly)) {
		log(NoticeFlowStart(), "Epona %s (compiled for %s) started (options:%s%s)",
				version_string().c_str(), program_protocol,
				(m_debuglevel) ? " debug" : "", 
				(m_flags & pfReadOnly) ? " readonly" : "");
	} else {
		log(NoticeFlowStart(), "Epona %s (compiled for %s) started",
				version_string().c_str(), program_protocol);
	}
}

/*************************************************************************/

void Epona::stop()
{
	if (!*exitmsg)
		strncpyzt(exitmsg, "Unknown reason", exitmsg_size);
	log(NoticeFlowStop(), "Stopping: %s", exitmsg);
	
	m_modules->stop();
	
	s_squit(m_cd->services.sname, exitmsg);
	disconn(servsock);
}

/*************************************************************************/

int Epona::version() const
{
	return program_version;
}

/*************************************************************************/

std::string const Epona::version_string() const
{
	char ver[32];
	
	switch (program_status)
	{
		case psDevel:
			snprintf(ver, sizeof(ver), "%d.%d(%d)",
					((program_version >> 24) & 0xFF),
					((program_version >> 16) & 0xFF), program_date);
			break;
		case psAlpha:
			snprintf(ver, sizeof(ver), "%d.%da%d",
					((program_version >> 24) & 0xFF),
					((program_version >> 16) & 0xFF),
					(program_version & 0xFFFF));
			break;
		case psBeta:
			snprintf(ver, sizeof(ver), "%d.%db%d",
					((program_version >> 24) & 0xFF),
					((program_version >> 16) & 0xFF),
					(program_version & 0xFFFF));
			break;
		case psPre:
			snprintf(ver, sizeof(ver), "%d.%drc%d",
					((program_version >> 24) & 0xFF),
					((program_version >> 16) & 0xFF),
					(program_version & 0xFFFF));
			break;
		default:
			snprintf(ver, sizeof(ver), "%d.%d.%d",
					((program_version >> 24) & 0xFF),
					((program_version >> 16) & 0xFF),
					(program_version & 0xFFFF));
			break;
	};
	
	return ver;
}

/*************************************************************************/

/* Main routine. (Disappointingly short :P) */

int main(int ac, char *av[], char *envp[])
{	
	/* Set our out-of-memory handler */
	set_new_handler(out_of_memory);
	
	/* Initialize random number generator */
	srand(time(0) * getpid());
	
	epona = new Epona(ac, av, envp);
	epona->run();
	
	exit(EXIT_SUCCESS);
}

/*************************************************************************/
