/* Horsepower.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file core/epona.h
  * \brief Epona core module header file. */

#ifndef CORE_EPONA_H
#define CORE_EPONA_H

/*************************************************************************/

#include <string>

#include <modules/core.h>

/*************************************************************************/

namespace conf { class ConfGlobalQ; }

/*************************************************************************/

/// Namespace for Epona's core declarations.

namespace core
{

/*************************************************************************/

namespace hidden
{
	class Directives;
}

/*************************************************************************/

/// Epona core module class
/** This module contains all the core functionality of Epona, namely,
  * the command-line and configuration file parsers, the modules manager,
  * the scheduler and the program logger. 
  * It is automatically created on startup and available to all modules. */ 

class Epona : public modules::Core
{
	private:
		friend class ECmdLineParser;
	public:
		/* Constants */
		/// --module is taking over Epona
		static uint32_t const efTakeOver = 0x80000000;
		/// No expiration checks
		static uint32_t const efNoExpire = 0x40000000;
		
		/* Typedefs */
		typedef hidden::Directives Directives;
	
		/* Functions */
	
		/// Core module constructor.
		Epona(int ac, char const *const *av, char const *const *envp);
		/// Core module destructor.
		~Epona();
		
		/// Implements Module::config().
		virtual conf::ConfQ *config(modules::Core &core,
				conf::ConfGlobalQ &global);
		
		/// Sets whether Epona should dump core or not.
		/** Sets whether Epona should dump core when receiving SIGBUS,
		  * SIGFPE, SIGILL, SIGSEGV and other fatal signals.
		  *
		  * \param[in] dump Set to true to dump a core, or false to
		  * use Epona's own handling. */
		void dump_core(bool dump);
		
		/// Implements Core::fatal().
		virtual void fatal(char const *msg = 0);
		
		/// Implements Module::notify().
		virtual void notify(Module &sender, int msg, void *arg);
		
		/// Removes a PID file.
		/** Removes the given PID file.
		  * \param[in] filename The name of the file to remove.
		  */
		void pid_erase(const char *filename);
		
		/// Writes a PID file.
		/** Writes the given PID file.
		  * \param[in] filename The name of the file to write.
		  */
		void pid_write(const char *filename);
		
		/// Epona main loop.
		/** This is called automatically from main(). It mustn't
		  * be called twice so ignore it. */
		void run();
		
		/// Implements Core::clp().
		virtual cmdline::CmdLineParser &clp() { return *m_clp; }
		
		/// Returns the core configuration directives.
		hidden::Directives const &cd() const { return *m_cd; }
		/// Returns the main configuration file.
		std::string const &conffile() const { return m_conffile; }
		
		/// Implements Core::conf_logger;
		virtual conf::ConfLogger &conf_logger() { return *m_conflogger; }
		
		/// Implements Core::logger().
		virtual xlog::Logger &logger() { return *m_logger; }
		
		/// Implements Core::modules().
		virtual modules::ModManager &modules() { return *m_modules; }
		
		/// Implements Core::tasks().
		virtual scheduler::Scheduler &tasks() { return *m_tasks; }
		
		/// Implements Module::version().
		virtual int version() const;
		/// Implements Core::version_string().
		virtual std::string const version_string() const;
	private:
		cmdline::CmdLineParser *m_clp;
		
		conf::ConfLogger *m_conflogger;
		hidden::Directives *m_cd;
		std::string m_conffile;
		conf::ConfGlobalQ *m_global;
		
		xlog::Logger *m_logger;
		
		modules::ModManager *m_modules;
		
		scheduler::Scheduler *m_tasks;
		
		bool config(bool reload);
		
		void init_signals();
		void load_modules();
		
		void load();
		void restart();
		void start();
		void stop();
};

/*************************************************************************/

} /* namespace core */

/*************************************************************************/

#endif /* CORE_EPONA_H */
