/* Database export module.
 *
 * Database export module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <cmdline/cmdline.h>
#include <core/version.h>
#include <databases/xml.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <modules/core.h>
#include <modules/manager.h>

#include "decl.h"
#include "module.h"

using namespace std;
using namespace modules;
using namespace dbexport;

/*************************************************************************/

extern "C" Module *dbexport_get(Core &core)
{
	/* Ensure the core version is the same as the version we were
	 * compiled for so that there are no compatibility problems later. */
	if (core.version() != core::core_version)
	{
		core.log(xlog::LogCritical(), "Core module version mismatch");
		return 0;
	}
	
	return (me = new DBExport(core));
}

/*************************************************************************/

extern "C" void dbexport_release(Core &core, Module *module)
{
	delete module;
	me = 0;
}

/*************************************************************************/

namespace dbexport
{

/*************************************************************************/

LOGLEVEL (InfoDBExport, xlog::LogInfo, "Info::DBExport");

/*************************************************************************/

DBExport *me;

/*************************************************************************/

DBExport::DBExport(Core &core) : Module("dbexport", "Database export utility"),
		m_core(&core), m_src(0), m_source(), m_dir(), m_index("index.xml"),
		m_db()
{
	
}

/*************************************************************************/

bool DBExport::load(modules::Core &core)
{
	/* Make the program run in foreground. */
	core.flags_add(Core::pfForeground);
	
	/* Parse command-line parameters. */
	typedef cmdline::CmdLineParser::ParamsList::const_iterator PIType;
	
	PIType it;
	
	it = core.clp().params().find("source");
	if (it == core.clp().params().end())
	{
		core.log(xlog::LogCritical(),
				"Missing --source command-line parameter.");
		return false;
	}
	m_source = it->second;
	
	it = core.clp().params().find("dir");
	if (it != core.clp().params().end())
		m_dir = it->second;

	it = core.clp().params().find("index");
	if (it != core.clp().params().end())
		m_index = it->second;
	
	pair<PIType, PIType> range;
	
	range = core.clp().params().equal_range("db");
	if (range.first != range.second)
	{
		PIType end;
		
		for (it = range.first, end = range.second; it != end; ++it)
			m_db.push_back(it->second);
	}
	
	if (!core.modules().load(m_source))
	{
		core.log(xlog::LogCritical(), "Can't load module %s",
				m_source.c_str());
		return false;
	}
	
	return Module::load(core);
}

/*************************************************************************/

void DBExport::notify(modules::Module &sender, int msg, void *arg)
{	
	Module::notify(sender, msg, arg);
	
	switch (msg)
	{
		case Module::mnLoaded:
			if (sender.name() == m_source)
			{
				m_src = &sender;
				m_src->references_add();
			}
			break;
		case Module::mnStopped:
			if (&sender == m_src)
				m_src->references_remove();
			m_src = 0;
		default:
			break;
	}
}

/*************************************************************************/

void DBExport::start(modules::Core &core)
{
	databases::XMLExport *xex;
	
	if (!m_src || !(m_src->state() & Module::msStarted))
		m_core->fatalf("Module %s has not been started", m_source.c_str());
	
	if (!(xex = (databases::XMLExport *) 
			m_src->factory(databases::XMLExport::IID)))
		m_core->fatalf("Module %s does not expose the XMLExport interface",
				m_source.c_str());
	
	databases::XMLIndex xi;
	
	if (m_db.empty())
	{
		if (!xex->export_data(m_dir, xi))
			m_core->fatalf("Can't export databases through module %s: %s",
					m_source.c_str(), xex->error());
	} else 
	{
		dblist_type::const_iterator it, end;
		
		for (it = m_db.begin(), end = m_db.end(); it != end; ++it)
		{
			if (!xex->export_data(m_dir, xi, *it))
				m_core->fatalf("Can't export database %s through module %s: %s",
						it->c_str(), m_source.c_str(), xex->error());
		}
	}
	
	delete xex;
	
	if (!xi.save(misc::absolute_path(m_dir, m_index)))
		m_core->fatalf("Can't save index file %s",
				misc::absolute_path(m_dir, m_index).c_str());
	
	m_core->log(InfoDBExport(), "Saved index file");
	m_core->log(InfoDBExport(), "Database export complete.");
	
	Module::start(core);
	exit(EXIT_SUCCESS);
}

/*************************************************************************/

} /* namespace dbexport */

/*************************************************************************/
