/* XMLExport interface.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <map>
#include <vector>

#include <cmdline/cmdline.h>
#include <core/epona.h>
#include <databases/field.h>
#include <databases/record.h>
#include <databases/xml.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <misc/string.h>
#include <tinyxml/tinyxml.h>

#include "datafiles.h"
#include "decl.h"
#include "module.h"
#include "xmlexport.h"

using namespace std;

using databases::Field;
using databases::Record;
using databases::XMLIndex;
using databases::XMLWriter;

using misc::absolute_path;
using misc::strncpyzt;

/*************************************************************************/

namespace
{

/*************************************************************************/

typedef std::map<std::string, Record::identifier_type, misc::insensitive_less> id_map;

id_map groups;
id_map aliases;
id_map bots;

/*************************************************************************/

class Forbid
{
	public:
		Forbid() : m_name(0), m_creator(0), m_reason(0), m_since(time(0)) { }
		
		char const *name() const { return m_name; }
		char const *name(char const *name) { return (m_name = name); }
		
		char const *reason() const { return m_reason; }
		char const *reason(char const *reason) { return (m_reason = reason); }
		
		char const *creator() const { return m_creator; }
		char const *creator(char const *creator)
				{ return (m_creator = creator); }
		
		time_t since() const { return m_since; }
		time_t since(time_t since) { return (m_since = since); }
	private:
		misc::cstring m_name;
		misc::cstring m_creator;
		misc::cstring m_reason;
		time_t m_since;
		
		Forbid(Forbid const &);
		Forbid &operator=(Forbid const &);
};

vector<Forbid *> forbids;

/*************************************************************************/

legacydb::CBModeInfo cbmi_bahamut[] = 
{
	/* Standard modes */
	
	{ 'i', CMODE_i, 0, -1 },
	{ 'k', CMODE_k, 0,  0 },
	{ 'l', CMODE_l, CBM_MINUS_NO_ARG, 1 },
	{ 'm', CMODE_m, 0, -1 },
	{ 'n', CMODE_n, 0, -1 },
	{ 'p', CMODE_p, 0, -1 },
	{ 's', CMODE_s, 0, -1 },
	{ 't', CMODE_t, 0, -1 },
	
	/* Extended modes */
	
	{ 'c', CMODE_BAHAMUT_c, 0, -1 },
	{ 'j', CMODE_BAHAMUT_j, CBM_MINUS_NO_ARG, 3 },
	{ 'r', CMODE_r, 0, -1 },
	{ 'M', CMODE_BAHAMUT_M, 0, -1 },
	{ 'O', CMODE_BAHAMUT_O, 0, -1 },
	{ 'R', CMODE_R, 0, -1 },
	{ 0 }
};

legacydb::CBModeInfo cbmi_dreamforge[] = 
{
	/* Standard modes */
	
	{ 'i', CMODE_i, 0, -1 },
	{ 'k', CMODE_k, 0,  0 },
	{ 'l', CMODE_l, CBM_MINUS_NO_ARG, 1 },
	{ 'm', CMODE_m, 0, -1 },
	{ 'n', CMODE_n, 0, -1 },
	{ 'p', CMODE_p, 0, -1 },
	{ 's', CMODE_s, 0, -1 },
	{ 't', CMODE_t, 0, -1 },
	
	/* Extended modes */
	
	{ 'r', CMODE_r, 0, -1 },
	{ 'R', CMODE_R, 0, -1 },
	{ 0 }
};

legacydb::CBModeInfo cbmi_ultimate[] = 
{
	/* Standard modes */
	
	{ 'i', CMODE_i, 0, -1 },
	{ 'k', CMODE_k, 0,  0 },
	{ 'l', CMODE_l, CBM_MINUS_NO_ARG, 1 },
	{ 'm', CMODE_m, 0, -1 },
	{ 'n', CMODE_n, 0, -1 },
	{ 'p', CMODE_p, 0, -1 },
	{ 's', CMODE_s, 0, -1 },
	{ 't', CMODE_t, 0, -1 },
	
	/* Extended modes */
	
	{ 'f', CMODE_ULTIMATE_f, 0,  2 },
	{ 'r', CMODE_r, 0, -1 },
	{ 'x', CMODE_ULTIMATE_x, 0, -1 },
	{ 'A', CMODE_ULTIMATE_A, 0, -1 },
	{ 'I', CMODE_ULTIMATE_I, 0, -1 },
	{ 'K', CMODE_ULTIMATE_K, 0, -1 },
	{ 'L', CMODE_ULTIMATE_L, 0,  4 },
	{ 'O', CMODE_ULTIMATE_O, 0, -1 },
	{ 'R', CMODE_R, 0, -1 },
	{ 'S', CMODE_ULTIMATE_S, 0, -1 },
	{ 0 }
};

legacydb::CBModeInfo cbmi_unreal[] = 
{
	/* Standard modes */
	
	{ 'i', CMODE_i, 0, -1 },
	{ 'k', CMODE_k, 0,  0 },
	{ 'l', CMODE_l, CBM_MINUS_NO_ARG, 1 },
	{ 'm', CMODE_m, 0, -1 },
	{ 'n', CMODE_n, 0, -1 },
	{ 'p', CMODE_p, 0, -1 },
	{ 's', CMODE_s, 0, -1 },
	{ 't', CMODE_t, 0, -1 },
	
	/* Extended modes */
	
	{ 'c', CMODE_UNREAL_c, 0, -1 },
	{ 'f', CMODE_UNREAL_f, 0,  2 },
	{ 'j', CMODE_UNREAL_j, CBM_MINUS_NO_ARG, 3 },
	{ 'r', CMODE_r, 0, -1 },
	{ 'u', CMODE_UNREAL_u, 0, -1 },
	{ 'z', CMODE_UNREAL_z, 0, -1 },
	{ 'A', CMODE_UNREAL_A, 0, -1 },
	{ 'C', CMODE_UNREAL_C, 0, -1 },
	{ 'G', CMODE_UNREAL_G, 0, -1 },
	{ 'K', CMODE_UNREAL_K, 0, -1 },
	{ 'L', CMODE_UNREAL_L, 0,  4 },
	{ 'M', CMODE_UNREAL_M, 0, -1 },
	{ 'N', CMODE_UNREAL_N, 0, -1 },
	{ 'O', CMODE_UNREAL_O, 0, -1 },
	{ 'Q', CMODE_UNREAL_Q, 0, -1 },
	{ 'R', CMODE_R, 0, -1 },
	{ 'S', CMODE_UNREAL_S, 0, -1 },
	{ 'T', CMODE_UNREAL_T, 0, -1 },
	{ 'V', CMODE_UNREAL_V, 0, -1 },
	{ 0 }
};

legacydb::CBModeInfo *cbmodeinfos = 0;

/*************************************************************************/

legacydb::DefEntry def_levels[CA_SIZE] = 
{
    legacydb::DefEntry(CA_INVITE,         "INVITE",        5),
    legacydb::DefEntry(CA_AKICK,          "AKICK",         10),
    legacydb::DefEntry(CA_SET,     		"SET",           ACCESS_INVALID),
    legacydb::DefEntry(CA_UNBAN,          "UNBAN",         5),
    legacydb::DefEntry(CA_AUTOOP,         "AUTOOP",        5),
    legacydb::DefEntry(CA_AUTODEOP,       "AUTODEOP",     -1),
    legacydb::DefEntry(CA_AUTOVOICE,      "AUTOVOICE",     3),
    legacydb::DefEntry(CA_OPDEOP,         "OPDEOP",        5),
    legacydb::DefEntry(CA_ACCESS_LIST,    "ACC-LIST",      1),
    legacydb::DefEntry(CA_CLEAR,          "CLEAR",         ACCESS_INVALID),
    legacydb::DefEntry(CA_NOJOIN,         "NOJOIN",       -2),
    legacydb::DefEntry(CA_ACCESS_CHANGE,  "ACC-CHANGE",    10),
    legacydb::DefEntry(CA_MEMO,           "MEMO",          10),
    legacydb::DefEntry(CA_ASSIGN,         "ASSIGN",        ACCESS_INVALID),
    legacydb::DefEntry(CA_BADWORDS,       "BADWORDS",      10),
    legacydb::DefEntry(CA_NOKICK,         "NOKICK",        1),
    legacydb::DefEntry(CA_FANTASIA,       "FANTASIA",      3),
    legacydb::DefEntry(CA_SAY,            "SAY",           5),
    legacydb::DefEntry(CA_GREET,          "GREET",         5),
    legacydb::DefEntry(CA_VOICEME,        "VOICEME",       3),
    legacydb::DefEntry(CA_VOICE,          "VOICE",         5),
    legacydb::DefEntry(CA_GETKEY,         "GETKEY",        5),
    legacydb::DefEntry(CA_AUTOHALFOP,     "AUTOHALFOP",    4),
    legacydb::DefEntry(CA_AUTOPROTECT,    "AUTOPROTECT",   10),
    legacydb::DefEntry(CA_OPDEOPME,       "OPDEOPME",      5),
    legacydb::DefEntry(CA_HALFOPME,       "HALFOPME",      4),
    legacydb::DefEntry(CA_HALFOP,         "HALFOP",        5),
    legacydb::DefEntry(CA_PROTECTME,      "PROTECTME",     10),
    legacydb::DefEntry(CA_PROTECT,        "PROTECT",       ACCESS_INVALID),
    legacydb::DefEntry(CA_KICKME,         "KICKME",        5),
    legacydb::DefEntry(CA_KICK,           "KICK",          5),
    legacydb::DefEntry(CA_SIGNKICK,       "SIGNKICK",      ACCESS_INVALID),
    legacydb::DefEntry(CA_BANME,          "BANME",         5),
    legacydb::DefEntry(CA_BAN,            "BAN",           5),
    legacydb::DefEntry(CA_TOPIC,          "TOPIC",         ACCESS_INVALID),
    legacydb::DefEntry(CA_INFO,           "INFO",          ACCESS_INVALID)
};

/*************************************************************************/

legacydb::DefEntry def_kbb[KBB_SIZE] =
{
	legacydb::DefEntry(KBB_BOLDS,      "BOLDS",      0),
	legacydb::DefEntry(KBB_COLORS,     "COLORS",     0),
	legacydb::DefEntry(KBB_REVERSES,   "REVERSES",   0),
	legacydb::DefEntry(KBB_UNDERLINES, "UNDERLINES", 0),
	legacydb::DefEntry(KBB_BADWORDS,   "BADWORDS",   0),
	legacydb::DefEntry(KBB_CAPS,       "CAPS",       0),
	legacydb::DefEntry(KBB_FLOOD,      "FLOOD",      0),
	legacydb::DefEntry(KBB_REPEAT,     "REPEAT",     0)
};

/*************************************************************************/

char *get_mlock_modes(uint32_t mlock_on, uint32_t mlock_off, char mlock_p[5][512])
{
	static char res[4096];
	
	char *end = res;

	if (mlock_on || mlock_off) 
	{
		legacydb::CBModeInfo *cbmi = cbmodeinfos;
		
		if (mlock_on) 
		{
			*end++ = '+';
	
			do {
				if (mlock_on & cbmi->flag)
					*end++ = cbmi->mode;
			} while ((++cbmi)->flag != 0);
			
			cbmi = cbmodeinfos;
		}
		
		if (mlock_off) 
		{
			*end++ = '-';
	
			do {
				if (mlock_off & cbmi->flag)
					*end++ = cbmi->mode;
			} while ((++cbmi)->flag != 0);
			
			cbmi = cbmodeinfos;
		}
		
		if (mlock_on) 
		{
			do {
				if (cbmi->pidx != -1 && (mlock_on & cbmi->flag)) 
				{
					char const *value = mlock_p[cbmi->pidx];
			
					if (value) 
					{
						*end++ = ' ';
						while (*value)
							*end++ = *value++;
					}
				}
			} while ((++cbmi)->flag != 0);
		}
	}
	
	*end = 0;
	
	return res;
}

/*************************************************************************/

} /* anonymous namespace */

/*************************************************************************/

namespace legacydb
{

/*************************************************************************/

bool XMLExport::add_field(databases::XMLWriter &xw, char const *name,
		uint16_t type, size_t size, uint16_t attrs, char const *defval,
		char const *filename)
{
	if (!xw.add_field(name, type, size, attrs, defval))
	{
		errorf("Can't add field %s to file %s", name, filename);
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool XMLExport::add_index(databases::XMLIndex &xi, std::string const &filename)
{
	if (!xi.add(filename))
	{
		errorf("Can't add file %s to index file", filename.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool XMLExport::add_ivalue(databases::XMLWriter &xw, char const *name,
		int32_t value, char const *filename)
{
	if (!xw.add_ivalue(name, value))
	{
		errorf("Can't add value %s to file %s", name, filename);
		return false;
	}
	
	return true;
}
				
/*************************************************************************/

bool XMLExport::add_rivalue(databases::XMLWriter &xw, char const *name,
		databases::Record::identifier_type value, char const *filename)
{
	if (!xw.add_rivalue(name, value))
	{
		errorf("Can't add value %s to file %s", name, filename);
		return false;
	}
	
	return true;
}
				
/*************************************************************************/

bool XMLExport::add_uivalue(databases::XMLWriter &xw, char const *name,
		uint32_t value, char const *filename)
{
	if (!xw.add_uivalue(name, value))
	{
		errorf("Can't add value %s to file %s", name, filename);
		return false;
	}
	
	return true;
}
				
/*************************************************************************/

bool XMLExport::add_utvalue(databases::XMLWriter &xw, char const *name,
		time_t value, char const *filename)
{
	if (!xw.add_utvalue(name, value))
	{
		errorf("Can't add value %s to file %s", name, filename);
		return false;
	}
	
	return true;
}
				
/*************************************************************************/

bool XMLExport::add_value(databases::XMLWriter &xw, char const *name,
		char const *value, char const *filename)
{
	if (!xw.add_value(name, value))
	{
		errorf("Can't add value %s to file %s", name, filename);
		return false;
	}
	
	return true;
}
				
/*************************************************************************/

bool XMLExport::begin_record(databases::XMLWriter &xw, char const *filename)
{
	if (!xw.begin_record())
	{
		errorf("Can't create new record in file %s", filename);
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool XMLExport::count(databases::XMLWriter &xw, uint32_t count,
		char const *filename)
{
	if (!xw.count(count))
	{
		errorf("Can't set count in file %s", filename);
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool XMLExport::end_record(databases::XMLWriter &xw, char const *filename)
{
	if (!xw.end_record())
	{
		errorf("Can't add new record in file %s", filename);
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool XMLExport::export_data(std::string const &dir, databases::XMLIndex &index,
		std::string const &database)
{	
	if (!export_nickserv(dir, index, database))
		return false;
	if (!export_botserv(dir, index, database))
		return false;
	if (!export_chanserv(dir, index, database))
		return false;
	if (!export_operserv(dir, index, database))
		return false;
	if (!export_news(dir, index, database))
		return false;
	if (!export_exceptions(dir, index, database))
		return false;
	
	return true;
}

/*************************************************************************/

#define SAFE(x) do {                          \
	if ((x) < 0)                              \
	{                                         \
		errorf("Read error on bot.db"); \
        return false;                         \
    }                                         \
} while (0)

bool XMLExport::export_botserv(std::string const &dir, XMLIndex &xi,
		std::string const &database)
{
	if ((database.empty() || database == "botserv_bots") &&
			!xi.exists("botserv_bots"))
	{
		XMLWriter xw;
		
		if (!version(xw, 1, "botserv_bots.xml"))
			return false;

		if (!add_field(xw, "nick", Field::ftString, 63, 0, NULL,
				"botserv_bots.xml"))
			return false;
		if (!add_field(xw, "username", Field::ftString, 31, 0, NULL,
				"botserv_bots.xml"))
			return false;
		if (!add_field(xw, "hostname", Field::ftString, 63, 0, NULL,
				"botserv_bots.xml"))
			return false;
		if (!add_field(xw, "realname", Field::ftString, 127, 0, NULL,
				"botserv_bots.xml"))
			return false;
		if (!add_field(xw, "created", Field::ftUnixTime, 0, 0, "0",
				"botserv_bots.xml"))
			return false;
		if (!add_field(xw, "flags", Field::ftUnsigned, 2, 0, "0",
				"botserv_bots.xml"))
			return false;
		
		dbFILE *f;
		
		if (!(f = open_db("bot.db")))
		{
			errorf("Can't open file bot.db");
			return false;
		}
		
		int ver;
		
		if ((ver = get_file_version(f)) < 10)
		{
			errorf("Unsupported version (%d) on bot.db", ver);
			return false;
		}
		
		size_t cnt = 0;
		Record::identifier_type last_id = 0;
		
		int c;
		uint16_t tmp16;
		uint32_t tmp32;
		
		while ((c = getc_db(f)) != 0) 
		{
			char *s;
			
			if (c != 1)
			{
				errorf("Invalid format in bot.db %d", c);
				return false;
			}
			
			if (!begin_record(xw, "botserv_bots.xml"))
				return false;
			
			if (ver >= 11)
				SAFE(read_int32(&tmp32, f));
			else
				tmp32 = ++last_id;
			
			if (!add_rivalue(xw, "id", tmp32, "botserv_bots.xml"))
				return false;
			
			SAFE(read_string(&s, f));
			if (!add_value(xw, "nick", s, "botserv_bots.xml"))
				return false;
			
			bots.insert(make_pair(string(s), tmp32));
			
			delete [] s;
			SAFE(read_string(&s, f));
			if (!add_value(xw, "username", s, "botserv_bots.xml"))
				return false;
			delete [] s;
			SAFE(read_string(&s, f));
			if (!add_value(xw, "hostname", s, "botserv_bots.xml"))
				return false;
			delete [] s;
			SAFE(read_string(&s, f));
			if (!add_value(xw, "realname", s, "botserv_bots.xml"))
				return false;
			delete [] s;
			SAFE(read_int16(&tmp16, f));
			if (!add_uivalue(xw, "flags", tmp16, "botserv_bots.xml"))
				return false;
			SAFE(read_int32(&tmp32, f));
			if (!add_utvalue(xw, "created", tmp32, "botserv_bots.xml"))
				return false;
			SAFE(read_int16(&tmp16, f)); /* chancount */
			
			if (!end_record(xw, "botserv_bots.xml"))
				return false;
			
			++cnt;
		}
		
		close_db(f);
		
		if (!count(xw, cnt, "botserv_bots.xml"))
			return false;
		
		if (!save(xw, absolute_path(dir, "botserv_bots.xml")))
			return false;
		
		if (!add_index(xi, "botserv_bots.xml"))
			return false;
			
		me->core().log(InfoLegacyDB(), "Exported botserv_bots");
	}
	
	return true;
}

#undef SAFE

/*************************************************************************/

#define SAFE(x) do {                          \
	if ((x) < 0)                              \
	{                                         \
		errorf("Read error on chan.db"); \
        return false;                         \
    }                                         \
} while (0)

bool XMLExport::export_chanserv(std::string const &dir, XMLIndex &xi,
		std::string const &database)
{
	bool channels = false, levels = false, access = false, akicks = false;
	bool kbb = false, badwords = false;
	
	if ((database.empty() || database == "chanserv_channels") &&
			!xi.exists("chanserv_channels"))
		channels = true;
	if ((database.empty() || database == "chanserv_levels") &&
			!xi.exists("chanserv_levels"))
		levels = true;
	if ((database.empty() || database == "chanserv_access") &&
			!xi.exists("chanserv_access"))
		access = true;
	if ((database.empty() || database == "chanserv_akicks") &&
			!xi.exists("chanserv_akicks"))
		akicks = true;
	if ((database.empty() || database == "botserv_kbb") &&
			!xi.exists("chanserv_kbb"))
		kbb = true;
	if ((database.empty() || database == "botserv_badwords") &&
			!xi.exists("chanserv_badwords"))
		badwords = true;
	
	if (!channels && !levels && !access && !akicks && !kbb && !badwords)
		return true; 
	
	char ircd[12];
	
	if (channels)
	{
		/* Parse command-line parameters. */
		typedef cmdline::CmdLineParser::ParamsList::const_iterator PIType;
		
		PIType it;
		
		it = me->core().clp().params().find("ircd");
		if (it == me->core().clp().params().end())
		{
			me->core().log(xlog::LogCritical(),
					"Missing --ircd=name command-line parameter (where name is"
					" either bahamut, dreamforge, ultimate or unreal).");
			return false;
		}
		
		char const *p_ircd = it->second.c_str();
		
		if (!stricmp(p_ircd, "bahamut"))
		{
			cbmodeinfos = cbmi_bahamut;
			snprintf(ircd, sizeof(ircd), "bahamut");
		} 
		else if (!stricmp(p_ircd, "dreamforge"))
		{
			cbmodeinfos = cbmi_dreamforge;
			snprintf(ircd, sizeof(ircd), "dreamforge");
		}
		else if (!stricmp(p_ircd, "ultimate"))
		{
			cbmodeinfos = cbmi_ultimate;
			snprintf(ircd, sizeof(ircd), "ultimate");
		}
		else if (!stricmp(p_ircd, "unreal"))
		{
			cbmodeinfos = cbmi_unreal;
			snprintf(ircd, sizeof(ircd), "unreal");
		}
	}
	
	XMLWriter *xw = 0;
	size_t chcount = 0;
	Record::identifier_type last_chid = 0;
	
	if (channels)
	{
		xw = new XMLWriter;
			
		if (!version(*xw, 1, "chanserv_channels.xml"))
			return false;
	
		if (!add_field(*xw, "name", Field::ftString, 63, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "owner", Field::ftIdentifier, 0, Field::faOwner,
				"0", "chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "successor", Field::ftIdentifier, 0,
				Field::faOwner, "0", "chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "created", Field::ftUnixTime, 0, 0, "0",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "last_used", Field::ftUnixTime, 0, 0, "0",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "enc", Field::ftString, 255, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "pass", Field::ftString, 192, Field::faCase,
				"", "chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "flags", Field::ftUnsigned, 4, 0, "1073741824",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "desc", Field::ftString, 255, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "email", Field::ftString, 127, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "url", Field::ftString, 255, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "entry_message", Field::ftString, 410, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "bantype", Field::ftUnsigned, 1, 0, "255",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "mlock_ircd", Field::ftString, 63, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "mlock_modes", Field::ftString, 450, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "topic", Field::ftString, 450, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "topic_setter", Field::ftString, 63, 0, "",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "topic_time", Field::ftUnixTime, 0, 0, "0",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "bot", Field::ftIdentifier, 0, Field::faOwner,
				"0", "chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "botflags", Field::ftUnsigned, 4, 0, "64",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "caps_min", Field::ftUnsigned, 1, 0, "10",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "caps_percent", Field::ftUnsigned, 1, 0, "25",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "flood_lines", Field::ftUnsigned, 1, 0, "6",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "flood_secs", Field::ftUnsigned, 1, 0, "10",
				"chanserv_channels.xml"))
			return false;
		if (!add_field(*xw, "repeats", Field::ftUnsigned, 1, 0, "3",
				"chanserv_channels.xml"))
			return false;
	}
	
	XMLWriter *xwacc = 0;
	size_t acccount = 0;
	Record::identifier_type last_accid = 0;
	
	if (access)
	{
		xwacc = new XMLWriter;
			
		if (!version(*xwacc, 1, "chanserv_access.xml"))
			return false;
	
		if (!add_field(*xwacc, "owner", Field::ftIdentifier, 0, Field::faOwner,
				"0", "chanserv_access.xml"))
			return false;
		if (!add_field(*xwacc, "group", Field::ftIdentifier, 0, Field::faOwner,
				"0", "chanserv_access.xml"))
			return false;
		if (!add_field(*xwacc, "level", Field::ftInteger, 2, 0, "0",
				"chanserv_access.xml"))
			return false;
		if (!add_field(*xwacc, "last_seen", Field::ftUnixTime, 0, 0, "0",
				"chanserv_access.xml"))
			return false;
	}
	
	XMLWriter *xwlv = 0;
	size_t lvcount = 0;
	Record::identifier_type last_lvid = 0;
	
	if (levels)
	{
		xwlv = new XMLWriter;
			
		if (!version(*xwlv, 1, "chanserv_levels.xml"))
			return false;
	
		if (!add_field(*xwlv, "owner", Field::ftIdentifier, 0, Field::faOwner,
				"0", "chanserv_levels.xml"))
			return false;
		if (!add_field(*xwlv, "name", Field::ftString, 31, 0, "",
				"chanserv_levels.xml"))
			return false;
		if (!add_field(*xwlv, "level", Field::ftInteger, 2, 0, "0",
				"chanserv_levels.xml"))
			return false;
	}
	
	XMLWriter *xwak = 0;
	size_t akcount = 0;
	Record::identifier_type last_akid = 0;
	
	if (akicks)
	{
		xwak = new XMLWriter;
			
		if (!version(*xwak, 1, "chanserv_akicks.xml"))
			return false;
	
		if (!add_field(*xwak, "owner", Field::ftIdentifier, 0, Field::faOwner,
				"0", "chanserv_akicks.xml"))
			return false;
		if (!add_field(*xwak, "flags", Field::ftUnsigned, 1, 0, "0",
				"chanserv_akicks.xml"))
			return false;
		if (!add_field(*xwak, "mask", Field::ftString, 191, 0, "",
				"chanserv_akicks.xml"))
			return false;
		if (!add_field(*xwak, "group", Field::ftIdentifier, 0, Field::faOwner,
				"0", "chanserv_akicks.xml"))
			return false;
		if (!add_field(*xwak, "creator", Field::ftString, 63, 0, "",
				"chanserv_akicks.xml"))
			return false;
		if (!add_field(*xwak, "reason", Field::ftString, 255, 0, "",
				"chanserv_akicks.xml"))
			return false;
		if (!add_field(*xwak, "since", Field::ftUnixTime, 0, 0, "0",
				"chanserv_akicks.xml"))
			return false;
	}
	
	XMLWriter *xwkbb = 0;
	size_t kbbcount = 0;
	Record::identifier_type last_kbbid = 0;
	
	if (kbb)
	{
		xwkbb = new XMLWriter;
			
		if (!version(*xwkbb, 1, "botserv_kbb.xml"))
			return false;
	
		if (!add_field(*xwkbb, "owner", Field::ftIdentifier, 0, Field::faOwner,
				"0", "botserv_kbb.xml"))
			return false;
		if (!add_field(*xwkbb, "name", Field::ftString, 31, 0, "",
				"botserv_kbb.xml"))
			return false;
		if (!add_field(*xwkbb, "count", Field::ftUnsigned, 1, 0, "0",
				"botserv_kbb.xml"))
			return false;
	}
	
	XMLWriter *xwbw = 0;
	size_t bwcount = 0;
	Record::identifier_type last_bwid = 0;
	
	if (badwords)
	{
		xwbw = new XMLWriter;
			
		if (!version(*xwbw, 1, "botserv_badwords.xml"))
			return false;
	
		if (!add_field(*xwbw, "owner", Field::ftIdentifier, 0, Field::faOwner,
				"0", "botserv_badwords.xml"))
			return false;
		if (!add_field(*xwbw, "word", Field::ftString, 63, 0, "",
				"botserv_badwords.xml"))
			return false;
		if (!add_field(*xwbw, "type", Field::ftUnsigned, 1, 0, "0",
				"botserv_badwords.xml"))
			return false;
	}
	
	dbFILE *f;
	
	if (!(f = open_db("chan.db")))
	{
		errorf("Can't open file chan.db");
		return false;
	}
	
	int ver = get_file_version(f);
	
	if (ver < 13)
	{
		errorf("Unsupported version %d on chan.db", ver);
		close_db(f);
		return false;
	}
	
	int c;
	char *s;
	uint16_t tmp16;
	uint32_t tmp32;
	char name[CHANMAX], pass[PASSMAX], nick[NICKMAX];
	
	char *reason, *creator;
	time_t since;
	uint32_t flags;
	
	uint32_t mlock_on, mlock_off;
	char mlock_p[5][512];
	
	Record::identifier_type id;
	id_map::const_iterator it;
	
	for (int i = 0; i < 256; ++i)
	{
		while ((c = getc_db(f)) != 0)
		{
			if (c != 1)
			{
				errorf("Invalid format in chan.db");
				return false;
			}
			
			bool dump = false, forbidden = false;
			
			if (channels && !begin_record(*xw, "chanserv_channels.xml"))
				return false;
			
			if (ver >= 18)
				SAFE(read_int32(&id, f));
			else
				id = ++last_chid;
			
			if (channels && !add_rivalue(*xw, "id", id,
					"chanserv_channels.xml"))
				return false;
			
			SAFE(read_buffer(name, f));
			
			if (channels && !add_value(*xw, "name", name,
					"chanserv_channels.xml"))
				return false;
			
			SAFE(read_string(&s, f));
			if (channels && s) 
			{
				it = groups.find(s);
				
				if (it == groups.end())
				{
					me->core().log(InfoLegacyDB(),
							"Ignoring channel %s with invalid founder %s",
							name, s);
					dump = true;
				} 
				else
				{
					if (channels && !add_rivalue(*xw, "owner", it->second,
							"chanserv_channels.xml"))
						return false;
				}
				
				delete [] s;
			}
			
		    SAFE(read_string(&s, f));
		    if (channels && s) 
		    {
				it = groups.find(s);
				
				if (it == groups.end())
				{
					if (channels && !add_rivalue(*xw, "successor", 0,
							"chanserv_channels.xml"))
						return false;
				} 
				else
				{
					if (channels && !add_rivalue(*xw, "successor", it->second,
							"chanserv_channels.xml"))
						return false;
				}
				
		    	delete [] s;
			} else
			{
				if (channels && !add_rivalue(*xw, "successor", 0,
						"chanserv_channels.xml"))
					return false;
			}
			
			SAFE(read_buffer(pass, f));
			//ci->founderpass(pass);
			
			SAFE(read_string(&s, f));
			if (channels && !add_value(*xw, "desc", s,
					"chanserv_channels.xml"))
				return false;
			if (s)
				delete [] s;
			SAFE(read_string(&s, f));
			if (channels && !add_value(*xw, "url", s,
					"chanserv_channels.xml"))
				return false;
			if (s)
				delete [] s;
			SAFE(read_string(&s, f));
			if (channels && !add_value(*xw, "email", s,
					"chanserv_channels.xml"))
				return false;
			if (s)
				delete [] s;
			
			SAFE(read_int32(&tmp32, f));
			since = tmp32;
			if (channels && !add_utvalue(*xw, "created", tmp32,
					"chanserv_channels.xml"))
				return false;
			SAFE(read_int32(&tmp32, f));
			if (channels && !add_utvalue(*xw, "last_used", tmp32,
					"chanserv_channels.xml"))
				return false;
			
			SAFE(read_string(&s, f));
			if (channels && !add_value(*xw, "topic", s,
					"chanserv_channels.xml"))
				return false;
			if (s)
				delete [] s;
			SAFE(read_buffer(nick, f));
			if (channels && !add_value(*xw, "topic_setter", nick,
					"chanserv_channels.xml"))
				return false;
			SAFE(read_int32(&tmp32, f));
			if (channels && !add_utvalue(*xw, "topic_time", tmp32,
					"chanserv_channels.xml"))
				return false;
			
			SAFE(read_int32(&flags, f));
			
			if (*pass && (flags & CI_ENCRYPTEDPW))
			{
				if (channels && !add_value(*xw, "enc", "legacydb",
						"chanserv_channels.xml"))
					return false;
				
				char truepass[33], *end = truepass;
				for (int k = 0; k < 16; ++k)
					end += sprintf(end, "%02x", (unsigned char)pass[k]);
				truepass[32] = 0;
				
				if (channels && !add_value(*xw, "pass", truepass,
						"chanserv_channels.xml"))
					return false;
			} else if (*pass)
			{
				if (channels && !add_value(*xw, "enc", 0,
						"chanserv_channels.xml"))
					return false;
				
				if (channels && !add_value(*xw, "pass", pass,
						"chanserv_channels.xml"))
					return false;
			}
			
			if (flags & CI_VERBOTEN)
				forbidden = true;
			
			flags &= ~(CI_UNUSED | CI_TEMPORARY);
			
			if (channels && !add_uivalue(*xw, "flags", flags,
					"chanserv_channels.xml"))
				return false;
			
			SAFE(read_string(&creator, f));
			//ci->forbidby(s);
			/*if (s)
				delete [] s;*/
			SAFE(read_string(&reason, f));
			//ci->forbidreason(s);
			/*if (s)
				delete [] s;*/
			
			SAFE(read_int16(&tmp16, f));
			if (channels && !add_uivalue(*xw, "bantype", tmp16,
					"chanserv_channels.xml"))
				return false;
			
			SAFE(read_int16(&tmp16, f));
			int n_levels = tmp16;
			//reset_levels(ci);
			for (int j = 0; j < n_levels; j++) 
			{
				SAFE(read_int16(&tmp16, f));
				int16_t tmplv = tmp16;
				
				if (j < CA_SIZE && def_levels[j].defval() != tmplv)
				{
					if (levels && !dump &&
							!begin_record(*xwlv, "chanserv_levels.xml"))
						return false;
					
					if (levels && !dump &&
							!add_rivalue(*xwlv, "id", ++last_lvid,
							"chanserv_levels.xml"))
						return false;
					
					if (levels && !dump &&
							!add_rivalue(*xwlv, "owner", id,
							"chanserv_levels.xml"))
						return false;
					
					if (levels && !dump &&
							!add_value(*xwlv, "name", def_levels[j].name(),
							"chanserv_levels.xml"))
						return false;
					
					if (levels && !dump &&
							!add_ivalue(*xwlv, "level", tmplv,
							"chanserv_levels.xml"))
						return false;
					
					if (levels && !dump && 
							!end_record(*xwlv, "chanserv_levels.xml"))
						return false;
						
					if (levels && !dump)
						++lvcount;
				}
			}
			
			uint16_t accesscount;
			SAFE(read_int16(&accesscount, f));
			//ci->accesscount(tmp16);
			if (accesscount) 
			{
			    for (int j = 0; j < accesscount; j++) 
			    {
			    	uint16_t in_use;
			    	
				    SAFE(read_int16(&in_use, f));
				    
					if (in_use) 
					{
						bool accdump = false;
						
						if (access && !dump &&
								!begin_record(*xwacc, "chanserv_access.xml"))
							return false;
						
						if (access && !dump &&
								!add_rivalue(*xwacc, "id", ++last_accid,
								"chanserv_access.xml"))
							return false;
						
						if (access && !dump &&
								!add_rivalue(*xwacc, "owner", id,
								"chanserv_access.xml"))
							return false;
						
						SAFE(read_int16(&tmp16, f));
						int16_t tmplv = tmp16;
						
						if (access && !dump &&
								!add_ivalue(*xwacc, "level", tmplv,
								"chanserv_access.xml"))
							return false;
					    
					    SAFE(read_string(&s, f));
					    if (access && !dump) 
					    {
					    	it = groups.find(s);
				
							if (it == groups.end())
							{
								me->core().log(InfoLegacyDB(),
										"Ignoring invalid access %s for channel %s",
										s, name);
								accdump = true;
							} 
							else
							{
								if (!add_rivalue(*xwacc, "group",
										it->second, "chanserv_access.xml"))
									return false;
							}
					    }
					    delete [] s;
						
						
						SAFE(read_int32(&tmp32, f));
						if (access && !dump &&
								!add_utvalue(*xwacc, "last_seen", tmp32,
								"chanserv_access.xml"))
							return false;
						
						if (access && !dump)
						{
							if (accdump)
								xwacc->dump_record();
							else
							{
								if (!end_record(*xwacc, "chanserv_access.xml"))
									return false;
							
								++acccount;
							}
						}
					}
			    }
			}
			
			uint16_t akickcount;
			SAFE(read_int16(&akickcount, f));
			
			if (akickcount)
			{
			    for (int j = 0; j < akickcount; j++) 
			    {
			    	uint16_t akflags;
			    	SAFE(read_int16(&akflags, f));
			    	
					if (akflags & AK_USED) 
					{
						bool akdump = false;
						
						if (akicks && !dump &&
								!begin_record(*xwak, "chanserv_akicks.xml"))
							return false;
						
						if (akicks && !dump &&
								!add_rivalue(*xwak, "id", ++last_akid,
								"chanserv_akicks.xml"))
							return false;
						
						if (akicks && !dump &&
								!add_rivalue(*xwak, "owner", id,
								"chanserv_akicks.xml"))
							return false;
						
						akflags &= ~AK_UNUSED;
						
						if (akicks && !dump &&
								!add_utvalue(*xwak, "flags", akflags,
								"chanserv_akicks.xml"))
							return false;
						
				    	SAFE(read_string(&s, f));
				    	if (akflags & AK_ISNICK) 
				    	{
				    		if (akicks && !dump) 
						    {
						    	it = groups.find(s);
					
								if (it == groups.end())
								{
									me->core().log(InfoLegacyDB(),
											"Ignoring invalid akick %s "
											"for channel %s", s, name);
									akdump = true;
								} 
								else
								{
									if (!add_value(*xwak, "mask",
											"", "chanserv_akicks.xml"))
										return false;
									
									if (!add_rivalue(*xwak, "group",
											it->second, "chanserv_akicks.xml"))
										return false;
								}
						    }
				    	} else 
				    	{
				    		if (akicks && !dump &&
									!add_value(*xwak, "mask", s,
									"chanserv_akicks.xml"))
								return false;
				    		
				    		if (akicks && !dump &&
									!add_rivalue(*xwak, "group", 0,
									"chanserv_akicks.xml"))
								return false;
				    	}
				    	delete [] s;
				    	
				    	SAFE(read_string(&s, f));
				    	if (akicks && !dump &&
								!add_value(*xwak, "reason", s,
								"chanserv_akicks.xml"))
							return false;
						if (s)
				    		delete [] s;
						SAFE(read_string(&s, f));
						if (akicks && !dump &&
								!add_value(*xwak, "creator", s,
								"chanserv_akicks.xml"))
							return false;
			    		if (s)
			    			delete [] s;
			    		SAFE(read_int32(&tmp32, f));
			    		if (akicks && !dump &&
								!add_utvalue(*xwak, "since", tmp32,
								"chanserv_akicks.xml"))
							return false;
			    		
			    		if (akicks && !dump)
						{
							if (akdump)
								xwak->dump_record();
							else
							{
								if (!end_record(*xwak, "chanserv_akicks.xml"))
									return false;
							
								++akcount;
							}
						}
					}
			    }
			}
			
			SAFE(read_int32(&tmp32, f));
			
			if (!stricmp(ircd, "unreal") && ver < 19)
				tmp32 &= ~REMOVED_CMODE_H;
			mlock_on = tmp32;

			SAFE(read_int32(&tmp32, f));
			if (!stricmp(ircd, "unreal") && ver < 19)
				tmp32 &= ~REMOVED_CMODE_H;
			mlock_off = tmp32;
			
			SAFE(read_int32(&tmp32, f));
			/* Limit */
			snprintf(mlock_p[1], sizeof(mlock_p[1]), "%u", tmp32);
			
			SAFE(read_string(&s, f));
			/* Key */
			if (s)
				strncpyzt(mlock_p[0], s, sizeof(mlock_p[0]) - 1);
			else
				mlock_p[0][0] = 0;
			if (s)
				delete [] s;
			
			SAFE(read_string(&s, f));
			/* Flood */
			if (s)
				strncpyzt(mlock_p[2], s, sizeof(mlock_p[2]) - 1);
			else
				mlock_p[2][0] = 0;
			if (s) 
				delete [] s;
			
			if (ver >= 17) 
			{
				SAFE(read_string(&s, f));
				/* Join rate */
				if (s)
					strncpyzt(mlock_p[3], s, sizeof(mlock_p[3]) - 1);
				else
					mlock_p[3][0] = 0;
				if (s) 
					delete [] s;
			}
			SAFE(read_string(&s, f));
			/* Redirect */
			if (s)
				strncpyzt(mlock_p[4], s, sizeof(mlock_p[4]) - 1);
			else
				mlock_p[4][0] = 0;
			if (s) 
				delete [] s;
			
			if (channels && !add_value(*xw, "mlock_ircd", ircd, 
					"chanserv_channels.xml"))
				return false;
			
			if (channels && !add_value(*xw, "mlock_modes",
					get_mlock_modes(mlock_on, mlock_off, mlock_p), 
					"chanserv_channels.xml"))
				return false;
			
			uint16_t memocount;
			SAFE(read_int16(&memocount, f));
			//ci->memos.memocount(tmp16);
			SAFE(read_int16(&tmp16, f));
			//ci->memos.memomax(tmp16);
			if (memocount) 
			{
			    char sender[NICKMAX];
			    
			    for (int j = 0; j < memocount; j++) 
			    {
					SAFE(read_int32(&tmp32, f));
					//memos->number(tmp32);
					SAFE(read_int16(&tmp16, f));
					//memos->flags_add(tmp16);
					SAFE(read_int32(&tmp32, f));
					//memos->time(tmp32);
					SAFE(read_buffer(sender, f));
					//memos->sender(sender);
					SAFE(read_string(&s, f));
					//memos->text(s);
					delete [] s;
			    }
			}
			
			SAFE(read_string(&s, f));
			if (channels && !add_value(*xw, "entry_message", s, 
					"chanserv_channels.xml"))
				return false;
			if (s)
				delete [] s;
			
			/* BotServ options */
			
			int n_kbb, bwc;
			
			SAFE(read_string(&s, f));
			if (s) 
		    {
				it = bots.find(s);
				
				if (it == bots.end())
				{
					if (channels && !add_rivalue(*xw, "bot", 0,
							"chanserv_channels.xml"))
						return false;
				} 
				else
				{
					if (channels && !add_rivalue(*xw, "bot", it->second,
							"chanserv_channels.xml"))
						return false;
				}
				
		    	delete [] s;
			} else
			{
				if (channels && !add_rivalue(*xw, "bot", 0,
						"chanserv_channels.xml"))
					return false;
			}
			
			SAFE(read_int32(&tmp32, f));
			tmp32 &= ~BS_UNUSED;
			
			if (channels && !add_uivalue(*xw, "botflags", tmp32,
					"chanserv_channels.xml"))
				return false;
			
			SAFE(read_int16(&tmp16, f));
			n_kbb = tmp16;
			for (int j = 0; j < n_kbb; j++) 
			{
				SAFE(read_int16(&tmp16, f));
				
				if (j < KBB_SIZE && def_kbb[j].defval() != tmp16)
				{
					if (kbb && !dump &&
							!begin_record(*xwkbb, "botserv_kbb.xml"))
						return false;
					
					if (kbb && !dump &&
							!add_rivalue(*xwkbb, "id", ++last_kbbid,
							"botserv_kbb.xml"))
						return false;
					
					if (kbb && !dump &&
							!add_rivalue(*xwkbb, "owner", id,
							"botserv_kbb.xml"))
						return false;
					
					if (kbb && !dump &&
							!add_value(*xwkbb, "name", def_kbb[j].name(),
							"botserv_kbb.xml"))
						return false;
					
					if (kbb && !dump &&
							!add_uivalue(*xwkbb, "count", tmp16,
							"botserv_kbb.xml"))
						return false;
					
					if (kbb && !dump && 
							!end_record(*xwkbb, "botserv_kbb.xml"))
						return false;
						
					if (kbb && !dump)
						++kbbcount;
				}
			}
			
			SAFE(read_int16(&tmp16, f));
			if (!tmp16)
				tmp16 = 10;
			if (channels && !add_uivalue(*xw, "caps_min", tmp16,
					"chanserv_channels.xml"))
				return false;
			SAFE(read_int16(&tmp16, f));
			if (!tmp16 || tmp16 > 100)
				tmp16 = 25;
			if (channels && !add_uivalue(*xw, "caps_percent", tmp16,
					"chanserv_channels.xml"))
				return false;
			SAFE(read_int16(&tmp16, f));
			if (tmp16 < 2)
				tmp16 = 6;
			if (channels && !add_uivalue(*xw, "flood_lines", tmp16,
					"chanserv_channels.xml"))
				return false;
			SAFE(read_int16(&tmp16, f));
			if (!tmp16)
				tmp16 = 10;
			if (channels && !add_uivalue(*xw, "flood_secs", tmp16,
					"chanserv_channels.xml"))
				return false;
			SAFE(read_int16(&tmp16, f));
			if (tmp16 < 2)
				tmp16 = 3;
			if (channels && !add_uivalue(*xw, "repeats", tmp16,
					"chanserv_channels.xml"))
				return false;
			SAFE(read_int16(&tmp16, f));
			bwc = tmp16;
			//ci->bwcount(bwcount);
			if (bwc) 
			{
				for (int j = 0; j < bwc; j++) 
				{
					uint16_t in_use;
					
					SAFE(read_int16(&in_use, f));
					
					if (in_use) 
					{
						if (badwords && !dump &&
								!begin_record(*xwbw, "botserv_badwords.xml"))
							return false;
						
						if (badwords && !dump &&
								!add_rivalue(*xwbw, "id", ++last_bwid,
								"botserv_badwords.xml"))
							return false;
						
						if (badwords && !dump &&
								!add_rivalue(*xwbw, "owner", id,
								"botserv_badwords.xml"))
							return false;
						
						SAFE(read_string(&s, f));
						if (badwords && !dump && !add_value(*xwbw, "word", s,
								"botserv_badwords.xml"))
							return false;
						delete [] s;
						
						SAFE(read_int16(&tmp16, f));
						if (badwords && !dump &&
								!add_uivalue(*xwbw, "type", tmp16,
								"botserv_badwords.xml"))
							return false;
						
						if (badwords && !dump && 
								!end_record(*xwbw, "botserv_badwords.xml"))
							return false;
						
						if (badwords && !dump)
							++bwcount;
					}
		    	}
			}
			
			if (forbidden)
			{
				Forbid *f = new Forbid;
				f->name(name);
				f->creator(creator ? creator : "Unknown");
				f->reason(reason ? reason : "Unknown reason");
				f->since(since ? since : time(0));
				
				forbids.push_back(f);
				
				me->core().log(InfoLegacyDB(),
						"Ignoring forbidden channel %s (ID #" RECORD_FORMAT
						")", name, id);
				dump = true;
			}
			
			if (creator)
				delete [] creator;
			if (reason)
				delete [] reason;
			
			if (channels && dump)
				xw->dump_record();
			else
			{
				if (channels && !end_record(*xw, "chanserv_channels.xml"))
					return false;		
				++chcount;
			}
		}
	}
	
	if (channels)
	{	
		if (!count(*xw, chcount, "chanserv_channels.xml"))
			return false;
		
		if (!save(*xw, absolute_path(dir, "chanserv_channels.xml")))
			return false;
		
		if (!add_index(xi, "chanserv_channels.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported chanserv_channels");
		delete xw;
	}
	
	if (levels)
	{	
		if (!count(*xwlv, lvcount, "chanserv_levels.xml"))
			return false;
		
		if (!save(*xwlv, absolute_path(dir, "chanserv_levels.xml")))
			return false;
		
		if (!add_index(xi, "chanserv_levels.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported chanserv_levels");
		delete xwlv;
	}
	
	if (access)
	{	
		if (!count(*xwacc, acccount, "chanserv_access.xml"))
			return false;
		
		if (!save(*xwacc, absolute_path(dir, "chanserv_access.xml")))
			return false;
		
		if (!add_index(xi, "chanserv_access.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported chanserv_access");
		delete xwacc;
	}
	
	if (akicks)
	{	
		if (!count(*xwak, akcount, "chanserv_akicks.xml"))
			return false;
		
		if (!save(*xwak, absolute_path(dir, "chanserv_akicks.xml")))
			return false;
		
		if (!add_index(xi, "chanserv_akicks.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported chanserv_akicks");
		delete xwak;
	}
	
	if (kbb)
	{	
		if (!count(*xwkbb, kbbcount, "botserv_kbb.xml"))
			return false;
		
		if (!save(*xwkbb, absolute_path(dir, "botserv_kbb.xml")))
			return false;
		
		if (!add_index(xi, "botserv_kbb.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported botserv_kbb");
		delete xwkbb;
	}
	
	if (badwords)
	{	
		if (!count(*xwbw, bwcount, "botserv_badwords.xml"))
			return false;
		
		if (!save(*xwbw, absolute_path(dir, "botserv_badwords.xml")))
			return false;
		
		if (!add_index(xi, "botserv_badwords.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported botserv_badwords");
		delete xwbw;
	}
	
	close_db(f);
	return true;
}

#undef SAFE

/*************************************************************************/

#define SAFE(x) do {                          \
	if ((x) < 0)                              \
	{                                         \
		errorf("Read error on exception.db"); \
        return false;                         \
    }                                         \
} while (0)

bool XMLExport::export_exceptions(std::string const &dir, XMLIndex &xi,
		std::string const &database)
{
	if ((database.empty() || database == "operserv_exceptions") &&
			!xi.exists("operserv_exceptions"))
	{
		XMLWriter xw;
		
		if (!version(xw, 1, "operserv_exceptions.xml"))
			return false;

		if (!add_field(xw, "mask", Field::ftString, 63, 0, NULL,
				"operserv_exceptions.xml"))
			return false;
		if (!add_field(xw, "limit", Field::ftUnsigned, 2, 0, "0",
				"operserv_exceptions.xml"))
			return false;
		if (!add_field(xw, "creator", Field::ftString, 63, 0, NULL,
				"operserv_exceptions.xml"))
			return false;
		if (!add_field(xw, "reason", Field::ftString, 255, 0, NULL,
				"operserv_exceptions.xml"))
			return false;
		if (!add_field(xw, "since", Field::ftUnixTime, 0, 0, "0",
				"operserv_exceptions.xml"))
			return false;
		if (!add_field(xw, "expires", Field::ftUnixTime, 0, 0, "0",
				"operserv_exceptions.xml"))
			return false;

		dbFILE *f;
		
		if (!(f = open_db("exception.db")))
		{
			errorf("Can't open file exception.db");
			return false;
		}
		
		int ver;
		
		if ((ver = get_file_version(f)) < 7)
		{
			errorf("Unsupported version (%d) on exception.db", ver);
			return false;
		}
		
		uint16_t cnt;
		
		SAFE(read_int16(&cnt, f));
		if (!count(xw, cnt, "operserv_exceptions.xml"))
			return false;
		
		char *s, creator[32];
		uint16_t tmp16;
		uint32_t tmp32;
		
		Record::identifier_type id = 0;
		
		for (int i = 0; i < cnt; i++) 
		{
			if (!begin_record(xw, "operserv_exceptions.xml"))
				return false;
			
			if (!add_rivalue(xw, "id", ++id, "operserv_exceptions.xml"))
				return false;
			
			SAFE(read_string(&s, f));
			if (!add_value(xw, "mask", s, "operserv_exceptions.xml"))
				return false;
			delete [] s;
			SAFE(read_int16(&tmp16, f));
			if (!add_uivalue(xw, "limit", tmp16, "operserv_exceptions.xml"))
				return false;
			SAFE(read_buffer(creator, f));
			if (!add_value(xw, "creator", creator, "operserv_exceptions.xml"))
				return false;
			SAFE(read_string(&s, f));
			if (!add_value(xw, "reason", s, "operserv_exceptions.xml"))
				return false;
			delete [] s;
			SAFE(read_int32(&tmp32, f));
			if (!add_utvalue(xw, "since", tmp32, "operserv_exceptions.xml"))
				return false;
			SAFE(read_int32(&tmp32, f));
			if (!add_utvalue(xw, "expires", tmp32, "operserv_exceptions.xml"))
				return false;
			
			if (!end_record(xw, "operserv_exceptions.xml"))
				return false;
		}
		
		close_db(f);
		
		if (!save(xw, absolute_path(dir, "operserv_exceptions.xml")))
			return false;
		
		if (!add_index(xi, "operserv_exceptions.xml"))
			return false;
	
		me->core().log(InfoLegacyDB(), "Exported operserv_exceptions");
	}
	
	return true;
}

#undef SAFE

/*************************************************************************/

#define SAFE(x) do {                          \
	if ((x) < 0)                              \
	{                                         \
		errorf("Read error on news.db"); \
        return false;                         \
    }                                         \
} while (0)

bool XMLExport::export_news(std::string const &dir, XMLIndex &xi,
		std::string const &database)
{
	if ((database.empty() || database == "operserv_news") &&
			!xi.exists("operserv_news"))
	{
		XMLWriter xw;
		
		if (!version(xw, 1, "operserv_news.xml"))
			return false;

		if (!add_field(xw, "type", Field::ftUnsigned, 1, 0, "1",
				"operserv_news.xml"))
			return false;
		if (!add_field(xw, "text", Field::ftString, 300, 0, NULL,
				"operserv_news.xml"))
			return false;
		if (!add_field(xw, "creator", Field::ftString, 63, 0, NULL,
				"operserv_news.xml"))
			return false;
		if (!add_field(xw, "since", Field::ftUnixTime, 0, 0, "0",
				"operserv_news.xml"))
			return false;

		dbFILE *f;
		
		if (!(f = open_db("news.db")))
		{
			errorf("Can't open file news.db");
			return false;
		}
		
		int ver;
		
		if ((ver = get_file_version(f)) < 7)
		{
			errorf("Unsupported version (%d) on news.db", ver);
			return false;
		}
		
		uint16_t cnt;
		
		SAFE(read_int16(&cnt, f));
		if (!count(xw, cnt, "operserv_news.xml"))
			return false;
		
		char *s, creator[32];
		uint16_t tmp16;
		uint32_t tmp32;
		
		Record::identifier_type id = 0;
		
		for (int i = 0; i < cnt; i++) 
		{
			if (!begin_record(xw, "operserv_news.xml"))
				return false;
			
			if (!add_rivalue(xw, "id", ++id, "operserv_news.xml"))
				return false;
			
			SAFE(read_int16(&tmp16, f));
			if (!add_uivalue(xw, "type", tmp16 + 1, "operserv_news.xml"))
				return false;
			SAFE(read_int32(&tmp32, f));
			SAFE(read_string(&s, f));
			if (!add_value(xw, "text", s, "operserv_news.xml"))
				return false;
			delete [] s;
			SAFE(read_buffer(creator, f));
			if (!add_value(xw, "creator", creator, "operserv_news.xml"))
				return false;
			SAFE(read_int32(&tmp32, f));
			if (!add_utvalue(xw, "since", tmp32, "operserv_news.xml"))
				return false;
			
			if (!end_record(xw, "operserv_news.xml"))
				return false;
		}
		
		close_db(f);
		
		if (!save(xw, absolute_path(dir, "operserv_news.xml")))
			return false;
		
		if (!add_index(xi, "operserv_news.xml"))
			return false;
	
		me->core().log(InfoLegacyDB(), "Exported operserv_news");
	}
	
	return true;
}

#undef SAFE

/*************************************************************************/

#define SAFE(x) do {                          \
	if ((x) < 0)                              \
	{                                         \
		errorf("Read error on nick.db"); \
        return false;                         \
    }                                         \
} while (0)

bool XMLExport::export_nickserv(std::string const &dir, XMLIndex &xi,
		std::string const &database)
{	
	XMLWriter *xw = 0, *xwa = 0, *xwm = 0;
	bool groups = false, aliases = false, access = false, memos = false;
	
	if ((database.empty() || database == "nickserv_groups") &&
			!xi.exists("nickserv_groups"))
		groups = true;
	if ((database.empty() || database == "nickserv_aliases") &&
			!xi.exists("nickserv_aliases"))
		aliases = true;
	if ((database.empty() || database == "nickserv_access") &&
			!xi.exists("nickserv_access"))
		access = true;
	if ((database.empty() || database == "nickserv_memos") &&
			!xi.exists("nickserv_memos"))
		memos = true;
	
	dbFILE *f;
	
	if (!(f = open_db("nick.db")))
	{
		errorf("Can't open file nick.db");
		return false;
	}
	
	int ver = get_file_version(f);
	
	if (ver < 13)
	{
		errorf("Unsupported version %d on nick.db", ver);
		close_db(f);
		return false;
	}
	
	int c;
	uint16_t tmp16;
	uint32_t tmp32;
	char *s;
	
	Record::identifier_type id;
	id_map::const_iterator it;
	
	Record::identifier_type last_gid = 0;
	
	for (int i = 0; i < 1024; ++i)
	{
		while ((c = getc_db(f)) != 0)
		{
			if (c != 1)
			{
				errorf("Invalid format in nick.db");
				return false;
			}
			
			if (ver >= 20)
				SAFE(read_int32(&id, f));
			else
				id = ++last_gid;
			
			SAFE(read_string(&s, f));
			::groups.insert(make_pair(string(s), id));
			//nc = new NickGroup(s);
			delete [] s;
			
			SAFE(read_string(&s, f));
			//nc->pass(s);
			if (s)
				delete [] s;
			
			if (ver >= 16) 
			{
				SAFE(read_int32(&tmp32, f));
				//nc->created(tmp32);
			}
			
			if (ver >= 17) 
			{
				SAFE(read_string(&s, f));
				//nc->chost(s);
				if (s)
					delete [] s;
				SAFE(read_int32(&tmp32, f));
				//nc->chost_changed(tmp32);
				SAFE(read_string(&s, f));
				//nc->chost_setby(s);
				if (s)
					delete [] s;
			}
			
			SAFE(read_string(&s, f));
			//nc->email(s);
			if (s)
				delete [] s;
			SAFE(read_string(&s, f));
			//nc->greet(s);
			if (s)
				delete [] s;
			SAFE(read_int32(&tmp32, f));
			//nc->icq(tmp32);
			if (ver >= 15 && ver <= 17) 
			{
				SAFE(read_string(&s, f));
				delete [] s;
			}
			SAFE(read_string(&s, f));
			//nc->url(s);
			if (s)
				delete [] s;
			
			SAFE(read_int32(&tmp32, f));
			if (ver <= 18)
				tmp32 &= ~0x00000004; /* Remove old NI_MSG */
			//nc->flags_add(tmp32);

			SAFE(read_int16(&tmp16, f));
			//nc->language(tmp16);
			if (ver >= 14 && ver <= 17)
				SAFE(read_int16(&tmp16, f));
			
			uint16_t accesscount;
			SAFE(read_int16(&accesscount, f));
			//nc->accesscount(tmp16);
			if (accesscount) 
			{
				/*char **access;
				access = static_cast<char **>(smalloc(sizeof(char *) * nc->accesscount()));
				nc->access = access;*/
				for (int j = 0; j < accesscount; j++)
				{
			    	SAFE(read_string(&s, f));
			    	delete [] s;
			    }
			}
			
			uint16_t memocount;
			
			SAFE(read_int16(&memocount, f));
			//nc->memos.memocount(tmp16);
			SAFE(read_int16(&tmp16, f));
			//nc->memos.memomax(tmp16);
			if (memocount)
			{
				char sender[NICKMAX];
				
				for (int j = 0; j < memocount; j++) 
				{
			    	SAFE(read_int32(&tmp32, f));
			    	//memos->number(tmp32);
			    	SAFE(read_int16(&tmp16, f));
			    	//memos->flags_add(tmp16);
			    	SAFE(read_int32(&tmp32, f));
			    	//memos->time(tmp32);
			    	SAFE(read_buffer(sender, f));
			    	//memos->sender(sender);
			    	SAFE(read_string(&s, f));
			    	//memos->text(s);
			    	delete [] s;
				}
			}
			
			SAFE(read_int16(&tmp16, f));
			//nc->channelcount(tmp16);
			SAFE(read_int16(&tmp16, f));
			//nc->channelmax(epona->cd().chanserv.max_reg);
		}
	}
	
	if (aliases)
	{
		xw = new XMLWriter;
		
		if (!version(*xw, 1, "nickserv_aliases.xml"))
			return false;
	
		if (!add_field(*xw, "nick", Field::ftString, 63, 0, "",
				"nickserv_aliases.xml"))
			return false;
		if (!add_field(*xw, "group", Field::ftIdentifier, 0, Field::faOwner,
				"0", "nickserv_aliases.xml"))
			return false;
		if (!add_field(*xw, "created", Field::ftUnixTime, 0, 0, "0",
				"nickserv_aliases.xml"))
			return false;
		if (!add_field(*xw, "last_seen", Field::ftUnixTime, 0, 0, "0",
				"nickserv_aliases.xml"))
			return false;
		if (!add_field(*xw, "last_usermask", Field::ftString, 127, 0, "",
				"nickserv_aliases.xml"))
			return false;
		if (!add_field(*xw, "last_realname", Field::ftString, 127, 0, "",
				"nickserv_aliases.xml"))
			return false;
		if (!add_field(*xw, "last_quit", Field::ftString, 400, 0, "",
				"nickserv_aliases.xml"))
			return false;
		if (!add_field(*xw, "status", Field::ftUnsigned, 2, 0, "0",
				"nickserv_aliases.xml"))
			return false;
	}
	
	size_t acount = 0;
	Record::identifier_type last_aid = 0;
	
	char *name, *reason, *creator;
	time_t since;
	uint16_t flags;
	
	for (int i = 0; i < 1024; ++i)
	{
		while ((c = getc_db(f)) != 0) 
		{
			if (c != 1)
			{
				errorf("Invalid format in nick.db");
				return false;
			}
			
			bool dump = false;
			
			if (aliases && !begin_record(*xw, "nickserv_aliases.xml"))
					return false;
			
			if (ver >= 20)
				SAFE(read_int32(&id, f));
			else
				id = ++last_aid;
			
			if (aliases && !add_rivalue(*xw, "id", id, "nickserv_aliases.xml"))
				return false;
			
			SAFE(read_string(&name, f));
			
			if (aliases && !add_value(*xw, "nick", name, "nickserv_aliases.xml"))
				return false;
			
			SAFE(read_string(&creator, f));
			
			if (aliases && !add_value(*xw, "last_usermask", creator,
					"nickserv_aliases.xml"))
				return false;
			
			SAFE(read_string(&reason, f));
			
			if (aliases && !add_value(*xw, "last_realname", reason,
					"nickserv_aliases.xml"))
				return false;
			
			SAFE(read_string(&s, f));
			
			if (aliases && !add_value(*xw, "last_quit", s,
					"nickserv_aliases.xml"))
				return false;
			
			if (s)
				delete [] s;
			
			SAFE(read_int32(&tmp32, f));
			
			if (aliases && !add_utvalue(*xw, "created", tmp32,
					"nickserv_aliases.xml"))
				return false;
			
			since = tmp32;
			
			SAFE(read_int32(&tmp32, f));
			
			if (aliases && !add_utvalue(*xw, "last_seen", tmp32,
					"nickserv_aliases.xml"))
				return false;
			
			SAFE(read_int16(&flags, f));
			
			if (aliases && !add_uivalue(*xw, "status", (flags & ~NS_UNUSED),
					"nickserv_aliases.xml"))
				return false;
			
			SAFE(read_string(&s, f));
			
			if (aliases)
			{
				it = ::groups.find(s);
				if (it == ::groups.end())
				{
					me->core().log(InfoLegacyDB(),
							"Ignoring nickname %s with invalid group %s",
							name, s);
					dump = true;
				}
				else
				{
					if (aliases && !add_rivalue(*xw, "group", it->second,
							"nickserv_aliases.xml"))
						return false;
				}
			}
			
			delete [] s;
			
			if (flags & NS_VERBOTEN)
			{	
				Forbid *f = new Forbid;
				f->name(name);
				f->creator(creator ? creator : "Unknown");
				f->reason(reason ? reason : "Unknown reason");
				f->since(since ? since : time(0));
				forbids.push_back(f);
				
				/* Which means: FORBID, ignore the group. */
				::aliases.insert(make_pair(string(name), 0));
			
				me->core().log(InfoLegacyDB(),
						"Ignoring forbidden nick %s (ID #" RECORD_FORMAT ")",
						name, id);
				dump = true;
			} else {
				::aliases.insert(make_pair(string(name), id));
			}
			
			delete [] name;
			if (creator)
				delete [] creator;
			if (reason)
				delete [] reason;
			
			if (aliases && dump) {
				xw->dump_record();
			} else {
				if (aliases && !end_record(*xw, "nickserv_aliases.xml"))
					return false;		
				++acount;
			}
		}
	}
	
	if (aliases)
	{	
		if (!count(*xw, acount, "nickserv_aliases.xml"))
			return false;
		
		if (!save(*xw, absolute_path(dir, "nickserv_aliases.xml")))
			return false;
		
		if (!add_index(xi, "nickserv_aliases.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported nickserv_aliases");
		delete xw;
	}
	
	close_db(f);
	
	if (groups || access || memos)
	{
		char *pass;	
		size_t gcount = 0, acount = 0, mcount = 0;
		
		last_gid = 0;
		Record::identifier_type last_aid = 0, last_mid = 0;
		
		if (groups)
		{
			xw = new XMLWriter;
			
			if (!version(*xw, 1, "nickserv_groups.xml"))
				return false;
		
			if (!add_field(*xw, "name", Field::ftIdentifier, 0, 0, "0",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "enc", Field::ftString, 255, 0, "",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "pass", Field::ftString, 192, Field::faCase,
					"", "nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "created", Field::ftUnixTime, 0, 0, "0",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "flags", Field::ftUnsigned, 4, 0, "2147483648",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "email", Field::ftString, 127, 0, "",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "url", Field::ftString, 255, 0, "",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "im_type", Field::ftString, 31, 0, "",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "im_id", Field::ftString, 127, 0, "",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "greet", Field::ftString, 300, 0, "",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "language", Field::ftUnsigned, 2, 0, "65535",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "chost", Field::ftString, 63, 0, "",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "chost_changed", Field::ftUnixTime, 0, 0, "0",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "chost_setby", Field::ftString, 63, 0, "",
					"nickserv_groups.xml"))
				return false;
			if (!add_field(*xw, "maxmemos", Field::ftInteger, 2, 0, "-2",
					"nickserv_groups.xml"))
				return false;
		}
		
		if (access)
		{
			xwa = new XMLWriter;
			
			if (!version(*xwa, 1, "nickserv_access.xml"))
				return false;
		
			if (!add_field(*xwa, "owner", Field::ftIdentifier, 0,
					Field::faOwner, "0", "nickserv_access.xml"))
				return false;
			if (!add_field(*xwa, "mask", Field::ftString, 127, 0, "",
					"nickserv_access.xml"))
				return false;
		}
		
		if (memos)
		{
			xwm = new XMLWriter;
			
			if (!version(*xwm, 1, "nickserv_memos.xml"))
				return false;
		
			if (!add_field(*xwm, "owner", Field::ftIdentifier, 0,
					Field::faOwner, "0", "nickserv_memos.xml"))
				return false;
			if (!add_field(*xwm, "number", Field::ftUnsigned, 4, 0, "0",
					"nickserv_memos.xml"))
				return false;
			if (!add_field(*xwm, "flags", Field::ftUnsigned, 2, 0, "1",
					"nickserv_memos.xml"))
				return false;
			if (!add_field(*xwm, "sender", Field::ftString, 63, 0, "",
					"nickserv_memos.xml"))
				return false;
			if (!add_field(*xwm, "target", Field::ftString, 63, 0, "",
					"nickserv_memos.xml"))
				return false;
			if (!add_field(*xwm, "created", Field::ftUnixTime, 0, 0, "0",
					"nickserv_memos.xml"))
				return false;
			if (!add_field(*xwm, "text", Field::ftString, 4096, 0, "",
					"nickserv_memos.xml"))
				return false;
		}
		
		
		if (!(f = open_db("nick.db")))
		{
			errorf("Can't open file nick.db");
			return false;
		}
		
		ver = get_file_version(f);
		
		if (ver < 13)
		{
			errorf("Unsupported version %d on nick.db", ver);
			close_db(f);
			return false;
		}
		
		for (int i = 0; i < 1024; ++i)
		{
			while ((c = getc_db(f)) != 0)
			{
				if (c != 1)
				{
					errorf("Invalid format in nick.db");
					return false;
				}
				
				bool dump = false;
				
				if (groups && !begin_record(*xw, "nickserv_groups.xml"))
					return false;
				
				if (ver >= 20)
					SAFE(read_int32(&id, f));
				else
					id = ++last_gid;
				
				if (groups && !add_rivalue(*xw, "id", id, "nickserv_groups.xml"))
					return false;
				
				SAFE(read_string(&s, f));
				
				if (groups)
				{
					it = ::aliases.find(s);
					if (it == ::aliases.end())
					{
						if (groups && !add_rivalue(*xw, "name", 0,
								"nickserv_groups.xml"))
							return false;
					}
					else
					{
						if (it->second == 0)
						{
							me->core().log(InfoLegacyDB(),
									"Ignoring forbidden group %s (ID #"
									RECORD_FORMAT ")", s, id);
							dump = true;
						} else {
							if (groups && !add_rivalue(*xw, "name", it->second,
									"nickserv_groups.xml"))
								return false;
						}
					}
				}
				
				delete [] s;
				
				SAFE(read_string(&pass, f));
				
				if (ver >= 16) 
				{
					SAFE(read_int32(&tmp32, f));
					if (groups && !add_utvalue(*xw, "created", tmp32,
							"nickserv_groups.xml"))
						return false;
				} else {
					if (groups && !add_utvalue(*xw, "created", time(0),
							"nickserv_groups.xml"))
						return false;
				}
				
				if (ver >= 17) 
				{
					SAFE(read_string(&s, f));
					if (groups && !add_value(*xw, "chost", s,
							"nickserv_groups.xml"))
						return false;
					if (s)
						delete [] s;
					SAFE(read_int32(&tmp32, f));
					if (groups && !add_utvalue(*xw, "chost_changed", tmp32,
							"nickserv_groups.xml"))
						return false;
					SAFE(read_string(&s, f));
					if (groups && !add_value(*xw, "chost_setby", s,
							"nickserv_groups.xml"))
						return false;
					if (s)
						delete [] s;
				} else {
					if (groups && !add_value(*xw, "chost", 0,
							"nickserv_groups.xml"))
						return false;
					if (groups && !add_utvalue(*xw, "chost_changed", 0,
							"nickserv_groups.xml"))
						return false;
					if (groups && !add_value(*xw, "chost_setby", 0,
							"nickserv_groups.xml"))
						return false;
				}
					
				SAFE(read_string(&s, f));
				if (groups && !add_value(*xw, "email", s,
						"nickserv_groups.xml"))
					return false;
				if (s)
					delete [] s;
				SAFE(read_string(&s, f));
				if (groups && !add_value(*xw, "greet", s,
						"nickserv_groups.xml"))
					return false;
				if (s)
					delete [] s;
				SAFE(read_int32(&tmp32, f));
				if (tmp32)
				{
					char imid[20];
					snprintf(imid, sizeof(imid), "%u", tmp32);
					
					if (groups && !add_value(*xw, "im_type", "ICQ",
							"nickserv_groups.xml"))
						return false;
					if (groups && !add_value(*xw, "im_id", imid,
							"nickserv_groups.xml"))
						return false;
				} else {
					if (groups && !add_value(*xw, "im_type", 0,
							"nickserv_groups.xml"))
						return false;
					if (groups && !add_value(*xw, "im_id", 0,
							"nickserv_groups.xml"))
						return false;
				}
				if (ver >= 15 && ver <= 17) 
				{
					SAFE(read_string(&s, f));
					delete [] s;
				}
				SAFE(read_string(&s, f));
				if (groups && !add_value(*xw, "url", s, "nickserv_groups.xml"))
					return false;
				if (s)
					delete [] s;
				
				SAFE(read_int32(&tmp32, f));
				
				if (pass && (tmp32 & NI_ENCRYPTEDPW))
				{
					if (groups && !add_value(*xw, "enc", "legacydb",
							"nickserv_groups.xml"))
						return false;
					
					char truepass[33], *end = truepass;
					for (int k = 0; k < 16; ++k)
						end += sprintf(end, "%02x", (unsigned char)pass[k]);
					truepass[32] = 0;
					
					if (groups && !add_value(*xw, "pass", truepass,
							"nickserv_groups.xml"))
						return false;
					
					delete [] pass;
					pass = 0;
				} else if (pass)
				{
					if (groups && !add_value(*xw, "enc", 0,
							"nickserv_groups.xml"))
						return false;
					
					if (groups && !add_value(*xw, "pass", pass,
							"nickserv_groups.xml"))
						return false;
					
					delete [] pass;
					pass = 0;
				}
				
				tmp32 &= ~NI_UNUSED;
				
				if (groups && !add_uivalue(*xw, "flags", tmp32,
						"nickserv_groups.xml"))
					return false;
	
				SAFE(read_int16(&tmp16, f));
				if (groups && !add_uivalue(*xw, "language", tmp16,
						"nickserv_groups.xml"))
					return false;
				if (ver >= 14 && ver <= 17)
					SAFE(read_int16(&tmp16, f));
				
				uint16_t accesscount;
				SAFE(read_int16(&accesscount, f));
				//nc->accesscount(tmp16);
				if (accesscount) 
				{
					/*char **access;
					access = static_cast<char **>(smalloc(sizeof(char *) * nc->accesscount()));
					nc->access = access;*/
					for (int j = 0; j < accesscount; j++)
					{
						if (access && !dump &&
								!begin_record(*xwa, "nickserv_access.xml"))
							return false;
						
						if (access && !dump &&
								!add_rivalue(*xwa, "id", ++last_aid,
								"nickserv_access.xml"))
							return false;
						
						if (access && !dump &&
								!add_rivalue(*xwa, "owner", id,
								"nickserv_access.xml"))
							return false;
						
				    	SAFE(read_string(&s, f));
				    	
				    	if (access && !dump &&
								!add_value(*xwa, "mask", s,
								"nickserv_access.xml"))
							return false;
						
				    	delete [] s;
				    	
				    	if (access && !dump)
				    	{
				    		if (!end_record(*xwa, "nickserv_access.xml"))
								return false;
							++acount;
						}
				    }
				}
				
				uint16_t memocount;
				
				SAFE(read_int16(&memocount, f));
				//nc->memos.memocount(tmp16);
				SAFE(read_int16(&tmp16, f));
				//nc->memos.memomax(tmp16);
				int16_t memomax = tmp16;
				if (groups && !add_ivalue(*xw, "maxmemos", memomax,
						"nickserv_groups.xml"))
					return false;
				
				if (memocount)
				{
					char sender[NICKMAX];
					
					for (int j = 0; j < memocount; j++) 
					{
						if (memos && !dump &&
								!begin_record(*xwm, "nickserv_memos.xml"))
							return false;
						
						if (memos && !dump &&
								!add_rivalue(*xwm, "id", ++last_mid,
								"nickserv_memos.xml"))
							return false;
						
						if (memos && !dump &&
								!add_rivalue(*xwm, "owner", id,
								"nickserv_memos.xml"))
							return false;
						
				    	SAFE(read_int32(&tmp32, f));
				    	if (memos && !dump &&
								!add_uivalue(*xwm, "number", tmp32,
								"nickserv_memos.xml"))
							return false;
				    	SAFE(read_int16(&tmp16, f));
				    	if (memos && !dump &&
								!add_uivalue(*xwm, "flags", tmp16,
								"nickserv_memos.xml"))
							return false;
				    	SAFE(read_int32(&tmp32, f));
				    	if (memos && !dump &&
								!add_utvalue(*xwm, "created", tmp32,
								"nickserv_memos.xml"))
							return false;
				    	SAFE(read_buffer(sender, f));
				    	//memos->sender(sender);
				    	if (memos && !dump &&
								!add_value(*xwm, "sender", sender,
								"nickserv_memos.xml"))
							return false;
						if (memos && !dump &&
								!add_value(*xwm, "target", 0,
								"nickserv_memos.xml"))
							return false;
				    	SAFE(read_string(&s, f));
				    	if (memos && !dump &&
								!add_value(*xwm, "text", s,
								"nickserv_memos.xml"))
							return false;
				    	delete [] s;
				    	
				    	if (memos && !dump)
				    	{
				    		if (!end_record(*xwm, "nickserv_memos.xml"))
								return false;
							++mcount;
						}
					}
				}
				
				SAFE(read_int16(&tmp16, f));
				//nc->channelcount(tmp16);
				SAFE(read_int16(&tmp16, f));
				//nc->channelmax(epona->cd().chanserv.max_reg);
				
				if (groups && dump)
					xw->dump_record();
				else
				{
					if (groups && !end_record(*xw, "nickserv_groups.xml"))
						return false;		
					++gcount;
				}
			}
		}
		
		if (groups)
		{	
			if (!count(*xw, gcount, "nickserv_groups.xml"))
				return false;
			
			if (!save(*xw, absolute_path(dir, "nickserv_groups.xml")))
				return false;
			
			if (!add_index(xi, "nickserv_groups.xml"))
				return false;
			
			me->core().log(InfoLegacyDB(), "Exported nickserv_groups");
			delete xw;
		}
		
		if (access)
		{	
			if (!count(*xwa, acount, "nickserv_access.xml"))
				return false;
			
			if (!save(*xwa, absolute_path(dir, "nickserv_access.xml")))
				return false;
			
			if (!add_index(xi, "nickserv_access.xml"))
				return false;
			
			me->core().log(InfoLegacyDB(), "Exported nickserv_access");
			delete xwa;
		}
		
		if (memos)
		{	
			if (!count(*xwm, mcount, "nickserv_memos.xml"))
				return false;
			
			if (!save(*xwm, absolute_path(dir, "nickserv_memos.xml")))
				return false;
			
			if (!add_index(xi, "nickserv_memos.xml"))
				return false;
			
			me->core().log(InfoLegacyDB(), "Exported nickserv_memos");
			delete xwm;
		}
		
		close_db(f);
	}
	
	return true;
}

#undef SAFE

/*************************************************************************/

#define SAFE(x) do {                          \
	if ((x) < 0)                              \
	{                                         \
		errorf("Read error on oper.db"); \
        return false;                         \
    }                                         \
} while (0)

bool XMLExport::export_operserv(std::string const &dir, XMLIndex &xi,
		std::string const &database)
{
	bool stats = false, akills = false, sglines = false, sqlines = false,
			szlines = false;
	
	if ((database.empty() || database == "operserv_stats") &&
			!xi.exists("operserv_stats"))
		stats = true;
	if ((database.empty() || database == "operserv_akills") &&
			!xi.exists("operserv_akills"))
		akills = true;
	if ((database.empty() || database == "operserv_sglines") &&
			!xi.exists("operserv_sglines"))
		sglines = true;
	if ((database.empty() || database == "operserv_sqlines") &&
			!xi.exists("operserv_sqlines"))
		sqlines = true;
	if ((database.empty() || database == "operserv_szlines") &&
			!xi.exists("operserv_szlines"))
		szlines = true;
	
	dbFILE *f;
	
	if (!(f = open_db("oper.db")))
	{
		errorf("Can't open file oper.db");
		return false;
	}
	
	int ver;
	
	if ((ver = get_file_version(f)) < 7)
	{
		errorf("Unsupported version (%d) on oper.db", ver);
		return false;
	}
	
	XMLWriter *xw = 0;
	
	if (stats)
	{
		xw = new XMLWriter;
		
		if (!version(*xw, 1, "operserv_stats.xml"))
			return false;
	
		if (!add_field(*xw, "maxusers", Field::ftUnsigned, 4, 0, "0",
				"operserv_stats.xml"))
			return false;
		if (!add_field(*xw, "maxusers_time", Field::ftUnixTime, 0, 0, "0",
				"operserv_stats.xml"))
			return false;
	}
	
	if (stats)
	{
		if (!count(*xw, 1, "operserv_stats.xml"))
			return false;
		
		if (!begin_record(*xw, "operserv_stats.xml"))
			return false;
	
		if (!add_rivalue(*xw, "id", 1, "operserv_stats.xml"))
			return false;
	}
	
	uint16_t cnt, i, tmp16;
	uint32_t tmp32;
	char *s;
	
	SAFE(read_int32(&tmp32, f)); /* maxusercnt */
	if (stats && !add_uivalue(*xw, "maxusers", tmp32, "operserv_stats.xml"))
		return false;
	SAFE(read_int32(&tmp32, f));
	if (stats && !add_utvalue(*xw, "maxusers_time", tmp32, "operserv_stats.xml"))
		return false;
		
	if (stats)
	{
		if (!end_record(*xw, "operserv_stats.xml"))
			return false;
	}
	
	if (stats)
	{	
		if (!save(*xw, absolute_path(dir, "operserv_stats.xml")))
			return false;
		
		if (!add_index(xi, "operserv_stats.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported operserv_stats");
		
		delete xw;
	}
	
	if (ver == 14) {
		uint16_t mcount;
		
		SAFE(read_int16(&mcount, f));
		
		for (i = 0; i < mcount; i++) {
			SAFE(read_int16(&tmp16, f));
			SAFE(read_int16(&tmp16, f));
		}
	}
	
	if (akills)
	{
		xw = new XMLWriter;
		
		if (!version(*xw, 1, "operserv_akills.xml"))
			return false;
	
		if (!add_field(*xw, "mask", Field::ftString, 127, 0, "",
				"operserv_akills.xml"))
			return false;
		if (!add_field(*xw, "creator", Field::ftString, 63, 0, "",
				"operserv_akills.xml"))
			return false;
		if (!add_field(*xw, "reason", Field::ftString, 255, 0, "",
				"operserv_akills.xml"))
			return false;
		if (!add_field(*xw, "since", Field::ftUnixTime, 0, 0, "0",
				"operserv_akills.xml"))
			return false;
		if (!add_field(*xw, "expires", Field::ftUnixTime, 0, 0, "0",
				"operserv_akills.xml"))
			return false;
	}
	
	char *s2;
	
	read_int16(&tmp16, f);
	cnt = tmp16;
	
	if (akills)
	{
		if (!count(*xw, cnt, "operserv_akills.xml"))
			return false;
	}
	
	for (i = 0; i < cnt; i++)
	{
		if (akills)
		{
			if (!begin_record(*xw, "operserv_akills.xml"))
				return false;
	
			if (!add_rivalue(*xw, "id", i + 1, "operserv_akills.xml"))
				return false;
		}
		
		SAFE(read_string(&s, f));
		SAFE(read_string(&s2, f));
		
		if (akills)
		{
			size_t len = strlen(s) + strlen(s2) + 2;
			misc::cstring mask;
			mask.reserve(len);
			snprintf(mask, len, "%s@%s", s, s2);
		
			if (!add_value(*xw, "mask", mask, "operserv_akills.xml"))
				return false;
		}
		
		delete [] s;
		delete [] s2;
		
		SAFE(read_string(&s, f));
		if (akills && !add_value(*xw, "creator", s, "operserv_akills.xml"))
			return false;
		delete [] s;
		SAFE(read_string(&s, f));
		if (akills && !add_value(*xw, "reason", (s ? s : "No reason"),
				"operserv_akills.xml"))
			return false;
		if (s)
			delete [] s;
		SAFE(read_int32(&tmp32, f));
		if (akills && !add_utvalue(*xw, "since", tmp32, "operserv_akills.xml"))
			return false;
		SAFE(read_int32(&tmp32, f));
		if (akills && !add_utvalue(*xw, "expires", tmp32,
				"operserv_akills.xml"))
			return false;
		
		if (akills)
		{
			if (!end_record(*xw, "operserv_akills.xml"))
				return false;
		}
	}
	
	if (akills)
	{	
		if (!save(*xw, absolute_path(dir, "operserv_akills.xml")))
			return false;
		
		if (!add_index(xi, "operserv_akills.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported operserv_akills");
		
		delete xw;
	}
	
	/* SGLINES */
	
	if (sglines)
	{
		xw = new XMLWriter;
		
		if (!version(*xw, 1, "operserv_sglines.xml"))
			return false;
	
		if (!add_field(*xw, "mask", Field::ftString, 127, 0, "",
				"operserv_sglines.xml"))
			return false;
		if (!add_field(*xw, "creator", Field::ftString, 63, 0, "",
				"operserv_sglines.xml"))
			return false;
		if (!add_field(*xw, "reason", Field::ftString, 255, 0, "",
				"operserv_sglines.xml"))
			return false;
		if (!add_field(*xw, "since", Field::ftUnixTime, 0, 0, "0",
				"operserv_sglines.xml"))
			return false;
		if (!add_field(*xw, "expires", Field::ftUnixTime, 0, 0, "0",
				"operserv_sglines.xml"))
			return false;
	}
	
	read_int16(&tmp16, f);
	cnt = tmp16;
	
	if (sglines)
	{
		if (!count(*xw, cnt, "operserv_sglines.xml"))
			return false;
	}
	
	for (i = 0; i < cnt; i++) 
	{
		if (sglines)
		{
			if (!begin_record(*xw, "operserv_sglines.xml"))
				return false;
	
			if (!add_rivalue(*xw, "id", i + 1, "operserv_sglines.xml"))
				return false;
		}
		
		SAFE(read_string(&s, f));
		if (sglines && !add_value(*xw, "mask", s, "operserv_sglines.xml"))
				return false;
		delete [] s;
		SAFE(read_string(&s, f));
		if (sglines && !add_value(*xw, "creator", s, "operserv_sglines.xml"))
			return false;
		delete [] s;
		SAFE(read_string(&s, f));
		if (sglines && !add_value(*xw, "reason", s, "operserv_sglines.xml"))
			return false;
		delete [] s;
		SAFE(read_int32(&tmp32, f));
		if (sglines && !add_utvalue(*xw, "since", tmp32,
				"operserv_sglines.xml"))
			return false;
		SAFE(read_int32(&tmp32, f));
		if (sglines && !add_utvalue(*xw, "expires", tmp32,
				"operserv_sglines.xml"))
			return false;
		
		if (sglines)
		{
			if (!end_record(*xw, "operserv_sglines.xml"))
				return false;
		}
	}
	
	if (sglines)
	{	
		if (!save(*xw, absolute_path(dir, "operserv_sglines.xml")))
			return false;
		
		if (!add_index(xi, "operserv_sglines.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported operserv_sglines");
		delete xw;
	}

	/* SQLINES */
	
	if (sqlines)
	{
		xw = new XMLWriter;
		
		if (!version(*xw, 1, "operserv_sqlines.xml"))
			return false;
	
		if (!add_field(*xw, "mask", Field::ftString, 127, 0, "",
				"operserv_sqlines.xml"))
			return false;
		if (!add_field(*xw, "creator", Field::ftString, 63, 0, "",
				"operserv_sqlines.xml"))
			return false;
		if (!add_field(*xw, "reason", Field::ftString, 255, 0, "",
				"operserv_sqlines.xml"))
			return false;
		if (!add_field(*xw, "since", Field::ftUnixTime, 0, 0, "0",
				"operserv_sqlines.xml"))
			return false;
		if (!add_field(*xw, "expires", Field::ftUnixTime, 0, 0, "0",
				"operserv_sqlines.xml"))
			return false;
	}
	
	read_int16(&tmp16, f);
	cnt = tmp16;
	
	if (sqlines)
	{
		if (!count(*xw, cnt + forbids.size(), "operserv_sqlines.xml"))
			return false;
	}
	
	int sqid = 0;

	for (i = 0; i < cnt; i++) 
	{
		if (sqlines)
		{
			if (!begin_record(*xw, "operserv_sqlines.xml"))
				return false;
	
			if (!add_rivalue(*xw, "id", ++sqid, "operserv_sqlines.xml"))
				return false;
		}
		
		SAFE(read_string(&s, f));
		if (sqlines && !add_value(*xw, "mask", s, "operserv_sqlines.xml"))
				return false;
		delete [] s;
		SAFE(read_string(&s, f));
		if (sqlines && !add_value(*xw, "creator", s, "operserv_sqlines.xml"))
			return false;
		delete [] s;
		SAFE(read_string(&s, f));
		if (sqlines && !add_value(*xw, "reason", s, "operserv_sqlines.xml"))
			return false;
		delete [] s;
		SAFE(read_int32(&tmp32, f));
		if (sqlines && !add_utvalue(*xw, "since", tmp32, "operserv_sqlines.xml"))
			return false;
		SAFE(read_int32(&tmp32, f));
		if (sqlines && !add_utvalue(*xw, "expires", tmp32,
				"operserv_sqlines.xml"))
			return false;
		
		if (sqlines)
		{
			if (!end_record(*xw, "operserv_sqlines.xml"))
				return false;
		}
	}
	
	if (sqlines)
	{
		vector<Forbid *>::const_iterator it, end;
		
		for (it = forbids.begin(), end = forbids.end(); it != end; ++it) 
		{
			Forbid *f = (*it);
			
			if (!begin_record(*xw, "operserv_sqlines.xml"))
				return false;
				
			if (!add_rivalue(*xw, "id", ++sqid, "operserv_sqlines.xml"))
				return false;
			
			if (!add_value(*xw, "mask", f->name(), "operserv_sqlines.xml"))
				return false;
			if (!add_value(*xw, "creator", f->creator(),
					"operserv_sqlines.xml"))
				return false;
			if (!add_value(*xw, "reason", f->reason(), "operserv_sqlines.xml"))
				return false;
			if (!add_utvalue(*xw, "since", f->since(), "operserv_sqlines.xml"))
				return false;
			if (!add_value(*xw, "expires", "0", "operserv_sqlines.xml"))
				return false;
			
			if (!end_record(*xw, "operserv_sqlines.xml"))
				return false;
		}
		
		if (!forbids.empty())
			me->core().log(InfoLegacyDB(), "Converted %u previously forbidden "
					"nicknames and channels into sqlines", forbids.size());
	}
	
	if (sqlines)
	{	
		if (!save(*xw, absolute_path(dir, "operserv_sqlines.xml")))
			return false;
		
		if (!add_index(xi, "operserv_sqlines.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported operserv_sqlines");
		delete xw;
	}
	
	/* SZLINES */
	
	if (szlines)
	{
		xw = new XMLWriter;
		
		if (!version(*xw, 1, "operserv_szlines.xml"))
			return false;
	
		if (!add_field(*xw, "mask", Field::ftString, 127, 0, "",
				"operserv_szlines.xml"))
			return false;
		if (!add_field(*xw, "creator", Field::ftString, 63, 0, "",
				"operserv_szlines.xml"))
			return false;
		if (!add_field(*xw, "reason", Field::ftString, 255, 0, "",
				"operserv_szlines.xml"))
			return false;
		if (!add_field(*xw, "since", Field::ftUnixTime, 0, 0, "0",
				"operserv_szlines.xml"))
			return false;
		if (!add_field(*xw, "expires", Field::ftUnixTime, 0, 0, "0",
				"operserv_szlines.xml"))
			return false;
	}
	
	read_int16(&tmp16, f);
	cnt = tmp16;
	
	if (szlines)
	{
		if (!count(*xw, cnt, "operserv_szlines.xml"))
			return false;
	}
	
	for (i = 0; i < cnt; i++) 
	{
		if (szlines)
		{
			if (!begin_record(*xw, "operserv_szlines.xml"))
				return false;
	
			if (!add_rivalue(*xw, "id", i + 1, "operserv_szlines.xml"))
				return false;
		}
		
		SAFE(read_string(&s, f));
		if (szlines && !add_value(*xw, "mask", s, "operserv_szlines.xml"))
				return false;
		delete [] s;
		SAFE(read_string(&s, f));
		if (szlines && !add_value(*xw, "creator", s, "operserv_szlines.xml"))
			return false;
		delete [] s;
		SAFE(read_string(&s, f));
		if (szlines && !add_value(*xw, "reason", s, "operserv_szlines.xml"))
			return false;
		delete [] s;
		SAFE(read_int32(&tmp32, f));
		if (szlines && !add_utvalue(*xw, "since", tmp32,
				"operserv_szlines.xml"))
			return false;
		SAFE(read_int32(&tmp32, f));
		if (szlines && !add_utvalue(*xw, "expires", tmp32,
				"operserv_szlines.xml"))
			return false;
		
		if (szlines)
		{
			if (!end_record(*xw, "operserv_szlines.xml"))
				return false;
		}
	}
	
	if (szlines)
	{	
		if (!save(*xw, absolute_path(dir, "operserv_szlines.xml")))
			return false;
		
		if (!add_index(xi, "operserv_szlines.xml"))
			return false;
		
		me->core().log(InfoLegacyDB(), "Exported operserv_szlines");
		delete xw;
	}
	
	close_db(f);
	
	return true;
}

#undef SAFE

/*************************************************************************/

bool XMLExport::save(XMLWriter &xw, std::string const &filename)
{
	if (!xw.save(filename))
	{
		errorf("Can't save XML file %s: %s", filename.c_str(), xw.error());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool XMLExport::version(XMLWriter &xw, uint16_t version,
				char const *filename)
{
	if (!xw.database(version))
	{
		errorf("Can't set database information in file %s", filename);
		return false;
	}
	
	return true;
}

/*************************************************************************/

} /* namespace legacydb */

/*************************************************************************/
