/* Log file.
 *
 * Log file module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <filecomp/filecomp.h>
#include <log/conf.h>
#include <misc/misc.h>
#include <modules/core.h>
#include <modules/manager.h>
#include <modules/module.h>

#include "conf.h"
#include "decl.h"
#include "file.h"
#include "module.h"

using namespace std;

using filecomp::Compressor;
using misc::absolute_path;
using modules::Module;
using xlog::LogRules;

/*************************************************************************/

static char const *absolute(char const *filename)
{
	static char path[PATH_MAX];
	absolute_path(path, sizeof(path) - 1, logfile::me->cd().directory,
			filename);
	return path;
}

/*************************************************************************/

namespace logfile
{

/*************************************************************************/

File::File(hidden::Directives::File const &directive) :
		m_directive(directive), m_stream(), m_rotator(0), m_task(0),
		m_compressor(0), m_marked(true), m_ready(false)
{
	/* Sets up file compressor */
	setup_compressor();
	
	/* Sets up log rotation manager */
	setup_rotator();
	
	/* Sets up rotation task */
	setup_task();
	
	/* Rotate the file if span is 0 */
	if (m_directive.rotate && (m_directive.rotate.span == 0))
		if (misc::file_exists(absolute(m_directive.filename)))
			rotate();
	
	/* Open the log file */
	open();
}

/*************************************************************************/

File::~File()
{
	close();
	cleanup_task();
	cleanup_rotator();
	cleanup_compressor();
}

/*************************************************************************/

void File::cleanup_compressor()
{
	if (m_compressor)
	{
		delete m_compressor;
		m_compressor = 0;
	}
}

/*************************************************************************/

void File::cleanup_rotator()
{
	if (m_rotator)
	{
		delete m_rotator;
		m_rotator = 0;
	}
}

/*************************************************************************/

void File::cleanup_task()
{
	if (m_task)
	{
		if (!me->core().tasks().remove(*m_task))
			me->core().fatalf("Can't remove task %s", m_task->name().c_str());
		
		delete m_task;
		m_task = 0;
	}
}

/*************************************************************************/

void File::close()
{
	if (!m_ready)
		return;
	
	if (m_stream)
		log(LogFileInfo(), "--- Log session stopped\n");
	
	m_stream.close();
	m_ready = false;
}

/*************************************************************************/

hidden::Directives::File const
				&File::directive(hidden::Directives::File const &directive)
{
	m_directive = directive;
	
	/* Apply changes to the file compressor */
	setup_compressor();
	
	/* Apply changes to the log rotation manager. */
	setup_rotator();
	
	/* Apply changes to the rotation task. */
	setup_task();
	
	/* Try to reopen the file in case it isn't */
	open();
	
	return m_directive;
}

/*************************************************************************/

void File::log(xlog::LogLevel const &level, char const *message)
{
	if (!m_ready)
		return;
	
	/* Always log the session start/stop message. */
	if (!(level == "Info::Logfile"))
	{
		LogRules::Rule const *rule = m_directive.rules.match(level);
		if (!rule || (rule->type() == LogRules::lrReject))
			return;
	}
	
	if (!m_directive.timestamp_format.value().empty())
	{
		time_t t = time(0);
		struct tm tm = *localtime(&t);
		
		char timebuf[timestamp_length];
		strftime(timebuf, sizeof(timebuf), m_directive.timestamp_format, &tm);
		
		if (*timebuf)
			if (!m_stream.write(timebuf, strlen(timebuf)) || !m_stream.put(' '))
				goto write_error;
	}
	
	if (m_directive.header && *(level.header()) &&
			!m_stream.write(level.header(), strlen(level.header())))
		goto write_error;
	
	if (*message && !m_stream.write(message, strlen(message)))
		goto write_error;
	
	if (!m_stream.put('\n') || !m_stream.flush())
		goto write_error;
	
	return;

write_error:
	close();
	me->core().log_perror(LogFileCritical(), "Write error to %s",
			m_directive.filename.value().c_str());
	return;
}

/*************************************************************************/

bool File::open()
{
	if (m_ready)
		return true;
	
	m_stream.clear();
	m_stream.open(absolute(m_directive.filename), ios::out | ios::app);
	
	if (!m_stream)
	{
		me->core().log_perror(LogFileCritical(), "Can't open file %s",
				m_directive.filename.value().c_str());
		return false;
	}
	
	m_ready = true;
	
	log(LogFileInfo(), "+++ Log session started");
	
	return true;
}

/*************************************************************************/

bool File::rotate(bool manual)
{
	close();
	
	/* If rotate() is called for a file that is not configured for
	 * rotation, delete it and reopen it. Do that
	 * as well when the number of files to keep is zero. */
	if (!m_directive.rotate || (m_directive.rotate.keep == 0))
	{
		if (remove(absolute(m_directive.filename)) == -1 && errno != ENOENT)
		{
			int save_errno = errno;
			
			open();
			
			errno = save_errno;
			me->core().log_perror(LogFileError(),
					"Can't delete file %s for log rotation",
					m_directive.filename.value().c_str());
			return false;
		}
	} else {
		char const *fp = "", *sp = "";
		rotation_suffix(&fp, &sp, manual);
		
		char tbuf_fp[timestamp_length], tbuf_sp[timestamp_length];
		
		time_t t = time(0), bt = (m_directive.rotate ?
				(t - m_directive.rotate.span) : t);
		
		struct tm tm = *localtime(&bt);
		strftime(tbuf_fp, sizeof(tbuf_fp), fp, &tm);
		
		tm = *localtime(&t);
		strftime(tbuf_sp, sizeof(tbuf_sp), sp, &tm);
		
		char newname[PATH_MAX], *end = newname;
		end += snprintf(newname, sizeof(newname), "%s%s%s",
				m_directive.filename.value().c_str(), tbuf_fp, tbuf_sp);
		
		char newpath[PATH_MAX];
		absolute_path(newpath, sizeof(newpath) - 1, me->cd().directory,
				newname);
		
		if (rename(absolute(m_directive.filename), newpath) == -1)
		{
			int save_errno = errno;
			
			open();
			
			errno = save_errno;
			me->core().log_perror(LogFileError(),
					"Can't rename file %s to %s for log rotation",
					m_directive.filename.value().c_str(), newname);
			return false;
		}
		
		if (m_compressor)
		{
			if (!m_compressor->compress(newpath,
					m_directive.rotate.compress.level))
			{
				me->core().log(LogFileError(),
						"Can't compress file %s after log rotation: %s",
						m_directive.filename.value().c_str(),
						m_compressor->error());
			} else 
			{
				if (remove(newpath) == -1)
				{
					me->core().log_perror(LogFileError(),
							"Can't remove file %s after log compression",
							m_directive.filename.value().c_str());
				}
			
				snprintf(end, sizeof(newname)-(end-newname), "%s",
						m_compressor->extension());
			}
		}
		
		if (m_rotator)
			if (!m_rotator->add(newname))
				cleanup_rotator();
	}
	
	open();	
	return true;
}

/*************************************************************************/

void File::rotation_suffix(char const **fp, char const **sp, bool manual)
{
	if (!m_directive.rotate) {
		*fp = "";
		*sp = ".%Y%m%d~%H%M%S";
	} else if (m_directive.rotate.suffix) {
		*fp = m_directive.rotate.suffix.first_part;
		*sp = m_directive.rotate.suffix.second_part;
	} else if (manual || (m_directive.rotate.span == 0)) {
		*fp = "";
		*sp = ".%Y%m%d~%H%M%S";
	} else if (m_directive.rotate.span < 60) {
		*fp = ".%Y%m%d~%H%M%S";
		*sp = "-%Y%m%d~%H%M%S";
	} else if (m_directive.rotate.span == 60) {
		*fp = ".%Y%m%d~%H%M";
		*sp = "";
	} else if (m_directive.rotate.span < 3600) {
		*fp = ".%Y%m%d~%H%M";
		*sp = "-%Y%m%d~%H%M";
	} else if (m_directive.rotate.span == 3600) {
		*fp = ".%Y%m%d~%H";
		*sp = "";
	} else if (m_directive.rotate.span < 86400) {
		*fp = ".%Y%m%d~%H";
		*sp = "-%Y%m%d~%H";
	} else if (m_directive.rotate.span == 86400) {
		*fp = ".%Y%m%d";
		*sp = "";
	} else {
		*fp = ".%Y%m%d";
		*sp = "-%Y%m%d";
	}
}

/*************************************************************************/

void File::set_compressor(modules::Module *module)
{
	cleanup_compressor();
	
	if (!module)
		return;
	
	m_compressor = (Compressor *)module->factory(Compressor::IID);
	
	if (!m_compressor)
		me->core().log(LogFileError(),
				"For file %s: %s is not a compression module",
				m_directive.filename.value().c_str(), module->name().c_str());
}

/*************************************************************************/

void File::setup_compressor()
{
	if (m_directive.rotate && m_directive.rotate.compress)
		set_compressor(me->core().modules().find(m_directive.rotate.compress.module,
				Module::msStarted));
	else
		set_compressor(0);
}

/*************************************************************************/

void File::setup_rotator()
{
	/* If there is a rotator and the rotate directive is set, it
	 * means the rotate directive was previously set as well. So, we
	 * just check if we need to remove some files.
	 *
	 * This will remove ALL files if keep has been set to zero (which
	 * is good!) */
	if (m_rotator && m_directive.rotate)
	{
		/* Remove files exceeding the limit set by the keep directive. */
		if (!m_rotator->remove())
			cleanup_rotator();
		return;
	}
	
	cleanup_rotator();
	
	if (m_directive.rotate)
	{
		m_rotator = new File::Rotator(*this);
		if (!m_rotator->ready())
			cleanup_rotator();
	}
}

/*************************************************************************/

void File::setup_task()
{
	/* Check if we can just update the running task. */
	if (m_task && m_directive.rotate && m_directive.rotate.span != 0)
	{
		m_task->step(m_directive.rotate.span);
		return;
	}
	
	cleanup_task();
	
	if (m_directive.rotate && m_directive.rotate.span != 0)
	{
		m_task = new File::RotateTask(*this);
		if (!me->core().tasks().add(*m_task))
		{
			me->core().log(LogFileError(),
					"Can't add rotation task for file %s",
					m_directive.filename.value().c_str());
			delete m_task;
			m_task = 0;
		}
	}
}

/*************************************************************************/

File::RotateTask::RotateTask(File &file) :
		scheduler::TaskStep(me->task_type(), file.directive().rotate.span, 0),
		m_file(&file)
{
	
}

/*************************************************************************/

std::string File::RotateTask::description() const
{
	char desc[PATH_MAX];
	snprintf(desc, sizeof(desc), "Rotating log file %s",
			m_file->directive().filename.value().c_str());
	return desc;
}

/*************************************************************************/

int File::RotateTask::run()
{
	m_file->rotate();
	return rOK;
}

/*************************************************************************/

File::Rotator::Rotator(File &file) : m_file(&file), m_index(), m_ready(false)
{
	m_ready = load();
}

/*************************************************************************/

bool File::Rotator::add(std::string const &name)
{
	if (!m_ready)
		return false;
	
	m_index.push_back(name);
	return remove();
}

/*************************************************************************/

bool File::Rotator::load()
{
	if (m_ready)
		return true;
	
	char name[PATH_MAX];
	snprintf(name, sizeof(name), "%s.idx",
			m_file->directive().filename.value().c_str());
	
	/* Make sure the file exists */
	ofstream out(absolute(name), ios::out | ios::app);
	if (!out)
	{
		me->core().log_perror(LogFileError(), "Can't open %s (log "
				"rotation index)", name);
		return false;
	}
	out.close();
	
	/* Now.. open the file for reading */
	ifstream in(absolute(name), ios::in);
	if (!in)
	{
		me->core().log_perror(LogFileError(), "Can't open %s (log "
				"rotation index)", name);
		return false;
	}
	
	std::string entry;
	
	while (getline(in, entry))
	{
		if (entry.empty() || entry[0] == '#')
			continue;
		m_index.push_back(entry);
	}
	
	if (in.fail() && !in.eof())
	{
		me->core().log_perror(LogFileError(),
				"Read error from %s (log rotation index)", name);
		return false;
	}
	
	in.close();
	
	m_ready = true;
	return remove();
}

/*************************************************************************/

bool File::Rotator::remove()
{
	if (!m_ready)
		return false;
	
	int howmany = (m_index.size() - m_file->directive().rotate.keep);
	
	for (int i = 0; i < howmany; ++i)
	{
		if (::remove(absolute(m_index.front().c_str())) == -1 && errno != ENOENT)
		{
			me->core().log_perror(LogFileError(),
					"Can't delete file %s for log rotation",
					m_index.front().c_str());
			return false;
		}
		
		m_index.pop_front();
	}
	
	return save();
}

/*************************************************************************/

bool File::Rotator::save()
{
	if (!m_ready)
		return false;
	
	char name[PATH_MAX];
	snprintf(name, sizeof(name), "%s.idx",
			m_file->directive().filename.value().c_str());
	
	ofstream out(absolute(name), ios::out | ios::trunc);
	if (!out)
	{
		me->core().log_perror(LogFileError(), "Can't open %s (log "
				"rotation index)", name);
		return (m_ready = false);
	}
	
	if (!(out << "# Log rotation index -- Do not remove or modify!") ||
			!(out << endl))
	{
		me->core().log_perror(LogFileError(),
				"Write error to %s (log rotation index)", name);
		return (m_ready = false);
	}
	
	container_type::iterator it, end;
	for (it = m_index.begin(), end = m_index.end(); it != end; ++it)
	{
		if (!(out << *it) || !(out << endl))
		{
			me->core().log_perror(LogFileError(),
					"Write error to %s (log rotation index)", name);
			return (m_ready = false);
		}
	}
	
	return true;
}

/*************************************************************************/

} /* namespace logfile */

/*************************************************************************/
