/* Log file.
 *
 * Log file module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef FILE_H
#define FILE_H

/*************************************************************************/

#include <fstream>
#include <string>
#include <deque>

#include <scheduler/scheduler.h>

#include "conf.h"

/*************************************************************************/

namespace filecomp { class Compressor; }
namespace modules { class Module; }
namespace xlog { class LogLevel; }

/*************************************************************************/

namespace logfile
{

/*************************************************************************/

class File
{
	public:
		static int const timestamp_length = 512;
	
		/// Constructor.
		File(hidden::Directives::File const &directive);
		/// Destructor.
		~File();
		
		/// Open the log file.
		bool open();
		
		/// Close the log file.
		void close();
		
		/// Logs a message
		void log(xlog::LogLevel const &level, char const *message);
		
		/// Rotates the log file.
		bool rotate(bool manual = false);
		
		/// Sets the compression module (and check if it is!).
		void set_compressor(modules::Module *module);
		
		/// Returns the associated configuration directive.
		hidden::Directives::File const &directive() const
				{ return m_directive; }
		/// Sets the associated configuration directive.
		hidden::Directives::File const
				&directive(hidden::Directives::File const &directive);
		
		/// Is the file marked?
		bool marked() const { return m_marked; }
		/// Changes the marked state.
		bool marked(bool new_value) { return (m_marked = new_value); }
		
		/// Is the log file ready?
		bool ready() const { return m_ready; }
	private:
		/* Classes */
		
		/// Log rotation manager -- keeps track of previously rotated files
		class Rotator
		{
			public:
				/* Typedefs */
				typedef std::deque<std::string> container_type;

				/* Functions */
				
				/// Constructor.
				Rotator(File &file);
				
				/// Adds a file to manage.
				bool add(std::string const &name);
				/// Removes older files as needed.
				bool remove();
				
				/// Is the rotator ready?
				bool ready() const { return m_ready; }
			private:
				File *m_file;
				container_type m_index;
				bool m_ready;

				Rotator(Rotator const &);
				Rotator &operator=(Rotator const &);
				
				/// Loads the index.
				bool load();
				/// Saves the index.
				bool save();
		};
		
		/// Rotation task
		class RotateTask : public scheduler::TaskStep
		{
			public:
				/// Constructor.
				RotateTask(File &file);
				
				/// Implements Task::run().
				virtual int run();
				/// Implements Task::name().
				virtual std::string name() const
					{ return (char const *) m_file->directive().filename; }
				/// Implements Task::description().
				virtual std::string description() const;
			private:
				File *m_file;
				
				RotateTask(RotateTask const &);
				RotateTask &operator=(RotateTask const &);
		};
		
		/* Data */
		hidden::Directives::File m_directive;
		
		std::ofstream m_stream;
		
		Rotator *m_rotator;
		RotateTask *m_task;
		filecomp::Compressor *m_compressor;
		
		bool m_marked;
		bool m_ready;
		
		File (File const &);
		File &operator=(File const &);
		
		/// Sets up file compressor.
		void setup_compressor();
		/// Cleans up file compressor.
		void cleanup_compressor();
		
		/// Sets up log rotation manager.
		void setup_rotator();
		/// Cleans up log rotation manager.
		void cleanup_rotator();
		
		/// Sets up rotation task.
		void setup_task();
		/// Cleans up rotation task.
		void cleanup_task();
		
		/// Determines usable suffix.
		void rotation_suffix(char const **fp, char const **sp,
				bool manual = false);
};

/*************************************************************************/

} /* namespace logfile */

/*************************************************************************/

#endif /* FILE_H */
