/* XML import/export.
 *
 * Twilight database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <map>

#include <databases/field.h>
#include <misc/misc.h>
#include <modules/core.h>

#include "dbfile.h"
#include "decl.h"
#include "module.h"
#include "xml.h"

using namespace std;

using databases::Field;
using databases::Record;
using databases::XMLIndex;
using databases::XMLReader;
using databases::XMLWriter;

/*************************************************************************/

namespace twilight
{

/*************************************************************************/

class XICallback : public XMLReader::Callback
{
	public:
		/// Constructor.
		XICallback(DBWriter &header, DBWriter &data) : m_header(&header),
				m_data(&data), m_fieldids() { }
	
		/// Overrides XMLReader::Callback::database().
		virtual bool database(uint16_t version);
		
		/// Overrides XMLReader::Callback::add_field().
		virtual bool add_field(char const *name, uint16_t type,
				size_t size, uint16_t attrs, char const *defval);
		
		/// Overrides XMLReader::Callback::count().
		virtual bool count(uint32_t count);
		
		/// Overrides XMLReader::Callback::begin_record().
		virtual bool begin_record();
		
		/// Overrides XMLReader::Callback::add_value().
		virtual bool add_value(char const *field,
				char const *value);
		
		/// Overrides XMLReader::Callback::add_ivalue().
		virtual bool add_ivalue(char const *field, int32_t value);
		/// Overrides XMLReader::Callback::add_uivalue().
		virtual bool add_uivalue(char const *field,
				uint32_t value);
		/// Overrides XMLReader::Callback::add_utvalue().
		virtual bool add_utvalue(char const *field, time_t value);
		/// Overrides XMLReader::Callback::add_rivalue().
		virtual bool add_rivalue(char const *field,
				Record::identifier_type value);
		
		/// Overrides XMLReader::Callback::end_record().
		virtual bool end_record();
	private:
		DBWriter *m_header;
		DBWriter *m_data;
		
		map<string, uint16_t> m_fieldids;
		
		XICallback(XICallback const &);
		XICallback &operator=(XICallback const &);
};

/*************************************************************************/

bool XICallback::database(uint16_t version)
{
	return m_header->write_dbinfo(version);
}

/*************************************************************************/

bool XICallback::add_field(char const *name, uint16_t type,
		size_t size, uint16_t attrs, char const *defval)
{
	uint16_t fieldid = m_fieldids.size();
	m_fieldids[name] = fieldid;
	
	return m_header->write_fieldinfo(fieldid, name, type, size, attrs, defval);
}

/*************************************************************************/

bool XICallback::count(uint32_t count)
{
	return m_data->write_count(count);
}

/*************************************************************************/

bool XICallback::begin_record()
{
	return true;
}

/*************************************************************************/

bool XICallback::add_value(char const *field,
		char const *value)
{
	map<string, uint16_t>::const_iterator it = m_fieldids.find(field);
	
	if (it == m_fieldids.end())
		return false;
	
	return (m_data->write_valueinfo(it->second, Field::ftString) &&
			m_data->write_string(value));
}

/*************************************************************************/

bool XICallback::add_ivalue(char const *field, int32_t value)
{
	map<string, uint16_t>::const_iterator it = m_fieldids.find(field);
	
	if (it == m_fieldids.end())
		return false;
	
	return (m_data->write_valueinfo(it->second, Field::ftInteger) &&
			m_data->write_integer(value));
}

/*************************************************************************/

bool XICallback::add_uivalue(char const *field,
		uint32_t value)
{
	map<string, uint16_t>::const_iterator it = m_fieldids.find(field);
	
	if (it == m_fieldids.end())
		return false;
	
	return (m_data->write_valueinfo(it->second, Field::ftUnsigned) &&
			m_data->write_unsigned(value));
}

/*************************************************************************/

bool XICallback::add_utvalue(char const *field, time_t value)
{
	map<string, uint16_t>::const_iterator it = m_fieldids.find(field);
	
	if (it == m_fieldids.end())
		return false;
	
	return (m_data->write_valueinfo(it->second, Field::ftUnixTime) &&
			m_data->write_time(value));
}

/*************************************************************************/

bool XICallback::add_rivalue(char const *field,
		Record::identifier_type value)
{
	uint16_t fieldid;
	
	if (!stricmp(field, "id"))
	{
		fieldid = DBFile::id_fieldid;
	} else
	{
		map<string, uint16_t>::const_iterator it = m_fieldids.find(field);
	
		if (it == m_fieldids.end())
			return false;
		
		fieldid = it->second;
	}
	
	return (m_data->write_valueinfo(fieldid, Field::ftIdentifier) &&
			m_data->write_id(value));
}

/*************************************************************************/

bool XICallback::end_record()
{
	return m_data->write_endrecord();
}

/*************************************************************************/

bool XMLExport::export_data(std::string const &dir, databases::XMLIndex &index,
		std::string const &database)
{
	databases::XMLIndex::const_iterator it, end;
	
	for (it = me->index().begin(), end = me->index().end(); it != end; ++it)
	{
		if (!database.empty() && (*it != database))
			continue;
		
		DBReader dbh(*it + ".dbh");
		DBReader dbd(*it + ".dbd");
		
		if (!dbh.open() || !dbd.open() || !dbh.read_version())
		{
			errorf("Could not open %s database files.", (*it).c_str());
			return false;
		}
		
		XMLWriter xw;
		
		uint16_t version;
		
		if (!dbh.read_dbinfo(&version))
		{
			errorf("Read error");
			return false;
		}
		
		if (!xw.database(version))
		{
			errorf("XML write error: %s", xw.error());
			return false;
		}
		
		uint32_t size;
		char const *name, *defval;
		uint16_t fieldid, type, attrs;
		
		map<uint16_t, string> fieldnames;
		
		for (;;)
		{	
			if (!dbh.read_fieldinfo(&fieldid, &name, &type, &size, &attrs,
					&defval))
			{
				errorf("Read error");
				return false;
			}
			
			if ((fieldid == DBFile::end_fieldid) && (type == DBFile::end_type))
				break;
			
			if (!xw.add_field(name, type, size, attrs, defval))
			{
				errorf("XML write error: %s", xw.error());
				return false;
			}
			
			fieldnames[fieldid] = name;
		}
		
		uint32_t count;
		
		if (!dbd.read_count(&count))
		{
			errorf("Read error");
			return false;
		}
		
		if (!xw.count(count))
		{
			errorf("XML write error: %s", xw.error());
			return false;
		}
		
		for (uint32_t i = 0; i < count; ++i)
		{
			if (!xw.begin_record())
			{
				errorf("XML write error: %s", xw.error());
				return false;
			}
			
			for (;;)
			{
			
				if (!dbd.read_valueinfo(&fieldid, &type))
				{
					errorf("Read error");
					return false;
				}
				
				if ((fieldid == DBFile::end_fieldid) && (type == DBFile::end_type))
					break;
				
				if (fieldid != DBFile::id_fieldid)
				{
					map<uint16_t, string>::iterator it = fieldnames.find(fieldid);
					if (it == fieldnames.end())
					{
						errorf("Got a value of field ID %d but no corresponding field "
								"(corrupted database?)", fieldid);
						return false;
					}
				
					name = it->second.c_str();
				} else {
					name = "id";
				}
				
				switch (type)
				{
					case Field::ftString:
					{
						char const *value;
						if (!dbd.read_string(&value))
						{
							errorf("Read error");
							return false;
						}
						
						if (!xw.add_value(name, value))
						{
							errorf("XML write error: %s", xw.error());
							return false;
						}
					}
						break;
					case Field::ftInteger:
					{
						int32_t value;
						if (!dbd.read_integer(&value))
						{
							errorf("Read error");
							return false;
						}
						
						if (!xw.add_ivalue(name, value))
						{
							errorf("XML write error: %s", xw.error());
							return false;
						}
					}
						break;
					case Field::ftUnsigned:
					{
						uint32_t value;
						if (!dbd.read_unsigned(&value))
						{
							errorf("Read error");
							return false;
						}
						
						if (!xw.add_uivalue(name, value))
						{
							errorf("XML write error: %s", xw.error());
							return false;
						}
					}
						break;
					case Field::ftUnixTime:
					{
						time_t value;
						if (!dbd.read_time(&value))
						{
							errorf("Read error");
							return false;
						}
						
						if (!xw.add_utvalue(name, value))
						{
							errorf("XML write error: %s", xw.error());
							return false;
						}
					}
						break;
					case Field::ftIdentifier:
					{
						Record::identifier_type value;
						if (!dbd.read_id(&value))
						{
							errorf("Read error");
							return false;
						}
						
						if (!xw.add_rivalue(name, value))
						{
							errorf("XML write error: %s", xw.error());
							return false;
						}
					}
						break;
					default:
						errorf("Bad type %d for field %s "
							"(corrupted database?)", type, name);
						break;
				}
			}
			
			if (!xw.end_record())
			{
				errorf("XML write error: %s", xw.error());
				return false;
			}
		}
		
		if (!xw.save(misc::absolute_path(dir, *it + ".xml")))
		{
			errorf("XML save error: %s", xw.error());
			return false;
		}
		
		if (!index.add(*it + ".xml"))
		{
			errorf("Could not add database %s to index file", it->c_str());
			return false;
		}
		
		me->core().log(InfoTwilight(), "Exported %s", it->c_str());
	}
	
	return true;
}

/*************************************************************************/

bool XMLImport::import_data(std::string const &dir,
		std::string const &database)
{
	if (me->index().exists(database))
	{
		errorf("Database %s already exists", database.c_str());
		return false;
	}
	
	DBWriter dbh(database + ".dbh");
	DBWriter dbd(database + ".dbd");
	
	if (!dbh.open() || !dbd.open())
	{
		errorf("Database creation failed.");
		return false;
	}
	
	if (!dbh.write_version())
	{
		errorf("Write error");
		return false;
	}
	
	XMLReader xr;
	XICallback xic(dbh, dbd);
	
	if (!xr.load(misc::absolute_path(dir, database + ".xml"), xic))
	{
		if (*xr.error())
			errorf("XML loading error: %s", xr.error());
		else
			errorf("Write error");
		return false;
	}
	
	if (!dbh.write_endfields())
	{
		errorf("Write error");
		return false;
	}
	
	if (!dbh.commit() || !dbd.commit() || !dbh.cleanup() || !dbd.cleanup())
	{
		errorf("Database cleanup failed.");
		return false;
	}
	
	if (!me->index().add(database) || !me->save_index())
	{
		errorf("Can't save database index file twilight.xml: %s",
				me->index().error().c_str());
		return false;
	}
	
	me->core().log(InfoTwilight(), "Imported database %s", database.c_str());
	return true;
}

/*************************************************************************/

} /* namespace twilight */

/*************************************************************************/
