/* NickServ records management.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#include <map>
#if HAVE_OPENSSL
# include <openssl/rand.h>
#endif

#include <core/directives.h>
#include <core/epona.h>
#include <databases/database.h>
#include <databases/field.h>
#include <log/levels.h>
#include <mail/mail.h>
#include <modules/manager.h>
#include <password/password.h>

#include "chanserv.h"
#include "epona.h"
#include "language.h"
#include "nickserv.h"
#include "operserv.h"
#include "services.h"

using namespace std;

using databases::Collection;
using databases::Database;
using databases::Field;
using databases::Record;

/*************************************************************************/

databases::Field *NickAlias::f_nick = 0;
databases::Field *NickAlias::f_group = 0;
databases::Field *NickAlias::f_created = 0;
databases::Field *NickAlias::f_last_seen = 0;
databases::Field *NickAlias::f_last_usermask = 0;
databases::Field *NickAlias::f_last_realname = 0;
databases::Field *NickAlias::f_last_quit = 0;
databases::Field *NickAlias::f_status = 0;

databases::Field *NickGroup::f_name = 0;
databases::Field *NickGroup::f_enc = 0;
databases::Field *NickGroup::f_pass = 0;
databases::Field *NickGroup::f_created = 0;
databases::Field *NickGroup::f_flags = 0;
databases::Field *NickGroup::f_email = 0;
databases::Field *NickGroup::f_auth_code = 0;
databases::Field *NickGroup::f_url = 0;
databases::Field *NickGroup::f_im_type = 0;
databases::Field *NickGroup::f_im_id = 0;
databases::Field *NickGroup::f_greet = 0;
databases::Field *NickGroup::f_language = 0;
databases::Field *NickGroup::f_chost = 0;
databases::Field *NickGroup::f_chost_changed = 0;
databases::Field *NickGroup::f_chost_setby = 0;
databases::Field *NickGroup::f_maxmemos = 0;

databases::Field *NickGroup::Access::f_owner = 0;
databases::Field *NickGroup::Access::f_mask = 0;

Aliases aliases;
Groups groups;

/*************************************************************************/

namespace
{

/*************************************************************************/

#define HASH(name)	((static_cast<unsigned char>(tolower((name)[0]))&31)<<5 | (static_cast<unsigned char>(tolower((name)[1]))&31))

class AliasHash : public databases::NameCollection::name_hash
{
	public:
		virtual int compare(char const *const &key,
				databases::NameRecord *const &value) const 
				{ return stricmp(key, value->name()); }	
		virtual uint32_t hash(char const *const &key, uint32_t max) const
				{ return HASH(key); }
		virtual bool match(char const *const &key,
				databases::NameRecord *const &value) const
				{ return !stricmp(key, value->name()); }
};

#undef HASH

/*************************************************************************/

typedef map<std::string, password::Encrypter *> EncMap;
EncMap enc_map;

AliasHash alias_hash;

time_t const ls_sensibility = 600;

/*************************************************************************/

} /* anonymous namespace */

/*************************************************************************/

Aliases::Aliases() : databases::NameCollection(alias_hash)
{
}

/*************************************************************************/

bool Aliases::add(databases::Record *record, int *err)
{	
	NickAlias *na = (NickAlias *)record;
	
	if (!na->ng())
		return false;
	
	if (!NameCollection::add(record, err))
		return false;
	
	/* Resolve group name here. */
	if (na->ng()->name_id() && (na->id() == na->ng()->name_id()))
	{	
		na->ng()->name_a(na);
		na->ng()->name_id(0);
		
		if (!(status() & csInitial))
			epona->log(InfoNickServList(),
					"Nickname group %s has been created",
							na->ng()->name());
	}
	
	/* Add the alias to the group */
	na->ng()->aliases().add(na);
	
	/* Make sure the group creation time is coherent
	 * (even if it's entirely cosmetic ;) */
	if ((na->ng()->created() > na->created()))
		na->ng()->created(na->created());
	
	/* Check for services roots */
	conf::ConfStrCharA::container_type::const_iterator it, end =
					epona->cd().operserv.root->end();
	
	if (!(na->ng()->flags() & NickGroup::gfServicesRoot))
	{
		for (it = epona->cd().operserv.root->begin(); it != end; ++it) 
		{
			if (!stricmp(*it, na->nick()))
			{
				na->ng()->flags_add(NickGroup::gfServicesRoot);
				break;
			}
		}
	}
	
	/* Set the user, if it has not already been set */
	if (na->u || (na->u = finduser(na->nick())))
		na->u->na = na;
	
	/* Do post registration stuff */
	na->post_registered((status() & (csInitial | csCreation)) ? true : false);
	
	if (status() & csCreation)
		epona->log(InfoNickServList(),
				"Nickname %s has been registered", na->nick());
	
	return true;
}

/*************************************************************************/

void Aliases::cleanup_groups()
{
	Groups::const_iterator it, end;
    	
    for (it = groups.begin(), end = groups.end(); it != end; )
    {
    	NickGroup *ng = (NickGroup *)*it++;
    	
    	if (!ng->name_a())
    	{
    		ng->change_name(0);
    		
    		if (ng->name_a())
    		{
	    		epona->log(InfoNickServList(),
    					"Named previously nameless group %s",
    					ng->name());
    		} else
	    	{
	    		epona->log(InfoNickServList(),
	    				"Removing group " RECORD_FORMAT " with no aliases",
	    				ng->id());
	    		groups.remove(ng);
	    		continue;
	    	}
	    }
	    
	    ng->name_id(0);
    }
}

/*************************************************************************/

databases::Record *Aliases::create(databases::Record::identifier_type id) const
{
	if (id)
		return new NickAlias(id);
	else
		return new NickAlias;
}

/*************************************************************************/

bool Aliases::end_import(uint16_t type)
{
	if (!NameCollection::end_import(type))
		return false;
	
	if (type == csCreation && groups.is_new())
	{
		cleanup_groups();
		groups.reset_is_new();
	}
		
	return true;
}

/*************************************************************************/

bool Aliases::end_initial()
{
	if (!NameCollection::end_initial())
		return false;
	
	if (groups.is_new())
	{
		cleanup_groups();
		groups.reset_is_new();
	}
	
	return true;
}

/*************************************************************************/

void Aliases::expire()
{
	if (!epona->cd().nickserv.expire)
		return;
	
	if (!writable() || !groups.writable())
		return;
	
	NickAlias *na;
	time_t now = time(0);
	const_iterator it, end;
	
	for (it = begin(), end = this->end(); it != end; ) 
	{
		na = (NickAlias *)*it++;
			
		if (na->status() & NickAlias::asNoExpire)
			continue;
		
		if (na->u && ((na->ng()->flags() & NickGroup::gfSecure) ?
				nick_identified(na->u) : nick_recognized(na->u)) &&
				(now - na->last_seen()) > 43200) 
		{
			if (epona->debuglevel() >= 2)
				epona->log(DebugNickServ(),
						"Updating last seen time of nickname %s", na->nick());
			na->last_seen(now);
			continue;
		}
		
		if (((now - na->last_seen()) >= (epona->cd().nickserv.expire.delay +
				 epona->cd().nickserv.expire.grace))) 
		{
			epona->log(InfoNickServList(),
					"Expiring nickname %s (group: %s) (e-mail: %s)",
					na->nick(), na->ng()->name(),
					(na->ng()->email() ? na->ng()->email() : "none"));
			
			if (!remove(na))
			{
				epona->log(ErrorNickServList(),
						"Expiration of nickname %s failed", na->nick());
				
				/* If it fails, we may have switched to read-only,
				 * so check for that. */
				if (!writable() || !groups.writable())
					return;
			}
			
			continue; /* Important. */
		}
		
		if ((now - na->last_seen()) >= (epona->cd().nickserv.expire.delay))
			na->status_add(NickAlias::asPendingExpire);
	}
}

/*************************************************************************/

databases::Field const *Aliases::owner_field(Collection const *coll) const
{
	return (coll == &groups) ? NickAlias::f_group : 0;
}

/*************************************************************************/

bool Aliases::remove(databases::Record *record, int *err)
{	
	NickAlias *na = (NickAlias *)record;
	
	/* Remove any timeout belonging to the nick we're deleting */
	clean_ns_timeouts(na);
	
	/* Do unregistration stuff */
	na->post_unregistered();
	
	User *u = na->u;
	NickGroup *ng = na->ng();
	
	misc::cstring nick(na->nick());
	
	/* Remove the record from the collection */
	if (!NameCollection::remove(record, err))
	{
		/* If the removal failed, register it again :P
		 * (do not revalidate) */
		na->post_registered(false);
		return false;
	}
	
	/* na IS INVALID FROM HERE */
	
	/* The user is no longer registered */
	if (u)
		u->na = 0;
	
	/* See if the group is still useful. */
	ng->aliases().remove(na);
	
	if (ng->aliases().size() == 0)
	{
		ng->name_a(0);
		
		if (!(ng->db_status() & NickGroup::dbsRemoving) && groups.remove(ng))
			epona->log(InfoNickServList(), "Deleted nickname group %s",
					nick.str());
	} else {
		/* Group name updating stuff */
		if (!ng->name_a() || na == ng->name_a())
		{
			ng->name_a(0); /* VERY IMPORTANT -- or it crashes */
			ng->change_name(0);
		}
	}
	
	if (status() & csDeletion)
		epona->log(InfoNickServList(),
				"Nickname %s has been deleted", nick.str());
	
	return true;
}

/*************************************************************************/

bool Aliases::remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err)
{
	if (coll != &groups)
		return true;
	
	NickGroup *ng = (NickGroup *)record;
	
	if (ng->aliases().empty())
		return true;
	
	return NameCollection::remove_owned_db(coll, record, err);
}

/*************************************************************************/

bool Aliases::remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err)
{
	if (coll != &groups)
		return true;
	
	NickGroup *ng = (NickGroup *)record;
	
	while (!ng->aliases().empty())
	{
		NickAlias *na = (NickAlias *)ng->aliases()[ng->aliases().size() - 1];
		
		na->db_status_remove(Record::dbsInsert);
		
		if (!remove(na, err))
			return false;
	}
	
	return true;
}

/*************************************************************************/

void Aliases::setup_database(databases::Database *db)
{
	db->coll(this);
	db->name("nickserv_aliases");
	db->version(Database::vtExport, 1);
	db->version(Database::vtMinimum, 1);
	db->perms_add(Database::dpInitial |
			Database::dpImport | Database::dpExport);
	
	NickAlias::f_nick = db->field_create("nick", 1, Field::ftString,
			NickAlias::nick_len, 0, 0, 1);
	NickAlias::f_group = db->field_create("group", 1, Field::ftIdentifier,
			0, Field::faOwner, "0", 2);
	NickAlias::f_created = db->field_create("created", 1, Field::ftUnixTime,
			0, 0, "0", 3);
	NickAlias::f_last_seen = db->field_create("last_seen", 1,
			Field::ftUnixTime, 0, 0, "0", 4);
	NickAlias::f_last_usermask = db->field_create("last_usermask", 1,
			Field::ftString, NickAlias::last_usermask_len, 0, 0, 5);
	NickAlias::f_last_realname = db->field_create("last_realname", 1,
			Field::ftString, NickAlias::last_realname_len, 0, 0, 6);
	NickAlias::f_last_quit = db->field_create("last_quit", 1,
			Field::ftString, NickAlias::last_quit_len, 0, 0, 7);
	NickAlias::f_status = db->field_create("status", 1, Field::ftUnsigned,
			2, 0, "0", 8);
	
	if (!db->field_add(NickAlias::f_nick) || 
			!db->field_add(NickAlias::f_group) ||
			!db->field_add(NickAlias::f_created) || 
			!db->field_add(NickAlias::f_last_seen) || 
			!db->field_add(NickAlias::f_last_usermask) || 
			!db->field_add(NickAlias::f_last_realname) ||
			!db->field_add(NickAlias::f_last_quit) || 
			!db->field_add(NickAlias::f_status))
		epona->fatal("Can't add fields to database nickserv_aliases");
	
	this->db(db);
}

/*************************************************************************/

void Aliases::setup_owner()
{
	groups.set_owned(*this);
}

/*************************************************************************/

bool Groups::add(databases::Record *record, int *err)
{	
	if (!Collection::add(record, err))
		return false;

	NickGroup *ng = (NickGroup *)record;
	
	if (ng->flags() & NickGroup::gfServicesAdmin)
		servadmins.add(ng);
	if (ng->flags() & NickGroup::gfServicesOper)
		servopers.add(ng);
	
	if (status() & (csInitial | csCreation))
		m_is_new = true;
	
	return true;
}

/*************************************************************************/

databases::Record *Groups::create(databases::Record::identifier_type id) const
{
	if (id)
		return new NickGroup(id);
	else
		return new NickGroup;
}

/*************************************************************************/

password::Encrypter *Groups::load_encrypter(char const *name)
{
	EncMap::const_iterator it = enc_map.find(name);
		
	if (it != enc_map.end())
	{
		it->second->module()->references_add();
		return it->second;
	}
	
	modules::Module *m = 0;
	
	if (!epona->modules().load(name) || !(m = epona->modules().find(name)))
	{
		epona->log(ErrorNickServ(), "Can't load password module %s: %s",
				name, epona->modules().error().c_str());
		return false;
	}
	
	password::Encrypter *enc;
	
	if (!(enc = (password::Encrypter *)m->interface(password::Encrypter::IID)))
	{
		epona->log(ErrorNickServ(), "%s is not a password module", name);
		return false;
	}
	
	enc_map[m->name()] = enc;
	enc->module()->references_add();
	
	return enc;
}

/*************************************************************************/

void Groups::release_encrypter(password::Encrypter *enc)
{
	if (enc)
		enc->module()->references_remove();
}

/*************************************************************************/

bool Groups::remove(databases::Record *record, int *err)
{
	NickGroup *ng = (NickGroup *)record;
	
	ng->db_status_add(NickGroup::dbsRemoving);
	
	if (ng->flags() & NickGroup::gfServicesAdmin)
		servadmins.remove(ng);
	if (ng->flags() & NickGroup::gfServicesOper)
		servopers.remove(ng);
    
    /* Remove aliases AFTER the other dependency cleanup (because
     * they may still refer to the group name) */
	
	if (!Collection::remove(record, err))
	{
		ng->db_status_remove(NickGroup::dbsRemoving);
		return false;
	}
	
	return true;
}

/*************************************************************************/

void Groups::setup_database(databases::Database *db)
{
	db->coll(this);
	db->name("nickserv_groups");
	db->version(Database::vtExport, 2);
	db->version(Database::vtMinimum, 1);
	db->perms_add(Database::dpInitial |
			Database::dpImport | Database::dpExport);
	
	NickGroup::f_name = db->field_create("name", 1, Field::ftIdentifier,
			0, 0, "0", 1);
	NickGroup::f_enc = db->field_create("enc", 1, Field::ftString,
			255, 0, 0, 2);
	NickGroup::f_pass = db->field_create("pass", 1, Field::ftString,
			NickGroup::pass_len, Field::faCase, 0, 3);
	NickGroup::f_created = db->field_create("created", 1, Field::ftUnixTime,
			0, 0, "0", 4);
	NickGroup::f_flags = db->field_create("flags", 1, Field::ftUnsigned,
			4, 0, "2147483648", 5);
	NickGroup::f_email = db->field_create("email", 1, Field::ftString,
			NickGroup::email_len, 0, 0, 6);
	NickGroup::f_auth_code = db->field_create("auth_code", 2, Field::ftUnsigned,
			4, 0, "0", 16);
	NickGroup::f_url = db->field_create("url", 1, Field::ftString,
			NickGroup::url_len, 0, 0, 7);
	NickGroup::f_im_type = db->field_create("im_type", 1, Field::ftString,
			NickGroup::im_type_len, 0, 0, 8);
	NickGroup::f_im_id = db->field_create("im_id", 1, Field::ftString,
			NickGroup::im_id_len, 0, 0, 9);
	NickGroup::f_greet = db->field_create("greet", 1, Field::ftString,
			NickGroup::greet_len, 0, 0, 10);
	NickGroup::f_language = db->field_create("language", 1, Field::ftUnsigned,
			2, 0, "65535", 11);
	NickGroup::f_chost = db->field_create("chost", 1, Field::ftString,
			NickGroup::chost_len, 0, 0, 12);
	NickGroup::f_chost_changed = db->field_create("chost_changed", 1,
			Field::ftUnixTime, 0, 0, "0", 13);
	NickGroup::f_chost_setby = db->field_create("chost_setby", 1,
			Field::ftString, NickGroup::chost_setby_len, 0, 0, 14);
	NickGroup::f_maxmemos =  db->field_create("maxmemos", 1, Field::ftInteger,
			2, 0, "-2", 15);
	
	if (!db->field_add(NickGroup::f_name) || 
			!db->field_add(NickGroup::f_enc) ||
			!db->field_add(NickGroup::f_pass) || 
			!db->field_add(NickGroup::f_created) || 
			!db->field_add(NickGroup::f_flags) || 
			!db->field_add(NickGroup::f_email) ||
			!db->field_add(NickGroup::f_auth_code) || 
			!db->field_add(NickGroup::f_url) || 
			!db->field_add(NickGroup::f_im_type) || 
			!db->field_add(NickGroup::f_im_id) ||
			!db->field_add(NickGroup::f_greet) || 
			!db->field_add(NickGroup::f_language) || 
			!db->field_add(NickGroup::f_chost) ||
			!db->field_add(NickGroup::f_chost_changed) || 
			!db->field_add(NickGroup::f_chost_setby) || 
			!db->field_add(NickGroup::f_maxmemos))
		epona->fatal("Can't add fields to database nickserv_groups");
	
	this->db(db);
}

/*************************************************************************/

bool Groups::Accesses::add(databases::Record *record, int *err)
{
	NickGroup::Access *nga = (NickGroup::Access *)record;
	
	NickGroup::accesses_type::const_iterator it, end;
	
	for (it = nga->owner()->accesses().begin(),
			end = nga->owner()->accesses().end(); it != end; ++it)
	{
		NickGroup::Access const *nga2 = (NickGroup::Access const *)(*it);
		
		if (!nga2)
			continue;
		
		if (*nga == nga2->name())
		{
			if (err)
				*err = ceName;
			return false;
		}
	}
	
	if (!Collection::add(record, err))
		return false;
	
	return true;
}

/*************************************************************************/

bool Groups::Accesses::add_list(Record *record)
{
	NickGroup::Access *nga = (NickGroup::Access *)record;
	
	NickGroup::accesses_type::iterator it, end;
	
	for (it = nga->owner()->accesses().begin(),
			end = nga->owner()->accesses().end(); it != end; ++it)
	{
		if (*it == 0)
		{
			*it = nga;
			break;
		}
	}

	if (it == nga->owner()->accesses().end())
		it = nga->owner()->accesses().insert(it, nga);
	
	if (!Collection::add_list(record))
	{
		*it = 0;
		return false;
	}

	return true;
}

/*************************************************************************/

Record *Groups::Accesses::create(Record::identifier_type id) const
{
	if (id)
		return new NickGroup::Access(id);
	else
		return new NickGroup::Access;
}

/*************************************************************************/

void Groups::Accesses::remove_list(Record *record)
{
	NickGroup::Access *nga = (NickGroup::Access *)record;
	
	NickGroup::accesses_type::iterator it, end;
	
	for (it = nga->owner()->accesses().begin(),
			end = nga->owner()->accesses().end(); it != end; ++it)
	{
		if (nga == (*it))
		{
			*it = 0;
			break;
		}
	}
	
	Collection::remove_list(record);
}

/*************************************************************************/

bool Groups::Accesses::remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err)
{
	if (coll != &groups)
		return true;
	
	NickGroup *ng = (NickGroup *)record;
	
	if (ng->accesses().empty())
		return true;
	
	return Collection::remove_owned_db(coll, record, err);
}

/*************************************************************************/

bool Groups::Accesses::remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err)
{
	if (coll != &groups)
		return true;
	
	NickGroup *ng = (NickGroup *)record;
	NickGroup::accesses_type::iterator it, end;
	
	for (it = ng->accesses().begin(), end = ng->accesses().end();
			it != end; ++it)
	{
		NickGroup::Access *nga = (NickGroup::Access *)(*it);
		
		if (nga)
		{
			nga->db_status_remove(Record::dbsInsert);
		
			if (!remove(nga, err))
				return false;
		}
	}
	
	return true;
}

/*************************************************************************/

void Groups::Accesses::setup_database(databases::Database *db)
{
	db->coll(this);
	db->name("nickserv_access");
	db->version(Database::vtExport, 1);
	db->version(Database::vtMinimum, 1);
	db->perms_add(Database::dpInitial |
			Database::dpImport | Database::dpExport);
	
	NickGroup::Access::f_owner = db->field_create("owner", 1,
			Field::ftIdentifier, 0, Field::faOwner, "0", 1);
	NickGroup::Access::f_mask = db->field_create("mask", 1, Field::ftString,
			NickGroup::Access::mask_len, 0, 0, 2);
	
	if (!db->field_add(NickGroup::Access::f_owner) ||
			!db->field_add(NickGroup::Access::f_mask))
		epona->fatal("Can't add fields to database nickserv_access");
	
	this->db(db);
}

/*************************************************************************/

void Groups::Accesses::setup_owner()
{
	groups.set_owned(*this);
}

/*************************************************************************/

NickAlias::NickAlias() : databases::NameRecord(), u(0), m_nick(0), m_ng(0),
		m_created(time(0)), m_last_seen(0), m_last_usermask(0),
		m_last_realname(0), m_last_quit(0), m_status(0)
{
}

/*************************************************************************/

NickAlias::NickAlias(uint32_t id) : databases::NameRecord(id), u(0), m_nick(0),
		m_ng(0), m_created(time(0)), m_last_seen(0), m_last_usermask(0),
		m_last_realname(0), m_last_quit(0), m_status(0)
{
}

/*************************************************************************/

time_t NickAlias::created(time_t created)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't accept dates in the future */
		if (created > time(0))
		{
			if (db_status() & dbsMember)
				return m_created;
			
			created = time(0);
		}
	}
	
	if (m_created != created)
	{
		if (!aliases.change_ut(this, f_created, created))
			return m_created;
		
		m_created = created;
	}
	
	return m_created;
}

/*************************************************************************/

bool NickAlias::end_update()
{
	if (!NameRecord::end_update())
		return false;
	
	if ((m_status & asPendingExpire) &&
			(time(0) - last_seen()) < epona->cd().nickserv.expire.delay)
	{
		if (aliases.change_ui(this, f_status, (m_status & ~asPendingExpire)))
			m_status &= ~asPendingExpire;
	}
	
	return true;
}

/*************************************************************************/

Record::identifier_type NickAlias::get_ri(Field const *field) const
{
	if (field == f_group)
		return (m_ng ? m_ng->id() : 0);
	else
		return NameRecord::get_ri(field);
}

/*************************************************************************/

char const *NickAlias::get_s(Field const *field) const
{
	if (field == f_nick)
		return nick();
	else if (field == f_last_usermask)
		return last_usermask();
	else if (field == f_last_realname)
		return last_realname();
	else if (field == f_last_quit)
		return last_quit();
	else
		return NameRecord::get_s(field);
}

/*************************************************************************/

uint32_t NickAlias::get_ui(Field const *field) const
{
	if (field == f_status)
		return (status() & ~NickAlias::asTemporary);
	else
		return NameRecord::get_ui(field);
}

/*************************************************************************/

time_t NickAlias::get_ut(Field const *field) const
{
	if (field == f_created)
		return created();
	else if (field == f_last_seen)
		return last_seen();
	else
		return NameRecord::get_ut(field);
}

/*************************************************************************/

char const *NickAlias::last_quit(char const *last_quit)
{
	if (m_last_quit.compare(last_quit, last_quit_len) != 0)
	{
		misc::cstring tlast_quit(last_quit, last_quit_len);
		
		if (!aliases.change_s(this, f_last_quit, tlast_quit))
			return m_last_quit;
		
		m_last_quit.swap(tlast_quit);
	}
	
	return m_last_quit;
}

/*************************************************************************/

char const *NickAlias::last_realname(char const *last_realname)
{
	if (m_last_realname.compare(last_realname, last_realname_len) != 0)
	{
		misc::cstring tlast_realname(last_realname, last_realname_len);
		
		if (!aliases.change_s(this, f_last_realname, tlast_realname))
			return m_last_realname;
		
		m_last_realname.swap(tlast_realname);
	}
	
	return m_last_realname;
}

/*************************************************************************/

time_t NickAlias::last_seen(time_t last_seen)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't accept dates in the future */
		if (last_seen > time(0))
		{
			if (db_status() & dbsMember)
				return m_last_seen;
			
			last_seen = time(0);
		}
		
		if (m_last_seen && (last_seen < m_last_seen))
			return m_last_seen;
	}
	
	if (m_last_seen != last_seen)
	{
		if (((last_seen - m_last_seen) > ls_sensibility) &&
				!aliases.change_ut(this, f_last_seen, last_seen))
			return m_last_seen;
		
		m_last_seen = last_seen;
	}
	
	return m_last_seen;
}

/*************************************************************************/

char const *NickAlias::last_usermask(char const *last_usermask)
{
	if (m_last_usermask.compare(last_usermask, last_usermask_len) != 0)
	{
		misc::cstring tlast_usermask(last_usermask, last_usermask_len);
		
		if (!aliases.change_s(this, f_last_usermask, tlast_usermask))
			return m_last_usermask;
		
		m_last_usermask.swap(tlast_usermask);
	}
	
	return m_last_usermask;
}

/*************************************************************************/

char const *NickAlias::last_usermask(char const *user, char const *host)
{
	if (!user || !host)
		return m_last_usermask;
	
	char tlast_usermask[last_usermask_len + 1];
	snprintf(tlast_usermask, sizeof(tlast_usermask), "%s@%s", user, host);
	
	if (m_last_usermask.compare(tlast_usermask) != 0)
	{
		if (!aliases.change_s(this, f_last_usermask, tlast_usermask))
			return m_last_usermask;
		
		m_last_usermask.assign(tlast_usermask);
	}
	
	return m_last_usermask;
}

/*************************************************************************/

size_t NickAlias::memory_usage() const
{
	size_t result = NameRecord::memory_usage();
	
	result += sizeof(NickAlias);
	if (m_nick)
		result += strlen(m_nick) + 1;
	if (m_last_usermask)
		result += strlen(m_last_usermask) + 1;
	if (m_last_realname)
		result += strlen(m_last_realname) + 1;
	if (m_last_quit)
		result += strlen(m_last_quit) + 1;
	
	return result;
}

/*************************************************************************/

NickGroup *NickAlias::ng(NickGroup *ng)
{
	if (m_ng)
		return m_ng;
	
	return (m_ng = ng);
}

/*************************************************************************/

Record::identifier_type NickAlias::ng_id(identifier_type ng_id)
{
	/* Assume (db_status() & dbsUpdating) == true */
	
	/* Don't allow group changes. */
	if (m_ng)
		return m_ng->id();
	
	NickGroup *ng;
	
	/* If the group doesn't exist, make end_update() fail. */
	if (!ng_id || !(ng = (NickGroup *)groups.find(ng_id)))
	{
		epona->log(ErrorNickServList(), "Ignoring nickname alias %s "
				"with bad group " RECORD_FORMAT, (nick() ? nick() : ""),
				ng_id);
		
		db_status_add(dbsUpdateError);
		return ng_id;
	}
	
	this->ng(ng);
	return m_ng->id();
}

/*************************************************************************/

char const *NickAlias::nick(char const *nick)
{
	if (db_status() & dbsUpdating)
	{
		/* Do not accept NULL nicks */
		if (!nick)
		{
			/* If we're already a member, just return the current nick */
			if (db_status() & dbsMember)
				return m_nick;
			/* Else, make end_update() fail. */
			db_status_add(dbsUpdateError);
			/* And return the NULL nick */
			return 0;
		}
		
		/* Do not accept nick changes (but allow case changes)  */
		if (m_nick && stricmp(m_nick, nick) != 0)
			return m_nick;
	}
	
	if (m_nick.compare(nick, nick_len) != 0)
	{
		misc::cstring tnick(nick, nick_len);
			
		if (!aliases.change_s(this, f_nick, tnick))
			return m_nick;
		
		m_nick.swap(tnick);
	}
	
	return m_nick;
}

/*************************************************************************/

void NickAlias::post_deidentified()
{
	unset_reg_mode();
	m_ng->post_deidentified(this);
}

/*************************************************************************/

void NickAlias::post_identified()
{
	/* Called for everything except do_register() and do_group(). */

	m_ng->post_identified(this);
	set_reg_mode();
	/* That's last so that the eventual custom hostname is already
	 * set for last_usermask stuff. */
	post_recognized();
}

/*************************************************************************/

void NickAlias::post_recognized()
{
	time_t now = time(0);
	
	char umask[last_usermask_len + 1];
	snprintf(umask, sizeof(umask), "%s@%s", u->username, GetHost(u));
	
	if (m_last_usermask.compare(umask) != 0 ||
			m_last_realname.compare(u->realname) != 0 ||
			(now - m_last_seen) > ls_sensibility)
	{
		if (!aliases.change_multi(this, 3, f_last_usermask, umask,
				f_last_realname, u->realname, f_last_seen, now))
			return;
		
		m_last_usermask = umask;
		m_last_realname = u->realname;
		m_last_seen = now;
	}
	
	check_memos(u);
}

/*************************************************************************/

void NickAlias::post_registered(bool validate)
{	
	if (validate) {
		if (u)
			validate_user(u);
	} else {
		m_ng->post_registered(this);
			
		if (u && nick_identified(u))
		{
			set_reg_mode();
			post_recognized();
		}
	}
}

/*************************************************************************/

void NickAlias::post_signoff(char const *quit)
{
	time_t now = time(0);
	
	if (m_last_quit.compare(quit, last_quit_len))
	{
		if ((now - m_last_seen) > ls_sensibility)
		{
			misc::cstring tquit(quit, last_quit_len);
				
			if (!aliases.change_multi(this, 2, f_last_quit,
					tquit.str(), f_last_seen, now))
				return;
			
			m_last_quit.swap(tquit);
			m_last_seen = now;
		} else {
			last_quit(quit);
			m_last_seen = now;
		}
	} else {
		last_seen(now);
	}
}

/*************************************************************************/

void NickAlias::post_unregistered()
{
	unset_reg_mode();
	m_ng->post_unregistered(this);
}

/*************************************************************************/

void NickAlias::set_reg_mode()
{
	if (u)
	{
		if ((time_t)u->svid != u->timestamp) 
		{
			char tsbuf[16];
			snprintf(tsbuf, sizeof(tsbuf), "%ld", (long) u->timestamp);
			change_user_mode(s_NickServ, u,
					(!(u->mode & UMODE_r) ? "+rd" : "+d"), tsbuf);
		} else {
			if (!(u->mode & UMODE_r))
	    		change_user_mode(s_NickServ, u, "+r", NULL);
		}
	}
}

/*************************************************************************/

Record::identifier_type NickAlias::set_ri(Field const *field,
		identifier_type value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_ri(field, value);
	
	if (field == f_group)
		return ng_id(value);
	else
		return NameRecord::set_ri(field, value);
}

/*************************************************************************/

char const *NickAlias::set_s(Field const *field, char const *value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_s(field, value);
	
	if (field == f_nick)
		return nick(value);
	else if (field == f_last_usermask)
		return last_usermask(value);
	else if (field == f_last_realname)
		return last_realname(value);
	else if (field == f_last_quit)
		return last_quit(value);
	else
		return NameRecord::set_s(field, value);
}

/*************************************************************************/

uint32_t NickAlias::set_ui(Field const *field, uint32_t value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_ui(field, value);
	
	if (field == f_status)
		return (status(value) & ~asTemporary);
	else
		return NameRecord::set_ui(field, value);
}

/*************************************************************************/

time_t NickAlias::set_ut(Field const *field, time_t value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_ut(field, value);
	
	if (field == f_created)
		return created(value);
	else if (field == f_last_seen)
		return last_seen(value);
	else
		return NameRecord::set_ut(field, value);
}

/*************************************************************************/

uint16_t NickAlias::status(uint16_t status)
{
	/* Assume db_status() & dbsUpdating */
	
	if ((status & asPendingExpire) &&
			(!epona->cd().nickserv.expire || (status & asNoExpire)))
		status &= ~asPendingExpire;
	
	status &= ~(asTemporary | asUnused);
	
	if ((m_status & ~asTemporary) != status)
	{
		uint16_t tstatus = (m_status & asTemporary);
		
		m_status = status;
		m_status |= tstatus;
	}
	
	return m_status;
}

/*************************************************************************/

uint16_t NickAlias::status_add(uint16_t status)
{
	if ((m_status & status) != status) 
	{
		if ((status & asTemporary) != status)
		{
			uint16_t tstatus = ((m_status | status) & ~asTemporary);
		
			if (status & (asNoExpire | asRecognized | asIdentified))
				tstatus &= ~asPendingExpire;
		
			if (!aliases.change_ui(this, f_status, tstatus))
				return m_status;
		} else {
			if (status & (asRecognized | asIdentified))
				status_remove(asPendingExpire);
		}
		
		m_status |= status;
		
		if (status & (asNoExpire | asTemporary))
			m_status &= ~asPendingExpire;
	}
	
	return m_status;
}

/*************************************************************************/

uint16_t NickAlias::status_remove(uint16_t status)
{
	if ((m_status & status) != 0) 
	{
		if ((status & asTemporary) != status)
		{
			uint16_t tstatus = (m_status & ~asTemporary);
			tstatus &= ~status;
			
			if (!aliases.change_ui(this, f_status, tstatus))
				return m_status;
		}
		
		m_status &= ~status;
	}
	
	return m_status;
}

/*************************************************************************/

void NickAlias::unset_reg_mode()
{
	if (u)	
		change_user_mode(s_NickServ, u,
				((u->mode & UMODE_r) ? "-r+d" : "+d"), "1");
}

/*************************************************************************/

NickGroup::NickGroup() : databases::NameRecord(), memos(this),
		m_name(0), m_name_id(0), m_enc(0), m_pass(0), m_created(time(0)),
		m_flags(gfDefaults | epona->cd().nickserv.default_options), m_email(0),
		m_auth_code(0), m_url(0), m_im_type(0), m_im_id(0), m_greet(0),
		m_language(0), m_chost(0), m_chost_changed(0), m_chost_setby(0),
		m_maxmemos(0), m_aliases(), m_accesses(), m_bad_auths(0),
		m_channelcount(0)
{
}

/*************************************************************************/

NickGroup::NickGroup(databases::Record::identifier_type id) : 
		databases::NameRecord(id), memos(this),
		m_name(0), m_name_id(0), m_enc(0), m_pass(0), m_created(time(0)),
		m_flags(gfDefaults | epona->cd().nickserv.default_options), m_email(0),
		m_auth_code(0), m_url(0), m_im_type(0), m_im_id(0), m_greet(0),
		m_language(0), m_chost(0), m_chost_changed(0), m_chost_setby(0),
		m_maxmemos(0), m_aliases(), m_accesses(), m_bad_auths(0),
		m_channelcount(0)
{
}

/*************************************************************************/

NickGroup::~NickGroup()
{
	groups.release_encrypter(m_enc);
}

/*************************************************************************/

uint32_t NickGroup::auth_code(uint32_t auth_code)
{
	if (m_auth_code != auth_code)
	{
		if (!groups.change_ui(this, f_auth_code, auth_code))
			return m_auth_code;
		
		m_auth_code = auth_code;
	}
	
	return m_auth_code;
}

/*************************************************************************/

bool NickGroup::begin_auth()
{
	uint32_t code = 0;
	
#if HAVE_OPENSSL
	while (!code)
	{
		if (RAND_bytes((unsigned char *)&code, sizeof(code)) != 1 &&
				RAND_pseudo_bytes((unsigned char *)&code, sizeof(code)) == -1)
			return false;
	}
#else
	while ((code = rand()) == 0)
		return false;
#endif

	if (auth_code(code) != code)
		return false;
	
	return true;
}

/*************************************************************************/

bool NickGroup::change_chost(char const *chost, time_t changed,
		char const *setby)
{
	if (m_chost.compare(chost, chost_len) != 0 || m_chost_changed != changed ||
			m_chost_setby.compare(setby, chost_setby_len) != 0)
	{
		misc::cstring tchost(chost, chost_len);
		misc::cstring tsetby(setby, chost_setby_len);
		
		if (!groups.change_multi(this, 3, f_chost, tchost.str(),
				f_chost_changed, changed, f_chost_setby, tsetby.str()))
			return false;
		
		m_chost.swap(tchost);
		m_chost_changed = changed;
		m_chost_setby.swap(tsetby);
	}
	
	return true;
}

/*************************************************************************/

bool NickGroup::change_im(char const *im_type, char const *im_id)
{
	if (m_im_type.compare(im_type, im_type_len) != 0)
	{
		if (m_im_id.compare(im_id, im_id_len) != 0)
		{
			misc::cstring tim_type(im_type, im_type_len);
			misc::cstring tim_id(im_id, im_id_len);
			
			if (!groups.change_multi(this, 2, f_im_type,
					tim_type.str(), f_im_id, tim_id.str()))
				return false;
			
			m_im_type.swap(tim_type);
			m_im_id.swap(tim_id);
			return true;
		}
		else {
			this->im_type(im_type);
			return (m_im_type.compare(im_type) == 0);
		}
	} else {
		this->im_id(im_id);
		return (m_im_id.compare(im_id) == 0);
	}
}

/*************************************************************************/

NickAlias *NickGroup::change_name(NickAlias *name)
{
	if (!name)
	{
		if (m_aliases.size() == 0)
			return m_name;
		
		name = (NickAlias *)m_aliases[0];
	}
	
	if (name != m_name)
	{
		if (!groups.change_ri(this, f_name, name->id()))
			return m_name;
		
		if (m_name)
		{
			epona->log(InfoNickServSettings(),
					"Changing %s nickname group name to %s",
					m_name->nick(), name->nick());
		} else
		{
			epona->log(InfoNickServList(),
					"Nickname group %s has been created",
					name->nick());
		}
		
		m_name = name;
	}
	
	return m_name;
}

/*************************************************************************/

bool NickGroup::change_password(char const *pass)
{
	if (!pass)
		return false;
	
	if (groups.encrypter())
	{
		/* Make sure we can store the password. */
		if (groups.encrypter()->length() > pass_len)
		{
			epona->log(ErrorNickServ(),
					"Encrypted passwords of module %s are too long "
					"(%u characters, max %u)",
					groups.encrypter()->module()->name().c_str(),
					groups.encrypter()->length(), pass_len);
			return false;
		}
		
		misc::cstring tpass;
		tpass.reserve(groups.encrypter()->length() * 2 + 1);
		
		if (!groups.encrypter()->encrypt(pass, tpass,
				groups.encrypter()->length() + 1))
		{
			epona->log(ErrorNickServ(), "Can't encrypt password of "
					"nickname group %s",
					(m_name ? m_name->nick() : "<unnamed>"));
			return false;
		}
				
		if (m_pass.compare(tpass, pass_len) != 0)
		{
			if (!groups.change_multi(this, 2, f_enc, 
					groups.encrypter()->module()->name().c_str(),
					f_pass, tpass.str()))
				return false;
			
			if (m_enc != groups.encrypter())
			{
				groups.release_encrypter(m_enc);
				m_enc = groups.encrypter();
				m_enc->module()->references_add();
			}
			
			m_pass.swap(tpass);
		}
	} else
	{
		if (m_pass.compare(pass, pass_len) != 0)
		{
			misc::cstring tpass(pass, pass_len);
		
			if (!groups.change_multi(this, 2, f_enc, 0,
					f_pass, tpass.str()))
				return false;
			
			if (m_enc)
			{
				groups.release_encrypter(m_enc);
				m_enc = 0;
			}
			
			m_pass.swap(tpass);
		}
	}
	
	return true;
}

/*************************************************************************/

bool NickGroup::check_password(char const *pass)
{
	/* Paranoia */
	if (!m_pass)
		return false;
	
	bool result;
	
	if (m_enc)
		result = m_enc->check(pass, m_pass);
	else
		result = (m_pass.compare(pass, pass_rlen) == 0);
	
	if (result && m_enc != groups.encrypter())
	{
		epona->log(InfoNickServ(), "Changing password encryption scheme "
				"of nickname group %s to %s",
				(m_name ? m_name->nick() : "<unnamed>"),
				(groups.encrypter() ?
				 groups.encrypter()->module()->name().c_str() : "cleartext"));
		change_password(pass);
	}
	
	return result;
}

/*************************************************************************/

char const *NickGroup::chost(char const *chost)
{	
	/* Assume (db_status() & dbsUpdating == true) */
	
	if (m_chost.compare(chost, chost_len) != 0)
		m_chost.assign(chost, chost_len);
	
	return m_chost;
}

/*************************************************************************/

void NickGroup::chost_apply()
{
#ifdef HAS_CHGHOST
	NickAlias *na;
	VectorNA::const_iterator it, end;

	for (it = m_aliases.begin(), end = m_aliases.end(); it != end; ++it)
	{
		na = (NickAlias *)(*it);
		
		if (na->u && nick_identified(na->u))
			custom_host_set(na->u);
	}
#endif
}

/*************************************************************************/

time_t NickGroup::chost_changed(time_t chost_changed)
{
	/* Assume (db_status() & dbsUpdating == true) */
	time_t now = time(0);
	
	if (chost_changed > now)
		chost_changed = now;
	
	if (m_chost_changed != chost_changed)
		m_chost_changed = chost_changed;
	
	return m_chost_changed;
}

/*************************************************************************/

char const *NickGroup::chost_setby(char const *chost_setby)
{
	/* Assume (db_status() & dbsUpdating == true) */
	if (m_chost_setby.compare(chost_setby, chost_setby_len) != 0)
		m_chost_setby.assign(chost_setby, chost_setby_len);
	
	return m_chost_setby;
}

/*************************************************************************/

void NickGroup::chost_unapply()
{
#ifdef HAS_CHGHOST
	NickAlias *na;
	VectorNA::const_iterator it, end;

	for (it = m_aliases.begin(), end = m_aliases.end(); it != end; ++it)
	{
		na = (NickAlias *)(*it);
		
		if (na->u)
			custom_host_unset(na->u);
	}
#endif
}

/*************************************************************************/

time_t NickGroup::created(time_t created)
{
	if (db_status() & dbsUpdating)
	{
		/* We don't allow time in the future. */
		if (created > time(0))
		{
			/* If it's already a member, just return the old time. */
			if (db_status() & dbsMember)
				return m_created;
			/* Else, assign a correct time. */
			created = time(0);
		}
	}
	
	if (m_created != created)
	{
		if (!groups.change_ut(this, f_created, created))
			return m_created;
		
		m_created = created;
	}
	
	return m_created;
}

/*************************************************************************/

char const *NickGroup::email(char const *email)
{
	if (m_email.compare(email, email_len) != 0)
	{	
		misc::cstring temail(email, email_len);
		
		if (!groups.change_s(this, f_email, temail))
			return m_email;
		
		m_email.swap(temail);
	}
	
	return m_email;
}

/*************************************************************************/

char const *NickGroup::enc() const
{
	return (m_enc ? m_enc->module()->name().c_str() : 0);
}

/*************************************************************************/

char const *NickGroup::enc(char const *enc_)
{
	if (!enc_)
	{
		if (m_enc)
		{
			groups.release_encrypter(m_enc);
			m_enc = 0;
		}
	} else 
	{
		if (!m_enc || strcmp(m_enc->module()->name().c_str(), enc_))
		{
			password::Encrypter *enc = groups.load_encrypter(enc_);
		
			if (!enc)
				epona->fatalf("Can't load password module %s required by "
						" nickname group %s", enc_,
						(m_name ? m_name->nick() : "<unnamed>"));
		
			groups.release_encrypter(m_enc);
			m_enc = enc;
		}
	}
	
	return this->enc();
}

/*************************************************************************/

bool NickGroup::end_auth(bool success)
{
	if (success && !(flags_add(gfAuth) & gfAuth))
		return false;
	
	auth_code(0);
	/* Just in case the database suddenly went read-only */
	if (m_auth_code)
		m_auth_code = 0;
	
	return true;
}

/*************************************************************************/

bool NickGroup::end_update()
{
	if (!NameRecord::end_update())
		return false;
	
	/* Don't allow memo forwarding when the e-mail address is not set. */
	if ((flags() & gfMemoForward) && !m_email)
	{
		flags_remove(gfMemoForward);
		/* Force the change in case the database went away */
		if (flags() & gfMemoForward)
			m_flags &= ~gfMemoForward;
	}
	
	/* Don't allow non-NULL IM types if the IM ID is NULL */
	if (m_im_type && !m_im_id)
	{
		im_type(0);
		/* Force the change in case the database went away */
		if (m_im_type)
			m_im_type = 0;
	}
	
	/* Don't allow non-NULL IM IDs if the IM type is NULL */
	if (m_im_id && !m_im_type)
	{
		im_id(0);
		/* Force the change in case the database went away */
		if (m_im_id)
			m_im_id = 0;
	}
	
	/* Clean up m_chost_setby if m_chost is not set */
	if (!m_chost && m_chost_setby)
	{
		groups.change_s(this, f_chost_setby, 0);
		m_chost_setby = 0;
	}
	
	/* Ensure all the chost data is present */
	if (m_chost && (!m_chost_setby || !m_chost_changed))
	{
		epona->log(ErrorNickServList(), "Ignoring bad group #"
				RECORD_FORMAT " with invalid custom hostname data",
				id());
		return false;
	}
	
	/* Don't allow the gfCustomHost flag if no custom hostname is set */
	if ((flags() & gfCustomHost) && !m_chost)
	{
		flags_remove(gfCustomHost);
		/* Force the change in case the database went away */
		if (flags() & gfCustomHost)
			m_flags &= ~gfCustomHost;
	}
	
	return true;
}

/*************************************************************************/

uint32_t NickGroup::flags(uint32_t flags)
{
	/* Assume (db_status() & dbsUpdating) == true */
	
	flags &= ~gfUnused;
	
	if (flags & gfDefaults)
		flags = (gfDefaults | epona->cd().nickserv.default_options);
	
	if (m_flags != flags)
	{
		if (!epona->cd().nickserv.kill_immed)
			flags &= ~gfKillImmed;
		if (!epona->cd().memoserv.forward)
			flags &= ~gfMemoForward;
		
		if (!(groups.status() & (Collection::csInitial | Collection::csCreation)))
		{
			if (!(m_flags & gfServicesAdmin) && (flags & gfServicesAdmin))
				servadmins.add(this);
			if (!(m_flags & gfServicesOper) && (flags & gfServicesOper))
				servopers.add(this);
			
			if ((m_flags & gfServicesAdmin) && !(flags & gfServicesAdmin))
				servadmins.remove(this);
			if ((m_flags & gfServicesOper) && !(flags & gfServicesOper))
				servopers.remove(this);
		}
		
		if (!(db_status() & Record::dbsMember))
			flags &= ~gfServicesRoot;
		
		bool ch_apply = false, ch_unapply = false;
		
		if (!(groups.status() & (Collection::csInitial | Collection::csCreation)))
		{
			if (!(m_flags & gfCustomHost) && (flags & gfCustomHost))
				ch_apply = true;
			if ((m_flags & gfCustomHost) && !(flags & gfCustomHost))
				ch_unapply = true;
		}
		
		m_flags = flags;
		
		if (ch_apply)
			chost_apply();
		if (ch_unapply)
			chost_unapply();
	}
	
	return m_flags;
}

/*************************************************************************/

uint32_t NickGroup::flags_add(uint32_t flags)
{
	if ((m_flags & flags) != flags) 
	{
		uint32_t tflags = (m_flags | flags);
		tflags &= ~gfDefaults;
		
		if (!groups.change_ui(this, f_flags, tflags))
			return m_flags;
		
		if (!(m_flags & gfServicesAdmin) && (flags & gfServicesAdmin))
			servadmins.add(this);
		if (!(m_flags & gfServicesOper) && (flags & gfServicesOper))
			servopers.add(this);
		
		bool ch_apply = false;
		
		if (!(m_flags & gfCustomHost) && (flags & gfCustomHost))
			ch_apply = true;
		
		m_flags = tflags;
		
		if (ch_apply)
			chost_apply();
	}
	
	return m_flags;
}

/*************************************************************************/

uint32_t NickGroup::flags_remove(uint32_t flags)
{
	if ((m_flags & flags) != 0) 
	{
		uint32_t tflags = (m_flags & ~(flags | gfDefaults));
		
		if (!groups.change_ui(this, f_flags, tflags))
			return m_flags;
		
		if ((m_flags & gfServicesAdmin) && (flags & gfServicesAdmin))
			servadmins.remove(this);
		if ((m_flags & gfServicesOper) && (flags & gfServicesOper))
			servopers.remove(this);
		
		bool ch_unapply = false;
		
		if ((m_flags & gfCustomHost) && (flags & gfCustomHost))
			ch_unapply = true;
		
		m_flags = tflags;
		
		if (ch_unapply)
			chost_unapply();
	}
	
	return m_flags;
}

/*************************************************************************/

bool NickGroup::generate_password(char *passbuf, size_t size)
{
#if HAVE_OPENSSL
	size_t rsize = (size - 1) - ((size - 1) % 4);
	rsize *= 3;
	rsize /= 4;
	
	misc::cstring rbuf;
	rbuf.reserve(rsize);
	
	if (RAND_bytes((unsigned char *)rbuf.str(), rsize) != 1 &&
			RAND_pseudo_bytes((unsigned char *)rbuf.str(), rsize) == -1)
		return false;
	
	size = base64_encode(rbuf, rsize, passbuf);
	passbuf[size] = 0;
	
	return change_password(passbuf);
#else
	return false;
#endif
}

/*************************************************************************/

int32_t NickGroup::get_i(Field const *field) const
{
	if (field == f_maxmemos)
		return maxmemos();
	else
		return NameRecord::get_i(field);
}

/*************************************************************************/

Record::identifier_type NickGroup::get_ri(Field const *field) const
{
	if (field == f_name)
		return (m_name ? m_name->id() : m_name_id);
	else
		return NameRecord::get_ri(field);
}

/*************************************************************************/

char const *NickGroup::get_s(Field const *field) const
{
	if (field == f_enc)
		return enc();
	else if (field == f_pass)
		return pass();
	else if (field == f_email)
		return email();
	else if (field == f_url)
		return url();
	else if (field == f_im_type)
		return im_type();
	else if (field == f_im_id)
		return im_id();
	else if (field == f_greet)
		return greet();
	else if (field == f_chost)
		return chost();
	else if (field == f_chost_setby)
		return chost_setby();
	else
		return NameRecord::get_s(field);
}

/*************************************************************************/

uint32_t NickGroup::get_ui(Field const *field) const
{
	if (field == f_flags)
		return flags();
	else if (field == f_auth_code)
		return auth_code();
	else if (field == f_language)
		return language();
	else
		return NameRecord::get_ui(field);
}

/*************************************************************************/

time_t NickGroup::get_ut(Field const *field) const
{
	if (field == f_created)
		return created();
	else if (field == f_chost_changed)
		return chost_changed();
	else
		return NameRecord::get_ut(field);
}

/*************************************************************************/

char const *NickGroup::greet(char const *greet)
{
	if (m_greet.compare(greet, greet_len) != 0)
	{	
		misc::cstring tgreet(greet, greet_len);
		
		if (!groups.change_s(this, f_greet, tgreet))
			return m_greet;
		
		m_greet.swap(tgreet);
	}
	
	return m_greet;
}

/*************************************************************************/

char const *NickGroup::im_id(char const *im_id)
{
	if (m_im_id.compare(im_id, im_id_len) != 0)
	{	
		misc::cstring tim_id(im_id, im_id_len);
		
		if (!groups.change_s(this, f_im_id, tim_id))
			return m_im_id;
		
		m_im_id.swap(tim_id);
	}
	
	return m_im_id;
}

/*************************************************************************/

char const *NickGroup::im_type(char const *im_type)
{
	if (m_im_type.compare(im_type, im_type_len) != 0)
	{	
		misc::cstring tim_type(im_type, im_type_len);
		
		if (!groups.change_s(this, f_im_type, tim_type))
			return m_im_type;
		
		m_im_type.swap(tim_type);
	}
	
	return m_im_type;
}

/*************************************************************************/

uint16_t NickGroup::language(uint16_t language)
{
	if (db_status() & dbsUpdating)
	{
		/* Set to the default language if it's out-of-range */
		if (language >= NUM_LANGS)
			language = epona->cd().nickserv.default_language;
	}
	
	if (m_language != language)
	{
		if (!groups.change_ui(this, f_language, language))
			return m_language;
	
		m_language = language;
	}
	
	return m_language;
}

/*************************************************************************/

bool NickGroup::mail_auth(char const *nick, char const *origin, bool force) const
{
	if (!mail::Mail::configured())
		return false;
	
	char subject[BUFSIZE];
	snprintf(subject, sizeof(subject),
			getstring2(this, NICK_AUTH_MAIL_SUBJECT), nick);
	
	mail::Mail mail;
	
	mail.subject(subject);
	mail.to_name(nick);
	mail.to_address(email());
	mail.origin(origin);
	
	mail.text_add(getstring2(this, NICK_AUTH_MAIL_HEAD));
	mail.text_add("");
	mail.text_addf(getstring2(this, NICK_AUTH_MAIL_LINE_1), s_NickServ);
	mail.text_add("");
	if (epona->cd().nickserv.auth.web_form)
	{
		char url[2048];
		snprintf(url, sizeof(url), epona->cd().nickserv.auth.web_form,
				id(), auth_code());
		mail.text_addf(getstring2(this, NICK_AUTH_MAIL_LINE_2_WEB), url);
	} else {
		mail.text_addf(getstring2(this, NICK_AUTH_MAIL_LINE_2),
				s_NickServ, auth_code());
	}
	mail.text_add("");
	mail.text_addf(getstring2(this, NICK_AUTH_MAIL_LINE_3), nick);
	mail.text_add("");
	mail.text_addf(getstring2(this, NICK_AUTH_MAIL_LINE_4),
			(char const *)epona->cd().services.help_channel);
	mail.text_add("");
	mail.text_addf(getstring2(this, NICK_AUTH_MAIL_LINE_5),
			(char const *)epona->cd().services.network_name);
	
	uint32_t err;
	std::string errdesc;
	
	if (!mail.send((force ? mail::Mail::mfNoQuota : 0), &err, &errdesc))
	{
		if (err & mail::Mail::meQuota)
			epona->log(InfoNickServ(),
					"Could not send authentication e-mail to %s: %s",
					email(), errdesc.c_str());
		else
			epona->log(ErrorNickServ(),
					"Could not send authentication e-mail to %s: %s",
					email(), errdesc.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

int16_t NickGroup::maxmemos(int16_t maxmemos)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't allow out-of-range max */
		if (maxmemos < -2)
			maxmemos = 32767;
		/* Set to the default maximum if it's -2 */
		if (maxmemos == -2)
			maxmemos = epona->cd().memoserv.max_memos;
	}
	
	if (m_maxmemos != maxmemos) 
	{
		if (!groups.change_i(this, f_maxmemos, maxmemos))
			return m_maxmemos;
		
		m_maxmemos = maxmemos;
	}
	
	return m_maxmemos;
}

/*************************************************************************/

size_t NickGroup::memory_usage() const
{
	size_t result = NameRecord::memory_usage();
	
	result += sizeof(NickGroup);
	if (m_pass)
		result += strlen(m_pass) + 1;
	if (m_email)
		result += strlen(m_email) + 1;
	if (m_url)
		result += strlen(m_url) + 1;
	if (m_im_type)
		result += strlen(m_im_type) + 1;
	if (m_im_id)
		result += strlen(m_im_id) + 1;
	if (m_greet)
		result += strlen(m_greet) + 1;
	if (m_chost)
		result += strlen(m_chost) + 1;
	if (m_chost_setby)
		result += strlen(m_chost_setby) + 1;

	result += m_aliases.capacity() * sizeof(NickAlias *);
	result += m_accesses.capacity() * sizeof(Access *);
	result += memos.memos().capacity() * sizeof(Memo *);
	
	return result;
}

/*************************************************************************/

char const *NickGroup::name() const
{ 
	return (m_name ? m_name->nick() : "(garbage)"); 
}

/*************************************************************************/

NickAlias *NickGroup::name_a(NickAlias *name_a)
{
	if (m_name != name_a)
	{	
		if (m_name && name_a)
		{
			epona->log(InfoNickServSettings(),
					"Changing %s nickname group name to %s",
					m_name->nick(), name_a->nick());
		}
		
		m_name = name_a;
	}
	
	return m_name;
}

/*************************************************************************/

Record::identifier_type NickGroup::name_id(identifier_type name_id)
{
	/* Allow us to set it to zero when we want ;) */
	if (!(db_status() & dbsUpdating) && !name_id)
		return (m_name_id = 0);
	
	/* We assume (db_status() & dbsUpdating == true) */
	
	NickAlias *na = (name_id ? ::aliases.ifind(name_id) : 0);
	
	/* Do special treatment to IDs zero... */
	if (!name_id)
	{
		/* If there is already a name, return the current ID */
		if (m_name)
			return m_name->id();
		/* Else, let it go that way for now... In
		 * Aliases::end_initial() and Aliases::end_import(),
		 * we will see if we can find it a name, or else
		 * delete it. */
		 return (m_name_id = 0);
	}
	
	/* If a matching alias has been found */
	if (na)
	{
		/* Don't accept to name a group after an alias it
		 * doesn't own ! */
		if (na->ng() != this)
		{
			/* If there is already a name, return the current ID */
			if (m_name)
				return m_name->id();
			/* Else, set the given ID to zero... then wait until
			 * we can try to find a name later! */
			name_id = 0;
		} else 
		{
			name_a(na);
			return m_name->id();
		}
	}
	
	if (m_name_id != name_id)
		m_name_id = name_id;
	
	return m_name_id;
}

/*************************************************************************/

bool NickGroup::on_access(User *u) const
{
	if (m_accesses.empty())
		return false;

	char buf[USERMAX+HOSTMAX+2];
#ifdef HAS_VHOST
	char buf2[USERMAX+HOSTMAX+2];
#endif

	snprintf(buf, sizeof(buf), "%s@%s", u->username, u->host);
#ifdef HAS_VHOST
	if (u->vhost)
		snprintf(buf2, sizeof(buf2), "%s@%s", u->username, u->vhost);
#endif
    
	NickGroup::Access *nga;
	accesses_type::const_iterator it, end;
	
	for (it = m_accesses.begin(), end = m_accesses.end(); it != end; ++it)
	{
		nga = (NickGroup::Access *)(*it);
		
		if (!nga)
			continue;
		
		if (match_wild_nocase(nga->mask(), buf))
			return true;

#ifdef HAS_VHOST
		if (match_wild_nocase(nga->mask(), buf2))
			return true;
#endif
	}

	return false;
}

/*************************************************************************/

char const *NickGroup::pass(char const *pass)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't allow NULL passwords. */
		if (!pass)
		{
			/* If the record is already a member, just keep the old value. */
			if (db_status() & dbsMember)
				return m_pass;
			/* Else make end_update() fail. */
			db_status_add(dbsUpdateError);
			/* Returns the NULL password... =) */
			return m_pass;
		}
	}
	
	if (m_pass.compare(pass, pass_len) != 0)
		m_pass.assign(pass, pass_len);
	
	return m_pass;
}

/*************************************************************************/

void NickGroup::post_deidentified(NickAlias *na)
{
#ifdef HAS_CHGHOST
	if (na->u)
		custom_host_unset(na->u);
#endif
}

/*************************************************************************/

void NickGroup::post_identified(NickAlias *na)
{
#ifdef HAS_CHGHOST
	if (na->u)
		custom_host_set(na->u);
#endif
}

/*************************************************************************/

void NickGroup::post_registered(NickAlias *na)
{
#ifdef HAS_CHGHOST
	if (na->u && nick_identified(na->u))
		custom_host_set(na->u);
#endif
}

/*************************************************************************/

void NickGroup::post_unregistered(NickAlias *na)
{
#ifdef HAS_CHGHOST
	if (na->u)
		custom_host_unset(na->u);
#endif
}

/*************************************************************************/

bool NickGroup::require_auth(User *u, bool force)
{
	if ((flags() & gfAuth) || auth_code() || !email())
		return true;
	
	if (!begin_auth())
		return false;
	
	if (!mail_auth(u->na->nick(), u->host, force))
	{
		end_auth(false);
		return false;
	}
	
	notice_lang(s_NickServ, u, (epona->cd().nickserv.auth.restrict.identify() ?
			NICK_AUTH_START_REQUIRED_NO_IDENTIFY : NICK_AUTH_START_REQUIRED));
	notice_lang(s_NickServ, u, NICK_AUTH_START_DONE, email());
	
	return true;
}

/*************************************************************************/

bool NickGroup::reset_auth()
{
	if (flags_remove(gfAuth) & gfAuth)
		return false;

	return true;
}

/*************************************************************************/

int32_t NickGroup::set_i(Field const *field, int32_t value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_i(field, value);
	
	if (field == f_maxmemos)
		return maxmemos(value);
	else
		return NameRecord::set_i(field, value);
}

/*************************************************************************/

Record::identifier_type NickGroup::set_ri(Field const *field,
		identifier_type value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_ri(field, value);
	
	if (field == f_name)
		return name_id(value);
	else
		return NameRecord::set_ri(field, value);
}

/*************************************************************************/

char const *NickGroup::set_s(Field const *field, char const *value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_s(field, value);
	
	if (field == f_enc)
		return enc(value);
	else if (field == f_pass)
		return pass(value);
	else if (field == f_email)
		return email(value);
	else if (field == f_url)
		return url(value);
	else if (field == f_im_type)
		return im_type(value);
	else if (field == f_im_id)
		return im_id(value);
	else if (field == f_greet)
		return greet(value);
	else if (field == f_chost)
		return chost(value);
	else if (field == f_chost_setby)
		return chost_setby(value);
	else
		return NameRecord::set_s(field, value);
}

/*************************************************************************/

uint32_t NickGroup::set_ui(Field const *field, uint32_t value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_ui(field, value);
	
	if (field == f_flags)
		return flags(value);
	else if (field == f_auth_code)
		return auth_code(value);
	else if (field == f_language)
		return language(value);
	else
		return NameRecord::set_ui(field, value);
}

/*************************************************************************/

time_t NickGroup::set_ut(Field const *field, time_t value)
{
	if (!(db_status() & dbsUpdating))
		return NameRecord::set_ut(field, value);
	
	if (field == f_created)
		return created(value);
	else if (field == f_chost_changed)
		return chost_changed(value);
	else
		return NameRecord::set_ut(field, value);
}

/*************************************************************************/

char const *NickGroup::url(char const *url)
{
	if (m_url.compare(url, url_len) != 0)
	{	
		misc::cstring turl(url, url_len);
		
		if (!groups.change_s(this, f_url, turl))
			return m_url;
		
		m_url.swap(turl);
	}
	
	return m_url;
}

/*************************************************************************/

NickGroup::Access::Access() : databases::NameRecord(), m_owner(0), m_mask()
{
}

/*************************************************************************/

NickGroup::Access::Access(identifier_type id) : databases::NameRecord(id),
		m_owner(0), m_mask()
{
}

/*************************************************************************/

Record::identifier_type NickGroup::Access::get_ri(Field const *field) const
{
	if (field == f_owner)
		return (m_owner ? m_owner->id() : 0);
	else
		return NameRecord::get_ri(field);
}

/*************************************************************************/

char const *NickGroup::Access::get_s(Field const *field) const
{
	if (field == f_mask)
		return mask();
	else
		return NameRecord::get_s(field);
}

/*************************************************************************/


char const *NickGroup::Access::mask(char const *mask)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't allow NULL mask */
		if (!mask)
		{
			/* If there's already a mask, just keep the old. */
			if (m_mask)
				return m_mask;
			/* Else, it is an error */
			db_status_add(dbsUpdateError);
			/* Return the NULL mask... */
			return mask;
		}
		
		/* Do not accept mask changes (but allow case changes)  */
		if (m_mask && stricmp(m_mask, mask) != 0)
			return m_mask;
	}
	
	if (m_mask.compare(mask, mask_len) != 0)
	{	
		misc::cstring tmask(mask, mask_len);
		
		if (!groups.accesses().change_s(this, f_mask, tmask))
			return m_mask; 
		
		m_mask.swap(tmask);
	}
	
	return m_mask;
}

/*************************************************************************/

size_t NickGroup::Access::memory_usage() const
{
	size_t result = NameRecord::memory_usage();
	
	result += sizeof(NickGroup::Access);
	
	result += strlen(m_mask) + 1;
	
	return result;
}

/*************************************************************************/

NickGroup *NickGroup::Access::owner(NickGroup *owner)
{
	if (m_owner)
		return m_owner;
	
	return (m_owner = owner);
}

/*************************************************************************/

Record::identifier_type NickGroup::Access::owner_id(identifier_type owner_id)
{
	/* Assume (db_status() & dbsUpdating == true). */
	
	/* Don't allow owner changes. */
	if (m_owner)
		return m_owner->id();
	
	NickGroup *owner;
	
	/* Don't allow no ownership. */
	if (!owner_id || !(owner = (NickGroup *)groups.find(owner_id)))
	{
		epona->log(ErrorNickServList(), "Ignoring access entry #" RECORD_FORMAT
				" with bad owner #" RECORD_FORMAT, id(), owner_id);
		
		/* Update error. */
		db_status_add(dbsUpdateError);
		/* Don't need to change the value, the record will
		 * be deleted. */
		return owner_id;
	}
	
	m_owner = owner;
	return m_owner->id();
}

/*************************************************************************/

Record::identifier_type NickGroup::Access::set_ri(Field const *field,
		identifier_type value)
{
	if (!(db_status() & dbsUpdating))
		return Record::set_ri(field, value);
	
	if (field == f_owner)
		return owner_id(value);
	else
		return NameRecord::set_ri(field, value);
}

/*************************************************************************/

char const *NickGroup::Access::set_s(Field const *field, char const *value)
{
	if (!(db_status() & dbsUpdating))
		return Record::set_s(field, value);
	
	if (field == f_mask)
		return mask(value);
	else
		return NameRecord::set_s(field, value);
}

/*************************************************************************/

void VectorNA::add(NickAlias *na)
{
	iterator it, end;
	
	/* Make sure the alias is not already in the list */
	for (it = begin(), end = this->end(); it != end; ++it)
		if (*it == na)
			return;
	
	for (it = begin(), end = this->end(); it != end; ++it)
	{	
		if (stricmp(na->nick(), ((NickAlias *)*it)->nick()) < 0)
		{
			insert(it, na);
			return;
		}
	}
	
	push_back(na);
}

/*************************************************************************/

bool VectorNA::remove(NickAlias *na)
{
	iterator it, end;
	
	for (it = begin(), end = this->end(); it != end; ++it)
	{
		if (*it == na)
		{
			erase(it);
			return true;
		}
	}
	
	return false;
}

/*************************************************************************/

void VectorNG::add(NickGroup *ng)
{
	iterator it, end;
	
	/* Make sure the group is not already in the list */
	for (it = begin(), end = this->end(); it != end; ++it)
		if (*it == ng)
			return;
	
	for (it = begin(), end = this->end(); it != end; ++it)
	{
		if (*it == 0)
		{
			*it = ng;
			return;
		}
	}
	
	push_back(ng);
}

/*************************************************************************/

bool VectorNG::remove(NickGroup *ng)
{
	iterator it, end;
	
	for (it = begin(), end = this->end(); it != end; ++it)
	{
		if (*it == ng)
		{
			*it = 0;
			return true;
		}
	}
	
	return false;
}

/*************************************************************************/

