/* OperServ header file.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#ifndef OPERSERV_H
#define OPERSERV_H

/*************************************************************************/

#include <databases/collection.h>
#include <databases/record.h>
#include <misc/string.h>

#include "commands.h"
#include "netbans.h"
#include "nickserv.h" 

#include "services.h"

/*************************************************************************/

namespace databases { class Database; class Field; }
namespace modules { class Module; }

/*************************************************************************/

/* Stats classes */

/// Stats class

class Stats : public databases::Record
{	
	public:
		/* Data */
		static databases::Field *f_maxusers;
		static databases::Field *f_maxusers_time;
	
		/* Functions */
		
		/// Constructor.
		Stats();
		/// Constructor setting the identifier.
		explicit Stats(identifier_type id);
		/// Copy constructor.
		Stats(Stats const &s);
		/// Assignment operator.
		Stats &operator=(Stats const &right);

		/// Overrides Record::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides Record::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);

		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;

		/// Returns the maximum number of users.
		uint32_t maxusers() const { return m_maxusers; }
		/// Sets the maximum number of users.
		uint32_t maxusers(uint32_t maxusers);
	    
	    /// Returns the date of the maximum number of users.
		time_t maxusers_time() const { return m_maxusers_time; }
		/// Sets the date of the maximum number of users.
		time_t maxusers_time(time_t maxusers_time);
		
		/// Sets both maxusers and maxusers_time at once.
		void new_max(uint32_t max, time_t time);
	private:
		uint32_t m_maxusers;     /* Maximum number of users */
		time_t m_maxusers_time;  /* Date of the maximum number of users */
};

/*************************************************************************/

/// The stats collection.

/** This is the stats collection class. */

class StatsList : public databases::Collection
{
	public:
		/// Constructor.
		StatsList() : databases::Collection() { }
	
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		
		/// Overrides Collection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		
		/// Creates the stats record.
		/** Creates the stats record, if it doesn't exist.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool make_record();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
};

/*************************************************************************/

extern StatsList stats;
extern Stats *stats_rec;

extern bool operserv_load();
extern void operserv_notify(modules::Module &sender, int msg, void *arg);
extern void operserv_start();

/*************************************************************************/

/// Network ban record class.

class NBInfo : public databases::NameRecord
{
	public:
		/* Constants */
		
		/// Maximum length of the mask field.
		static size_t const mask_len = 127;
		/// Maximum length of the creator field.
		static size_t const creator_len = 63;
		/// Maximum length of the reason field.
		static size_t const reason_len = 255;
	
		/* Functions */
		
		/// Constructor.
		NBInfo();
		/// Constructor that sets the record identifier.
		explicit NBInfo(identifier_type id);
		/// Destructor.
		~NBInfo();
		
		/// Changes the creator value.
		virtual bool change_creator(char const *value) = 0;
		/// Changes the reason value.
		virtual bool change_reason(char const *value) = 0;
		/// Changes the since value.
		virtual bool change_since(time_t value) = 0;
		/// Changes the expires value.
		virtual bool change_expires(time_t value) = 0;
		
		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Overrides NameRecord::matches().
		virtual bool matches(char const *pattern);
		
		/// Implements NameRecord::name().
		virtual char const *name() const { return m_nb->mask(); }
		
		/// Returns network ban mask.
		char const *mask() const { return m_nb->mask(); }
		/// Sets network ban mask.
		char const *mask(char const *mask);

	    /// Returns the nickname of the person who set the network ban. 
		char const *creator() const { return m_nb->creator(); }
		/// Sets the nickname of the person who set the network ban.
		char const *creator(char const *creator);
	    
	    /// Returns the reason the network ban was set.
		char const *reason() const { return m_nb->reason(); }
		/// Sets the reason the network ban was set.
		char const *reason(char const *reason);
	    
	    /// Returns when the network ban was set.
		time_t since() const { return m_nb->creation_time(); }
		/// Sets when the network ban was set.
		time_t since(time_t since);
	    
	    /// Returns when the network ban expires.
		time_t expires() const { return m_expires; }
		/// Sets when the network ban expires.
		time_t expires(time_t expires);
		
		/// Returns the NetBan pointer.
		NetBan *netban() { return m_nb; }
	protected:
		NetBan *m_nb;
		time_t m_expires;           /* Time when it expires. 0 == no expiry */
	private:
		NBInfo(NBInfo const &);
		NBInfo &operator=(NBInfo const &);
};

/*************************************************************************/

/* Akill classes */

/// The AKILL class

class Akill : public NBInfo
{
	public:
		/* Data */
		static databases::Field *f_mask;
		static databases::Field *f_creator;
		static databases::Field *f_reason;
		static databases::Field *f_since;
		static databases::Field *f_expires;
	
		/* Functions */
		
		/// Constructor.
		Akill(UmaskNetBan *nb);
		/// Constructor that sets the record identifier.
		Akill(UmaskNetBan *nb, identifier_type id);
		
		/// Changes the creator value.
		virtual bool change_creator(char const *value);
		/// Changes the reason value.
		virtual bool change_reason(char const *value);
		/// Changes the since value.
		virtual bool change_since(time_t value);
		/// Changes the expires value.
		virtual bool change_expires(time_t value);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		
		/// Returns the AKILL user part.
		char const *user() const { return ((UmaskNetBan *)m_nb)->user(); }
		/// Returns the AKILL host part.
		char const *host() const { return ((UmaskNetBan *)m_nb)->host(); }
};

/*************************************************************************/

/// The akill collection.

/** This is the akill collection class. */

class AkillList : public databases::ArrayCollection
{
	public:
		/// Constructor.
		AkillList();
	
		/// Implements ArrayCollection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		
		/// Expires AKILLs.
		void expire();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Is the banner available?
		bool available() const { return (m_banner != 0); }
	protected:
		/// Overrides ArrayCollection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides ArrayCollection::remove_list().
		virtual void remove_list(databases::Record *record);
		/// Overrides ArrayCollection::clear_list().
		virtual void clear_list();
	private:
		UmaskNetBanner *m_banner;
};

/*************************************************************************/

class AkillCmd : public ListCmd
{
	public:
		/// Constructor.
		AkillCmd();
		
		/// Adds a record to the AKILL list.
		/** Adds a record to the AKILL list.
		  *
		  * \param[in] u The user issuing the command. May be NULL, in
		  * which case no feedback is issued.
		  * \param[in] mask The AKILL mask.
		  * \param[in] creator The person who sets the AKILL.
		  * \param[in] expires The expiration time of the AKILL.
		  * \param[in] reason The reason of the AKILL.
		  * \return 0 or the number of obsoleted AKILL entries that
		  * have been deleted on success, -1 on failure, -2 if only
		  * the expiration time of an existing AKILL entry has been
		  * changed. */
		int add_record(User *u, char const *mask, char const *creator,
				time_t expires, char const *reason);
	protected:
		/// Overrides ListCmd::container().
		virtual container_type *container(User *u);
		
		/// Overrides ListCmd::dispatch().
		virtual int dispatch(User *u, container_type &ctn, char const *sub);
		/// Overrides ListCmd::changeable().
		virtual bool changeable(User *u, container_type &ctn,
				char const *sub) const;
		
		/// Overrides ListCmd::add().
		virtual int add(User *u, container_type &ctn, char const *sub);
		
		/// Overrides ListCmd::del_record().
		virtual bool del_record(User *u, container_type &ctn,
				char const *sub, databases::Record *rec);
		
		/// Overrides ListCmd::list_record().
		virtual bool list_record(User *u, container_type &ctn,
				char const *sub, uint16_t num, databases::Record const *rec,
				uint16_t count);
		
		/// Overrides ListCmd::clear_records().
		virtual bool clear_records(User *u, container_type &ctn,
				char const *sub);
};

/*************************************************************************/

/* Akill declarations */

extern AkillList akills;
extern AkillCmd *akill_cmd;

/*************************************************************************/

/* SGLINE classes */

/// The SGLINE class.

class SGLine : public NBInfo
{
	public:
		/* Data */
		static databases::Field *f_mask;
		static databases::Field *f_creator;
		static databases::Field *f_reason;
		static databases::Field *f_since;
		static databases::Field *f_expires;
	
		/* Functions */
		
		/// Constructor.
		SGLine(RealNetBan *nb);
		/// Constructor that sets the record identifier.
		SGLine(RealNetBan *nb, identifier_type id);
		
		/// Changes the creator value.
		virtual bool change_creator(char const *value);
		/// Changes the reason value.
		virtual bool change_reason(char const *value);
		/// Changes the since value.
		virtual bool change_since(time_t value);
		/// Changes the expires value.
		virtual bool change_expires(time_t value);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
};

/*************************************************************************/

/// The SGLINE collection.

/** This is the SGLINE collection class. */

class SGLineList : public databases::ArrayCollection
{
	public:
		/// Constructor.
		SGLineList();
	
		/// Implements ArrayCollection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		
		/// Expires SGLINEs.
		void expire();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Is the banner available?
		bool available() const { return (m_banner != 0); }
	protected:
		/// Overrides ArrayCollection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides ArrayCollection::remove_list().
		virtual void remove_list(databases::Record *record);
		/// Overrides ArrayCollection::clear_list().
		virtual void clear_list();
	private:
		RealNetBanner *m_banner;
};

/*************************************************************************/

extern SGLineList sglines;

/*************************************************************************/

/* SQLINE classes */

/// The SQLINE class

class SQLine : public NBInfo
{
	public:
		/* Data */
		static databases::Field *f_mask;
		static databases::Field *f_creator;
		static databases::Field *f_reason;
		static databases::Field *f_since;
		static databases::Field *f_expires;
	
		/* Functions */
		
		/// Constructor.
		SQLine(NickChanNetBan *nb);
		/// Constructor that sets the record identifier.
		SQLine(NickChanNetBan *nb, identifier_type id);
		
		/// Changes the creator value.
		virtual bool change_creator(char const *value);
		/// Changes the reason value.
		virtual bool change_reason(char const *value);
		/// Changes the since value.
		virtual bool change_since(time_t value);
		/// Changes the expires value.
		virtual bool change_expires(time_t value);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
};

/*************************************************************************/

/// The SQLine collection.

/** This is the SQLINE collection class. */

class SQLineList : public databases::ArrayCollection
{
	public:
		/// Constructor.
		SQLineList();
	
		/// Implements ArrayCollection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		
		/// Expires SQLINEs.
		void expire();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
	protected:
		/// Overrides ArrayCollection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides ArrayCollection::remove_list().
		virtual void remove_list(databases::Record *record);
		/// Overrides ArrayCollection::clear_list().
		virtual void clear_list();
	private:
		NickNetBanner *m_nbanner;
		ChanNameNetBanner *m_cbanner;
};

/*************************************************************************/

extern SQLineList sqlines;

/*************************************************************************/

/* SZLINE classes */

/// The SZLINE class.

class SZLine : public NBInfo
{
	public:
		/* Data */
		static databases::Field *f_mask;
		static databases::Field *f_creator;
		static databases::Field *f_reason;
		static databases::Field *f_since;
		static databases::Field *f_expires;
	
		/* Functions */
		
		/// Constructor.
		SZLine(IPNetBan *nb);
		/// Constructor that sets the record identifier.
		SZLine(IPNetBan *nb, identifier_type id);
		
		/// Changes the creator value.
		virtual bool change_creator(char const *value);
		/// Changes the reason value.
		virtual bool change_reason(char const *value);
		/// Changes the since value.
		virtual bool change_since(time_t value);
		/// Changes the expires value.
		virtual bool change_expires(time_t value);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
};

/*************************************************************************/

/// The SZLINE collection.

/** This is the SZLINE collection class. */

class SZLineList : public databases::ArrayCollection
{
	public:
		/// Constructor.
		SZLineList();
	
		/// Implements ArrayCollection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		
		/// Expires SZLINEs.
		void expire();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Is the banner available?
		bool available() const { return (m_banner != 0); }
	protected:
		/// Overrides ArrayCollection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides ArrayCollection::remove_list().
		virtual void remove_list(databases::Record *record);
		/// Overrides ArrayCollection::clear_list().
		virtual void clear_list();
	private:
		IPNetBanner *m_banner;
};

/*************************************************************************/

extern SZLineList szlines;

/*************************************************************************/

/* News classes */

/// News class

class News : public databases::Record
{	
	public:
		/* Data */
		static databases::Field *f_type;
		static databases::Field *f_text;
		static databases::Field *f_creator;
		static databases::Field *f_since;
	
		/* Constants */
		
		/// News types.
		enum
		{
			/// Logon news 
			ntLogon = 1,
			/// Oper news
			ntOper = 2,
			/// Random news
			ntRandom = 3,
			/// First news type allowed
			ntFirst = 1,
			/// Last news type allowed
			ntLast = 3
		};
		
		/// Maximum length of the text field.
		static size_t const text_len = 300;
		/// Maximum length of the creator field.
		static size_t const creator_len = 63;
		
		/* Functions */
		
		/// Constructor.
		News();
		/// Constructor setting the identifier.
		explicit News(identifier_type id);
		/// Copy constructor.
		News(News const &n);
		/// Assignment operator.
		News &operator=(News const &news);

		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides Record::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);

		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;

		/// Returns the type of the news.
		uint8_t type() const { return m_type; }
		/// Sets the type of the news.
		uint8_t type(uint8_t type);

		/// Returns the text of the news.
		char const *text() const { return m_text; }
		/// Sets the text of the news.
		char const *text(char const *text);
	    
	    /// Returns the nick of the person who created the news.
		char const *creator() const { return m_creator; }
		/// Sets the nick of the person who created the news.
		char const *creator(char const *creator);
	    
	    /// Returns the creation date of the news.
		time_t since() const { return m_since; }
		/// Sets the creation date of the news.
		time_t since(time_t since);
	private:
		uint8_t m_type;           /* Type of the news */
		misc::cstring m_text;     /* Text of the news */
		misc::cstring m_creator;  /* Nick of person who created the news */
		time_t m_since;           /* Creation date of the news */
};

/*************************************************************************/

/// The news collection.

/** This is the news collection class. */

class NewsList : public databases::ArrayCollection
{
	public:
		/// Constructor.
		NewsList() : databases::ArrayCollection(false) { }
	
		/// Implements ArrayCollection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
};

/*************************************************************************/

/* News declarations */

extern NewsList news;

extern bool news_load();
extern void news_notify(modules::Module &sender, int msg, void *arg);
extern void news_start();

extern void display_news(User *u, uint16_t type);
extern void do_news(User *u);
extern void do_oldnews(User *u);

/*************************************************************************/

/* Sessions classes */

struct Session {
    Session *prev, *next;
    char *host;
    int count;			/* Number of clients with this host */
    int hits;			/* Number of subsequent kills for a host */
};

/*************************************************************************/

/// Exception record class.

class Exception : public databases::NameRecord
{
	public:
		/* Data */
		static databases::Field *f_mask;
		static databases::Field *f_limit;
		static databases::Field *f_creator;
		static databases::Field *f_reason;
		static databases::Field *f_since;
		static databases::Field *f_expires;
		
		/* Constants */
		
		/// Maximum length of the mask field.
		static size_t const mask_len = 63;
		/// Maximum length of the creator field.
		static size_t const creator_len = 63;
		/// Maximum length of the reason field.
		static size_t const reason_len = 255;
	
		/* Functions */
		
		/// Constructor.
		Exception();
		/// Constructor that sets the record identifier.
		explicit Exception(identifier_type id);
		/// Copy constructor.
		Exception(Exception const &ex);
		/// Assignment operator.
		Exception &operator=(Exception const &right);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides Record::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		
		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Implements NameRecord::name().
		virtual char const *name() const { return m_mask; }
		
		/// Returns exception mask.
		char const *mask() const { return m_mask; }
		/// Sets exception mask.
		char const *mask(char const *mask);
	    
	    /// Returns exception limit.
		uint16_t limit() const { return m_limit; }
		/// Sets exception limit.
		uint16_t limit(uint16_t limit);
	    
	    /// Returns the nickname of the person who set the exception. 
		char const *creator() const { return m_creator; }
		/// Sets the nickname of the person who set the exception.
		char const *creator(char const *creator);
	    
	    /// Returns the reason the exception was set.
		char const *reason() const { return m_reason; }
		/// Sets the reason the exception was set.
		char const *reason(char const *reason);
	    
	    /// Returns when the exception was set.
		time_t since() const { return m_since; }
		/// Sets when the exception was set.
		time_t since(time_t since);
	    
	    /// Returns when the exception expires.
		time_t expires() const { return m_expires; }
		/// Sets when the exception expires.
		time_t expires(time_t expires);
	private:
		misc::cstring m_mask;       /* Hosts to which this exception applies */
		uint16_t m_limit;           /* Session limit for exception */
		misc::cstring m_creator;    /* Nick of person who added the exception */
		misc::cstring m_reason;     /* Reason for exception's addition */
		time_t m_since;             /* When this exception was added */
		time_t m_expires;           /* Time when it expires. 0 == no expiry */
};

/*************************************************************************/

/// The exception collection.

/** This is the exception collection class. It has a method to find a
  * match in the collection list easily. */

class ExceptionList : public databases::ArrayCollection
{
	public:
		/// Constructor.
		ExceptionList() : databases::ArrayCollection(true) { }
	
		/// Implements ArrayCollection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		
		/// Expires an exception
		/** Deletes an exception that has been determined to have
		  * expired.
		  *
		  * \param[in] exception The exception that expired.
		  * \return true if successful, false otherwise. */
		bool expire(Exception *ex);
		
		/// Finds an exception matching a hostname.
		/** Finds an exception matching the given hostname.
		  *
		  * \param[in] host The hostname.
		  * \return An exception matching the hostname, or 0 if
		  * no exception matched. */
		Exception *find_match(char const *host);
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
};

/*************************************************************************/

LOGLEVELH(InfoOperServ, xlog::LogInfo, "Info::OperServ", "operserv: ");
LOGLEVELH(InfoOperServList, InfoOperServ, "Info::OperServ::List", "operserv list: ");

LOGLEVELH(ErrorOperServ, xlog::LogError, "Error::OperServ",
		"Error: operserv: ");

/*************************************************************************/

/* Sessions declarations */

extern ExceptionList exceptions;

extern void get_session_stats(long *nrec, long *memuse);

extern void do_session(User *u);
extern int add_session(const char *nick, const char *host);
extern void del_session(const char *host);

extern bool exceptions_load();
extern void exceptions_notify(modules::Module &sender, int msg, void *arg);
extern void exceptions_start();
extern void do_exception(User *u);

/*************************************************************************/

extern VectorNG servadmins;
extern VectorNG servopers;

E void operserv(User *u, char *buf);

E void os_init(void);

E int is_services_root(User *u);
E int is_services_admin(User *u);
E int is_services_oper(User *u);
E int nick_is_services_admin(NickGroup *ng);
E int nick_is_services_oper(NickGroup *ng);

/*************************************************************************/

#endif /* OPERSERV_H */
